
/***************************************************************************

    68K.C (formerly GEM.C)

    - 68000/ST virtual machine routines

    Copyright (C) 1991-2008 by Darek Mihocka. All Rights Reserved.
    Branch Always Software. http://www.emulators.com/

    11/30/2008  darekm      Gemulator 9.0 release
    04/20/2008  darekm      HRESULT based memory accessors
    08/26/2007  darekm      video blit uses new memory API

***************************************************************************/

#include "gemtypes.h"

#ifdef ATARIST

#ifndef NDEBUG

#define  TRACEHW   0
#define  TRACEDISK 1

// enables DEBUG output
BOOL fLogTraps = TRUE;
BOOL fLogHWAccess = FALSE;

#else

#define  TRACEHW   0
#define  TRACEDISK 0

#endif


//
// Atari ST specific flags
//

typedef union _srgf
    {
 struct
    {
    unsigned :1;
    unsigned :1;
    unsigned :1;
    unsigned :1;
    unsigned :1;
    unsigned :1;
    unsigned :1;
    unsigned :1;
    unsigned :1;
    unsigned :1;
    unsigned :1;
    unsigned :1;
    unsigned :1;
    unsigned fSTE:2;
    unsigned fVBIPending:1;

    unsigned fCentronics:1;
    unsigned fRS232:1;
    unsigned fRS232CTS:1;
    unsigned fReserved:1;
    unsigned fTimerD:1;
    unsigned fTimerC:1;
    unsigned fKeyMIDI:1;
    unsigned fFDCHDC:1;

    unsigned fTimerB:1;
    unsigned fSenderError:1;
    unsigned fSenderBuffEmpty:1;
    unsigned fReceiveError:1;
    unsigned fReceiveBuffFull:1;
    unsigned fTimerA:1;
    unsigned fRS232Ring:1;
    unsigned fMonoDetect:1;
    };
 struct
    {
    unsigned char x;
    unsigned char y;
    unsigned char B;
    unsigned char A;
    };
    } RGF;

RGF  vrgf;

static  BOOL  __cdecl   Stub();
static  BOOL  __cdecl   Install();
static  BOOL  __cdecl   Init();
static  BOOL  __cdecl   UnInit();
static  BOOL  __cdecl   InitDisks();
static  BOOL  __cdecl   MountDisk(int i);
static  BOOL  __cdecl   UnInitDisks();
static  BOOL  __cdecl   UnmountDisk(int i);
static  ULONG __cdecl   VPollInterrupts();
static  ULONG __cdecl   Reset();
static  ULONG __cdecl   DumpHW();
static  ULONG __cdecl   WinMsg();
static  ULONG __cdecl   Exec(BOOL, BOOL);
static  HRESULT __cdecl   ReadHWByte(ULONG ea, BYTE *);
static  HRESULT __cdecl   ReadHWWord(ULONG ea, WORD *);
static  HRESULT __cdecl   ReadHWLong(ULONG ea, ULONG *);
static  HRESULT __cdecl   WriteHWByte(ULONG ea, BYTE *);
static  HRESULT __cdecl   WriteHWWord(ULONG ea, WORD *);
static  HRESULT __cdecl   WriteHWLong(ULONG ea, ULONG *);
static  PBYTE __cdecl   MapAddress(ULONG ea);


#ifdef MAKEDLL
_declspec(dllexport)
#endif
VMINFO const vmiST =
{
    NULL,
    0,
    "Atari ST / Mega ST / STE",
    vmAtariST | vmAtariSTE,
    cpu68000 | cpu68010 | cpu68020 | cpu68030 | cpu68040,
    ram512K | ram1M | ram2M | ram2_5M | ram4M | ram8M | ram14M,
    osTOS3X_4 |
    osTOS1X_2 | osTOS1X_6 | osTOS2X_2 | osTOS1X_D | osMagic_2 | osMagic_4 | \
                osMagicPC | osTOS2X_D | osTOS3X_D | osTOS4X_D,
    monSTColor | monSTMono,
    Install,
    Init,
    UnInit,
    InitDisks,
    MountDisk,
    UnInitDisks,
    UnmountDisk,
    Stub,
    Reset,
    Exec,
    Stub,
    WinMsg,
    m68k_DumpRegs,
    DumpHW,
    Stub,
    ReadHWByte,
    ReadHWWord,
    ReadHWLong,
    WriteHWByte,
    WriteHWWord,
    WriteHWLong,
    NULL,
    NULL,
    MapAddress,
    MapAddress,
    NULL,
    NULL,
};

BOOL __cdecl Stub()
{
    return TRUE;
}

BOOL __cdecl Install(PVMINFO pvmi, PVM pvm)
{
    // initialize the default Atari ST VM

    pvm->pvmi = pvmi;
    pvm->bfHW = BfFromWfI(pvmi->wfHW, 0);
    pvm->bfCPU = BfFromWfI(pvmi->wfCPU, 0);
    pvm->iOS = -1;
    pvm->bfMon = BfFromWfI(pvmi->wfMon, 0);
    pvm->bfRAM  = BfFromWfI(pvmi->wfRAM, 2); // 2M default
    pvm->ivdMac = sizeof(v.rgvm[0].rgvd)/sizeof(VD);

    pvm->rgvd[0].dt = DISK_FLOPPY;

    pvm->fValidVM = fTrue;

    return TRUE;
}

BOOL __cdecl Init()
{
    BOOL f;

    SetCurrentDirectory(vi.szDefaultDir);

    // Wipe away any stale 680x0 state

    memset((BYTE *)&vregs,  0, sizeof(REGS));
    memset((BYTE *)&vi.regs2, 0, sizeof(REGS2));

    // HACK: set fSTE to 0 = ST, 1 = STE, 2 = TT

    vrgf.fSTE = osCur.version >= 0x0106;
    vmCur.bfHW = vrgf.fSTE ? vmAtariSTE : vmAtariST;
    vmCur.ivdMac = sizeof(v.rgvm[0].rgvd)/sizeof(VD);
    vi.fFake040 = FALSE;

    if (osCur.version >= 0x0300)
        {
        // TT ROMs require a 68030

        vmCur.bfCPU = cpu68030;

        vrgf.fSTE = 2;
        vmCur.bfHW = vmAtariTT;
        }

    // Enable blitter when emulating STE or TT

    vmCur.fBlitter = vrgf.fSTE != 0;

    // Set up ROM and RAM address ranges

    vi.eaRAM[0] = 0;                // 4M main RAM
    vi.cbRAM[0] = CbRamFromBf(vmCur.bfRAM);

    switch (vrgf.fSTE)
        {
    default:
        vi.eaROM[0] = 0x00FC0000;   // 192K TOS ROM
        vi.cbROM[0] = 192*1024;
        break;

    case 1:
        vi.eaROM[0] = 0x00E00000;   // 256K TOS ROM
        vi.cbROM[0] = 256*1024;
        break;

    case 2:
        vi.eaROM[0] = 0x00E00000;   // 512K TOS ROM
        vi.cbROM[0] = 512*1024;
        break;
        }

    vi.eaROM[1] = 0x00FA0000;       // 128K cartridge ROM
    vi.cbROM[1] = 128*1024;

    // Don't allow illegal CPU type

    if ((vmCur.bfCPU & vmCur.pvmi->wfCPU) == 0)
        vmCur.bfCPU = BfFromWfI(vmCur.pvmi->wfCPU, 0);

    // initialize emulator globals
    //
    // vi.cbRAM, vi.cbROM, and vi.eaROM must have already been set

    f = memAllocAddressSpace(TRUE);

#if !defined(NDEBUG)
    printf("vpregs:%08X, cb:%08X, pbRAM:%08x, pbROM:%08X\n",
             vpregs, vi.cbAlloc, vi.pbRAM[0], vi.pbROM[0]);
#endif

    if (!f)
        {
        MessageBox(NULL, "Unable to allocate Atari memory. "
            "Reduce the Memory setting or increase the Windows swap file size.",
            vi.szAppName, MB_OK);

        memFreeAddressSpace();
        return FALSE;
        }

    // allocate profile and history buffers last to protect the previous allocs

    vi.pProfile = malloc(65536 * sizeof(DWORD));  // profile array;
    vi.pHistory = malloc(MAX_HIST*4);  // history array;

    if (NULL != vi.pProfile)
        memset(vi.pProfile, 0, 65536 * sizeof(DWORD));

    if (NULL != vi.pHistory)
        memset(vi.pHistory, 0, MAX_HIST*4);

    vpci = &cpi68K;

    vpci->pfnInit(
        vpregs,
        ReadHWByte,
        ReadHWWord,
        WriteHWByte,
        WriteHWWord,
        VPollInterrupts,
        Reset,
        vmCur.bfCPU);

    vpci->rgfExcptHook[0] =  0;
    vpci->rgfExcptHook[1] =  0;

    if (!vmCur.fUseVHD)
        vpci->rgfExcptHook[1] += 0x0002; // hook TRAP #1 for GEMDOS
 // vpci->rgfExcptHook[1] += 0x0004; // hook TRAP #2
    vpci->rgfExcptHook[0] += 0x0004; // hook Bus Error vector
    vpci->rgfExcptHook[0] += 0x0008; // hook Address Error vector
    vpci->rgfExcptHook[0] += 0x0010; // hook Illegal Instr. vector

    if (vrgf.fSTE > 1)
        vpci->rgfExcptHook[0] += 0x0800; // hook Line F vector (for 68020+)

    // set address mask and invalidate the whole page table

    vsthw.lAddrMask = 0x00FFFFFF; // 24-bit addressing
    InvalidatePages(0x00000000, 0xFFFFFFFF);

    // Initialize the ROMs

    if (!FInstallROMs())
        {
        MessageBox(NULL, "Unable to install TOS. "
            "Please run the ROM BIOS Setup to configure TOS on your computer.",
            vi.szAppName, MB_OK);
        return FALSE;
        }

    PatchTOS(osCur.version, vi.eaROM[0]);

    // Start the PC at the start of the ROMs

    vpregs->PC = vmPeekL(vi.eaROM[0] + 8);  // start PC at first byte of TOS
    vpregs->SSP = vi.cbRAM[0] - 32768;
    vpregs->USP = vi.cbRAM[0] - 32768;
    vpregs->A7 = vi.cbRAM[0] - 32768;
    vpregs->SR = 0x2700;

    // Initialize disk image files

    InitDisks();

    // Initialize the VM

    Reset();

    ClearProfile();

#ifndef NDEBUG
    DebugStr("pp = %08X, ph = %08X\n", vi.pProfile, vi.pHistory);
    DebugStr("vi.cbROM[0] = %08X  ", vi.cbROM[0]);
    DebugStr("vi.cbRAM[0] = %08X  ", vi.cbRAM[0]);
    DebugStr("vi.pregs    = %08X  ", vi.pregs);
    DebugStr("vi.pbRAM[0] = %08X\n", vi.pbRAM[0]);

    DebugStr("PC = %08X  ", vpregs->PC);
    DebugStr("SR = %08X  ", vpregs->SR);
    DebugStr("A7 = %08X\n", vpregs->A7);

    DebugStr("&D0 = %08X\n", &vpregs->D0);
#endif

    MarkAllPagesDirty();

    return TRUE;
}


BOOL __cdecl UnInit()
{
    UnInitDisks();

    memFreeAddressSpace();

    vi.fFake040 = FALSE;

    return TRUE;
}


//
// Set up the disk image files. This function is separate from the Init routine so
// that it can be called on the fly as the user changes disk settings on the fly.
//

BOOL __cdecl MountDisk(int i)
{
    PVD pvd = &vmCur.rgvd[i];

    vi.rgpdi[i] = NULL;

    // when opening a floppy don't bother to test the file system
    // but all other devives you have to (in case of boot partitions
    // and multiple partitions in general)

    if (pvd->dt == DISK_IMAGE)
        {
        // First open the disk image with file system checking
        // since we are assuming this is a .VHD file or floppy image.

        vi.rgpdi[i] = PdiOpenDisk(DISK_IMAGE, pvd->sz, (pvd->mdWP ? DI_READONLY : 0));

        if ((i >= 2) && vi.rgpdi[i] && (vi.rgpdi[i]->fst == FS_FAT16)
                 && vi.rgpdi[i]->offsec != 0)
        //  if (vmCur.fUseVHD)
            {
            // The boot sector is not the first sector, so this is
            // a raw disk image. Re-open it raw.

            CloseDiskPdi(vi.rgpdi[i]);

            vi.rgpdi[i] = PdiOpenDisk(DISK_IMAGE, pvd->sz,
                (pvd->mdWP ? DI_READONLY : 0) | DI_QUERY);
            }
        }
    else if (pvd->dt == DISK_FLOPPY)
        {
        vi.rgpdi[i] = PdiOpenDisk(DISK_FLOPPY, pvd->id,
            (pvd->mdWP ? DI_READONLY : 0) | DI_QUERY);

        // Atari ST uses track/side/sector, so override any size
        // detection and set it to 0

        if (vi.rgpdi[i])
            vi.rgpdi[i]->size = 0;
        }
    else if (pvd->dt == DISK_SCSI)
        {
        vi.rgpdi[i] = PdiOpenDisk(DISK_SCSI,
             LongFromScsi(pvd->ctl, pvd->id),
             (pvd->mdWP ? DI_READONLY : 0) | DI_QUERY);
        }

    return TRUE;
}

BOOL __cdecl InitDisks()
{
    int i;

    for (i = 0; i < vmCur.ivdMac; i++)
        MountDisk(i);

    return TRUE;
}


//
// Close any open disk image file handles
//

BOOL __cdecl UnmountDisk(int i)
{
    if (vi.rgpdi[i])
        CloseDiskPdi(vi.rgpdi[i]);

    vi.rgpdi[i] = NULL;

    return TRUE;
}

BOOL __cdecl UnInitDisks()
{
    int i;

    for (i = 0; i < vmCur.ivdMac; i++)
        UnmountDisk(i);

    // flush any open files

    FResetGemDosHooks();

    return TRUE;
}


void ClearProfile()
{
    ULONG *pl = vi.pProfile;
    int cl = 65536;

    while (cl--)
        *pl++ = 0x00;
}


//
// Initialize low memory TOS globals for boot and
// also apply any ROM patches as necessary
//

void PatchTOS(version, lLoadAddr)
int version;
ULONG lLoadAddr;
{
    unsigned i;

    // MMU copies first 8 bytes of ROM to address 0
    PokeL(0x000, vmPeekL(lLoadAddr));
    PokeL(0x004, vmPeekL(lLoadAddr+4));

    // set up some low memory variables used during boot
    PokeL(0x420, 0x752019f3); // memvalid
    PokeB(0x424, 10);         // memctrl
    PokeL(0x426, 0);          // resvalid
    PokeL(0x42A, 0);          // resvector
    PokeL(0x42E, vi.cbRAM[0]); // phystop **
    PokeL(0x432, 0x0003B900); // _membot
    PokeL(0x436, vi.cbRAM[0] - 32768); // _memtop **
    PokeL(0x43A, 0x237698aa); // memval2
    PokeW(0x440, 3);          // seekrate
    PokeW(0x442, 0x14);       // _timer_ms
    PokeW(0x444, 0xFF);       // _fverify
    PokeW(0x446, (WORD)vmCur.iBootDisk);// _bootdev
    PokeW(0x448, 0);          // palmode
    PokeW(0x44A, 0);          // defshiftmod
    PokeB(0x44C, 2);          // sshiftmod
    PokeL(0x44E, vi.cbRAM[0] - 32768); // v_bas_ad **
    PokeW(0x452, 1);          // vblsem
    PokeW(0x462, 0);          // vbclock
    PokeW(0x466, 0);          // _frclock
    PokeW(0x482, 0);          // _comload
    PokeW(0x4A6, 2);          // _nflops
    PokeW(0x4BA, 0);          // _hz_200
    PokeL(0x4C2, 0x00000003); // _drvbits
    PokeW(0x4EE, 0xFFFF);     // _dumpflg
    PokeW(0x4F0, 0);          // _prtabt
    PokeL(0x4F2, lLoadAddr);  // _sysbase
    PokeL(0x4F6, 0);          // _shell_p
    PokeL(0x51A, 0x5555AAAA); // memval3 ???

    // patch TOS for now
    switch (version)
        {
    default:
        break;

    case 0x8200:  // MagiC 2.0
        lLoadAddr = vpregs->PC;

        PokeL(0x000, vmPeekL(lLoadAddr));
        PokeL(0x004, vmPeekL(lLoadAddr+4));
        PokeL(0x008, 0xE00030);

        PokeL(0x426, 0x31415926); // resvalid
        PokeL(0x42A, lLoadAddr);  // resvector
        PokeL(0x432, 0x0002cc58); // _membot
        PokeL(0x486, 0x602e2000); // tr14ret
        PokeL(0x492, 0x0002cc58); // themd
        PokeL(0x496, 0x003cb3a8); // themd
        PokeL(0x49E, 0x01000000); // _md
        PokeL(0x4F2, lLoadAddr);  // _sysbase
        PokeL(0x4FA, 0x0002cc58); // end_os
        PokeL(0x51A, 0x5555AAAA);

        for (i = 0; i < 262144; i+=4)
            memWriteLongROM(0xE00000 + i, vmPeekL(lLoadAddr + i));
        break;

    case 0x0104:  // TOS 1.04 in ROM
    case 0x0200:  // TOS 2.06 on disk
    case 0x0206:  // TOS 2.06 in ROM
    case 0x0306:  // TOS 3.06 in ROM
        break;
        }

}


typedef struct data16
{
    unsigned short off;
    unsigned char rgb[16];
} data16;

data16 const d16HDX[] =
{
    { 0x0000,(BYTE)0x60,(BYTE)0x04,(BYTE)0x60,(BYTE)0x00,(BYTE)0x00,(BYTE)0xb2,(BYTE)0x08,(BYTE)0xb8,(BYTE)0x00,(BYTE)0x02,(BYTE)0x04,(BYTE)0xc5,(BYTE)0x31,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x00 },
    { 0x0010,(BYTE)0x04,(BYTE)0x46,(BYTE)0x48,(BYTE)0xe7,(BYTE)0x14,(BYTE)0x00,(BYTE)0x3f,(BYTE)0x3c,(BYTE)0xff,(BYTE)0xff,(BYTE)0x3f,(BYTE)0x3c,(BYTE)0x00,(BYTE)0x0b,(BYTE)0x4e,(BYTE)0x4d },
    { 0x0020,(BYTE)0x58,(BYTE)0x4f,(BYTE)0x08,(BYTE)0x00,(BYTE)0x00,(BYTE)0x03,(BYTE)0x66,(BYTE)0x42,(BYTE)0x70,(BYTE)0x03,(BYTE)0x72,(BYTE)0xf8,(BYTE)0x41,(BYTE)0xfa,(BYTE)0x01,(BYTE)0x98 },
    { 0x0030,(BYTE)0x0c,(BYTE)0x83,(BYTE)0x44,(BYTE)0x4d,(BYTE)0x41,(BYTE)0x72,(BYTE)0x67,(BYTE)0x06,(BYTE)0x78,(BYTE)0xff,(BYTE)0x74,(BYTE)0x80,(BYTE)0x60,(BYTE)0x04,(BYTE)0x14,(BYTE)0x05 },
    { 0x0040,(BYTE)0x67,(BYTE)0x16,(BYTE)0x08,(BYTE)0x10,(BYTE)0x00,(BYTE)0x00,(BYTE)0x67,(BYTE)0x06,(BYTE)0xc3,(BYTE)0x10,(BYTE)0xb4,(BYTE)0x10,(BYTE)0x67,(BYTE)0x36,(BYTE)0xd0,(BYTE)0xfc },
    { 0x0050,(BYTE)0x00,(BYTE)0x0c,(BYTE)0x51,(BYTE)0xc8,(BYTE)0xff,(BYTE)0xee,(BYTE)0x60,(BYTE)0x12,(BYTE)0x08,(BYTE)0x10,(BYTE)0x00,(BYTE)0x00,(BYTE)0x67,(BYTE)0x04,(BYTE)0xc3,(BYTE)0x10 },
    { 0x0060,(BYTE)0x66,(BYTE)0x22,(BYTE)0xd0,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x0c,(BYTE)0x51,(BYTE)0xc8,(BYTE)0xff,(BYTE)0xf0,(BYTE)0x50,(BYTE)0x4f,(BYTE)0x20,(BYTE)0x78,(BYTE)0x04,(BYTE)0xf2 },
    { 0x0070,(BYTE)0x20,(BYTE)0x28,(BYTE)0x00,(BYTE)0x18,(BYTE)0x48,(BYTE)0x40,(BYTE)0xb0,(BYTE)0xbc,(BYTE)0x19,(BYTE)0x87,(BYTE)0x04,(BYTE)0x22,(BYTE)0x64,(BYTE)0x04,(BYTE)0x3e,(BYTE)0x3c },
    { 0x0080,(BYTE)0x00,(BYTE)0xe0,(BYTE)0x4e,(BYTE)0x75,(BYTE)0x2c,(BYTE)0x28,(BYTE)0x00,(BYTE)0x04,(BYTE)0x7a,(BYTE)0x01,(BYTE)0x49,(BYTE)0xfa,(BYTE)0x01,(BYTE)0x74,(BYTE)0x61,(BYTE)0x26 },
    { 0x0090,(BYTE)0x4a,(BYTE)0x40,(BYTE)0x66,(BYTE)0xd6,(BYTE)0x20,(BYTE)0x4c,(BYTE)0x32,(BYTE)0x3c,(BYTE)0x00,(BYTE)0xff,(BYTE)0x70,(BYTE)0x00,(BYTE)0xd0,(BYTE)0x58,(BYTE)0x51,(BYTE)0xc9 },
    { 0x00a0,(BYTE)0xff,(BYTE)0xfc,(BYTE)0xb0,(BYTE)0x7c,(BYTE)0x12,(BYTE)0x34,(BYTE)0x66,(BYTE)0xc2,(BYTE)0x47,(BYTE)0xfa,(BYTE)0x00,(BYTE)0x0c,(BYTE)0x4b,(BYTE)0xfa,(BYTE)0xff,(BYTE)0xbe },
    { 0x00b0,(BYTE)0x4c,(BYTE)0xdf,(BYTE)0x00,(BYTE)0x28,(BYTE)0x4e,(BYTE)0xd4,(BYTE)0x4a,(BYTE)0x44,(BYTE)0x6b,(BYTE)0x16,(BYTE)0x3f,(BYTE)0x04,(BYTE)0x2f,(BYTE)0x0c,(BYTE)0x3f,(BYTE)0x05 },
    { 0x00c0,(BYTE)0x2f,(BYTE)0x06,(BYTE)0x3f,(BYTE)0x3c,(BYTE)0x00,(BYTE)0x2a,(BYTE)0x4e,(BYTE)0x4e,(BYTE)0xde,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x0e,(BYTE)0x60,(BYTE)0x00,(BYTE)0x00,(BYTE)0x9e },
    { 0x00d0,(BYTE)0x50,(BYTE)0xf8,(BYTE)0x04,(BYTE)0x3e,(BYTE)0x20,(BYTE)0x38,(BYTE)0x04,(BYTE)0xba,(BYTE)0x54,(BYTE)0x80,(BYTE)0xb0,(BYTE)0xb8,(BYTE)0x04,(BYTE)0xba,(BYTE)0x64,(BYTE)0xfa },
    { 0x00e0,(BYTE)0x3c,(BYTE)0x7c,(BYTE)0x86,(BYTE)0x09,(BYTE)0x48,(BYTE)0xe7,(BYTE)0x02,(BYTE)0x08,(BYTE)0x1d,(BYTE)0x6f,(BYTE)0x00,(BYTE)0x07,(BYTE)0x00,(BYTE)0x04,(BYTE)0x1d,(BYTE)0x6f },
    { 0x00f0,(BYTE)0x00,(BYTE)0x06,(BYTE)0x00,(BYTE)0x02,(BYTE)0x1c,(BYTE)0xaf,(BYTE)0x00,(BYTE)0x05,(BYTE)0x4d,(BYTE)0xfa,(BYTE)0x00,(BYTE)0x74,(BYTE)0x1d,(BYTE)0x6f,(BYTE)0x00,(BYTE)0x01 },
    { 0x0100,(BYTE)0x00,(BYTE)0x01,(BYTE)0x1d,(BYTE)0x6f,(BYTE)0x00,(BYTE)0x02,(BYTE)0x00,(BYTE)0x05,(BYTE)0x1d,(BYTE)0x6f,(BYTE)0x00,(BYTE)0x03,(BYTE)0x00,(BYTE)0x09,(BYTE)0x1d,(BYTE)0x45 },
    { 0x0110,(BYTE)0x00,(BYTE)0x0d,(BYTE)0x50,(BYTE)0x4f,(BYTE)0x3c,(BYTE)0x7c,(BYTE)0x86,(BYTE)0x06,(BYTE)0x3a,(BYTE)0x7c,(BYTE)0x86,(BYTE)0x04,(BYTE)0x3c,(BYTE)0xbc,(BYTE)0x01,(BYTE)0x98 },
    { 0x0120,(BYTE)0x3c,(BYTE)0xbc,(BYTE)0x00,(BYTE)0x98,(BYTE)0x3a,(BYTE)0x85,(BYTE)0x3c,(BYTE)0xbc,(BYTE)0x00,(BYTE)0x88,(BYTE)0x70,(BYTE)0x00,(BYTE)0x10,(BYTE)0x07,(BYTE)0x80,(BYTE)0x3c },
    { 0x0130,(BYTE)0x00,(BYTE)0x08,(BYTE)0x48,(BYTE)0x40,(BYTE)0x30,(BYTE)0x3c,(BYTE)0x00,(BYTE)0x8a,(BYTE)0x61,(BYTE)0x44,(BYTE)0x41,(BYTE)0xfa,(BYTE)0x00,(BYTE)0x32,(BYTE)0x74,(BYTE)0x03 },
    { 0x0140,(BYTE)0x20,(BYTE)0x18,(BYTE)0x61,(BYTE)0x3a,(BYTE)0x51,(BYTE)0xca,(BYTE)0xff,(BYTE)0xfa,(BYTE)0x2a,(BYTE)0xbc,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x0a,(BYTE)0x22,(BYTE)0x3c },
    { 0x0150,(BYTE)0x00,(BYTE)0x00,(BYTE)0x01,(BYTE)0x90,(BYTE)0x61,(BYTE)0x2c,(BYTE)0x3c,(BYTE)0xbc,(BYTE)0x00,(BYTE)0x8a,(BYTE)0x30,(BYTE)0x15,(BYTE)0xc0,(BYTE)0x7c,(BYTE)0x00,(BYTE)0xff },
    { 0x0160,(BYTE)0x67,(BYTE)0x02,(BYTE)0x70,(BYTE)0xff,(BYTE)0x3c,(BYTE)0xbc,(BYTE)0x00,(BYTE)0x80,(BYTE)0x51,(BYTE)0xf8,(BYTE)0x04,(BYTE)0x3e,(BYTE)0x4e,(BYTE)0x75,(BYTE)0x00,(BYTE)0x00 },
    { 0x0170,(BYTE)0x00,(BYTE)0x8a,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x8a,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x8a,(BYTE)0x00,(BYTE)0x01,(BYTE)0x00,(BYTE)0x8a,(BYTE)0x2a,(BYTE)0x80 },
    { 0x0180,(BYTE)0x72,(BYTE)0x0a,(BYTE)0xd2,(BYTE)0xb8,(BYTE)0x04,(BYTE)0xba,(BYTE)0x08,(BYTE)0x38,(BYTE)0x00,(BYTE)0x05,(BYTE)0xfa,(BYTE)0x01,(BYTE)0x67,(BYTE)0xde,(BYTE)0xb2,(BYTE)0xb8 },
    { 0x0190,(BYTE)0x04,(BYTE)0xba,(BYTE)0x64,(BYTE)0xf2,(BYTE)0x58,(BYTE)0x4f,(BYTE)0x60,(BYTE)0xca,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00 },
    { 0x01c0,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x04,(BYTE)0x00,(BYTE)0x00,(BYTE)0x81,(BYTE)0x47,(BYTE)0x45,(BYTE)0x4d,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x02,(BYTE)0x00,(BYTE)0x00 },
    { 0x01d0,(BYTE)0x07,(BYTE)0xfe,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00 },
    { 0x01f0,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x01,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x01,(BYTE)0x10,(BYTE)0x75 },
    { 0x0200,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0xa5,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00 },
    { 0x0400,(BYTE)0x60,(BYTE)0x1c,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x75,(BYTE)0x43,(BYTE)0xd6,(BYTE)0x00,(BYTE)0x02,(BYTE)0x02,(BYTE)0x01,(BYTE)0x00 },
    { 0x0410,(BYTE)0x02,(BYTE)0x00,(BYTE)0x01,(BYTE)0xfe,(BYTE)0x07,(BYTE)0xf8,(BYTE)0x05,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x76,(BYTE)0x48 },
    { 0x0420,(BYTE)0x70,(BYTE)0xff,(BYTE)0x2f,(BYTE)0x00,(BYTE)0x3f,(BYTE)0x03,(BYTE)0x4e,(BYTE)0x41,(BYTE)0x2a,(BYTE)0x00,(BYTE)0x2f,(BYTE)0x40,(BYTE)0x00,(BYTE)0x02,(BYTE)0x3e,(BYTE)0x83 },
    { 0x0430,(BYTE)0x4e,(BYTE)0x41,(BYTE)0x5c,(BYTE)0x4f,(BYTE)0x4a,(BYTE)0x80,(BYTE)0x67,(BYTE)0x5c,(BYTE)0x28,(BYTE)0x40,(BYTE)0x12,(BYTE)0x3a,(BYTE)0xff,(BYTE)0xd0,(BYTE)0xe0,(BYTE)0x59 },
    { 0x0440,(BYTE)0x12,(BYTE)0x3a,(BYTE)0xff,(BYTE)0xc9,(BYTE)0xe0,(BYTE)0x49,(BYTE)0xe2,(BYTE)0x49,(BYTE)0x74,(BYTE)0x00,(BYTE)0x14,(BYTE)0x3a,(BYTE)0xff,(BYTE)0xc6,(BYTE)0xe0,(BYTE)0x5a },
    { 0x0450,(BYTE)0x14,(BYTE)0x3a,(BYTE)0xff,(BYTE)0xbf,(BYTE)0x36,(BYTE)0x3a,(BYTE)0xff,(BYTE)0xb8,(BYTE)0xe0,(BYTE)0x5b,(BYTE)0xc6,(BYTE)0xc1,(BYTE)0xd6,(BYTE)0x86,(BYTE)0x3c,(BYTE)0x3a },
    { 0x0460,(BYTE)0xff,(BYTE)0xb6,(BYTE)0xe0,(BYTE)0x5e,(BYTE)0xcc,(BYTE)0xc1,(BYTE)0xdc,(BYTE)0x86,(BYTE)0xdc,(BYTE)0x83,(BYTE)0xe8,(BYTE)0x8a,(BYTE)0xd4,(BYTE)0x86,(BYTE)0x22,(BYTE)0x42 },
    { 0x0470,(BYTE)0x48,(BYTE)0xe7,(BYTE)0xd4,(BYTE)0x00,(BYTE)0x7a,(BYTE)0x01,(BYTE)0x48,(BYTE)0xe7,(BYTE)0x00,(BYTE)0x44,(BYTE)0x4e,(BYTE)0x93,(BYTE)0x4c,(BYTE)0xdf,(BYTE)0x22,(BYTE)0x00 },
    { 0x0480,(BYTE)0x4a,(BYTE)0x40,(BYTE)0x67,(BYTE)0x1e,(BYTE)0xdc,(BYTE)0x85,(BYTE)0xbc,(BYTE)0x89,(BYTE)0x65,(BYTE)0xea,(BYTE)0x3f,(BYTE)0x3c,(BYTE)0x00,(BYTE)0x49,(BYTE)0x4e,(BYTE)0x41 },
    { 0x0490,(BYTE)0xde,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x12,(BYTE)0x4e,(BYTE)0xd5,(BYTE)0x53,(BYTE)0x48,(BYTE)0x44,(BYTE)0x52,(BYTE)0x49,(BYTE)0x56,(BYTE)0x45,(BYTE)0x52,(BYTE)0x53,(BYTE)0x59 },
    { 0x04a0,(BYTE)0x53,(BYTE)0x00,(BYTE)0x20,(BYTE)0x4c,(BYTE)0x72,(BYTE)0x0f,(BYTE)0x70,(BYTE)0x0a,(BYTE)0x16,(BYTE)0x30,(BYTE)0x00,(BYTE)0x00,(BYTE)0xb6,(BYTE)0x3b,(BYTE)0x00,(BYTE)0xe8 },
    { 0x04b0,(BYTE)0x67,(BYTE)0x0a,(BYTE)0xd0,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x20,(BYTE)0x51,(BYTE)0xc9,(BYTE)0xff,(BYTE)0xee,(BYTE)0x60,(BYTE)0xc8,(BYTE)0x51,(BYTE)0xc8,(BYTE)0xff,(BYTE)0xea },
    { 0x04c0,(BYTE)0x2a,(BYTE)0x28,(BYTE)0x00,(BYTE)0x1c,(BYTE)0xe0,(BYTE)0x5d,(BYTE)0x48,(BYTE)0x45,(BYTE)0xe0,(BYTE)0x5d,(BYTE)0x06,(BYTE)0x85,(BYTE)0x00,(BYTE)0x00,(BYTE)0x08,(BYTE)0x00 },
    { 0x04d0,(BYTE)0xba,(BYTE)0xaf,(BYTE)0x00,(BYTE)0x0c,(BYTE)0x62,(BYTE)0xb4,(BYTE)0x24,(BYTE)0x4c,(BYTE)0xd9,(BYTE)0xef,(BYTE)0x00,(BYTE)0x0c,(BYTE)0x98,(BYTE)0xfc,(BYTE)0x06,(BYTE)0x00 },
    { 0x04e0,(BYTE)0x34,(BYTE)0x28,(BYTE)0x00,(BYTE)0x1a,(BYTE)0xe0,(BYTE)0x5a,(BYTE)0x76,(BYTE)0xff,(BYTE)0x7a,(BYTE)0x00,(BYTE)0x1a,(BYTE)0x3a,(BYTE)0xff,(BYTE)0x21,(BYTE)0xca,(BYTE)0xef },
    { 0x04f0,(BYTE)0x00,(BYTE)0x06,(BYTE)0x3c,(BYTE)0x02,(BYTE)0x55,(BYTE)0x46,(BYTE)0xcc,(BYTE)0xc5,(BYTE)0xdc,(BYTE)0x89,(BYTE)0xc5,(BYTE)0x4c,(BYTE)0x48,(BYTE)0xe7,(BYTE)0x20,(BYTE)0x64 },
    { 0x0500,(BYTE)0x4e,(BYTE)0x93,(BYTE)0x4c,(BYTE)0xdf,(BYTE)0x26,(BYTE)0x04,(BYTE)0x4a,(BYTE)0x40,(BYTE)0x66,(BYTE)0x80,(BYTE)0xc5,(BYTE)0x4c,(BYTE)0xe1,(BYTE)0x8d,(BYTE)0xda,(BYTE)0x85 },
    { 0x0510,(BYTE)0xd5,(BYTE)0xc5,(BYTE)0x3c,(BYTE)0x02,(BYTE)0x48,(BYTE)0xc6,(BYTE)0xe0,(BYTE)0x8e,(BYTE)0xdc,(BYTE)0xaf,(BYTE)0x00,(BYTE)0x08,(BYTE)0xbc,(BYTE)0x83,(BYTE)0x67,(BYTE)0x14 },
    { 0x0520,(BYTE)0x7a,(BYTE)0x01,(BYTE)0x26,(BYTE)0x06,(BYTE)0x48,(BYTE)0xe7,(BYTE)0x20,(BYTE)0x64,(BYTE)0x4e,(BYTE)0x93,(BYTE)0x4c,(BYTE)0xdf,(BYTE)0x26,(BYTE)0x04,(BYTE)0x4a,(BYTE)0x40 },
    { 0x0530,(BYTE)0x66,(BYTE)0x00,(BYTE)0xff,(BYTE)0x58,(BYTE)0x02,(BYTE)0x42,(BYTE)0x00,(BYTE)0xff,(BYTE)0xd4,(BYTE)0x42,(BYTE)0x34,(BYTE)0x34,(BYTE)0x20,(BYTE)0x00,(BYTE)0xe0,(BYTE)0x5a },
    { 0x0540,(BYTE)0x6a,(BYTE)0xa6,(BYTE)0x24,(BYTE)0x6f,(BYTE)0x00,(BYTE)0x00,(BYTE)0x20,(BYTE)0x4a,(BYTE)0x0c,(BYTE)0x58,(BYTE)0x60,(BYTE)0x1a,(BYTE)0x66,(BYTE)0x00,(BYTE)0xff,(BYTE)0x3c },
    { 0x0550,(BYTE)0x43,(BYTE)0xea,(BYTE)0x00,(BYTE)0x1c,(BYTE)0xd3,(BYTE)0xd8,(BYTE)0xd3,(BYTE)0xd8,(BYTE)0xd3,(BYTE)0xe8,(BYTE)0x00,(BYTE)0x04,(BYTE)0x4a,(BYTE)0x91,(BYTE)0x67,(BYTE)0x20 },
    { 0x0560,(BYTE)0x41,(BYTE)0xea,(BYTE)0x00,(BYTE)0x1c,(BYTE)0x22,(BYTE)0x08,(BYTE)0x70,(BYTE)0x00,(BYTE)0xd1,(BYTE)0xd9,(BYTE)0xd3,(BYTE)0x90,(BYTE)0x10,(BYTE)0x19,(BYTE)0x67,(BYTE)0x10 },
    { 0x0570,(BYTE)0xb0,(BYTE)0x3c,(BYTE)0x00,(BYTE)0x01,(BYTE)0x66,(BYTE)0x06,(BYTE)0xd0,(BYTE)0xfc,(BYTE)0x00,(BYTE)0xfe,(BYTE)0x60,(BYTE)0xf0,(BYTE)0xd0,(BYTE)0xc0,(BYTE)0x60,(BYTE)0xea },
    { 0x0580,(BYTE)0x20,(BYTE)0x2f,(BYTE)0x00,(BYTE)0x0c,(BYTE)0xde,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x10,(BYTE)0x4e,(BYTE)0xea,(BYTE)0x00,(BYTE)0x20,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00 },
    { 0x05f0,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x87,(BYTE)0x42 },
    { 0x0600,(BYTE)0xf8,(BYTE)0xff,(BYTE)0xff,(BYTE)0xff,(BYTE)0x03,(BYTE)0x00,(BYTE)0x04,(BYTE)0x00,(BYTE)0x05,(BYTE)0x00,(BYTE)0x06,(BYTE)0x00,(BYTE)0x07,(BYTE)0x00,(BYTE)0x08,(BYTE)0x00 },
    { 0x0610,(BYTE)0x09,(BYTE)0x00,(BYTE)0x0a,(BYTE)0x00,(BYTE)0x0b,(BYTE)0x00,(BYTE)0x0c,(BYTE)0x00,(BYTE)0xff,(BYTE)0xff,(BYTE)0xff,(BYTE)0xff,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00 },
    { 0x1000,(BYTE)0xf8,(BYTE)0xff,(BYTE)0xff,(BYTE)0xff,(BYTE)0x03,(BYTE)0x00,(BYTE)0x04,(BYTE)0x00,(BYTE)0x05,(BYTE)0x00,(BYTE)0x06,(BYTE)0x00,(BYTE)0x07,(BYTE)0x00,(BYTE)0x08,(BYTE)0x00 },
    { 0x1010,(BYTE)0x09,(BYTE)0x00,(BYTE)0x0a,(BYTE)0x00,(BYTE)0x0b,(BYTE)0x00,(BYTE)0x0c,(BYTE)0x00,(BYTE)0xff,(BYTE)0xff,(BYTE)0xff,(BYTE)0xff,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00 },
    { 0x1a00,(BYTE)0x53,(BYTE)0x48,(BYTE)0x44,(BYTE)0x52,(BYTE)0x49,(BYTE)0x56,(BYTE)0x45,(BYTE)0x52,(BYTE)0x53,(BYTE)0x59,(BYTE)0x53,(BYTE)0x20,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00 },
    { 0x1a10,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x23,(BYTE)0x6d,(BYTE)0x69,(BYTE)0x25,(BYTE)0x02,(BYTE)0x00,(BYTE)0xfb,(BYTE)0x2b,(BYTE)0x00,(BYTE)0x00 },
    { 0x1a20,(BYTE)0x4e,(BYTE)0x45,(BYTE)0x57,(BYTE)0x44,(BYTE)0x45,(BYTE)0x53,(BYTE)0x4b,(BYTE)0x20,(BYTE)0x49,(BYTE)0x4e,(BYTE)0x46,(BYTE)0x20,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00 },
    { 0x1a30,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0xdb,(BYTE)0x76,(BYTE)0x6c,(BYTE)0x25,(BYTE)0x0d,(BYTE)0x00,(BYTE)0x18,(BYTE)0x03,(BYTE)0x00,(BYTE)0x00 },
    { 0x3a00,(BYTE)0x60,(BYTE)0x1a,(BYTE)0x00,(BYTE)0x00,(BYTE)0x2a,(BYTE)0x62,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x02,(BYTE)0x0a,(BYTE)0x00,(BYTE)0x00 },
    { 0x3a10,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x60,(BYTE)0x00,(BYTE)0x00,(BYTE)0x48 },
    { 0x3a20,(BYTE)0x60,(BYTE)0x00,(BYTE)0x00,(BYTE)0x60,(BYTE)0x60,(BYTE)0x00,(BYTE)0x00,(BYTE)0x5c,(BYTE)0xf0,(BYTE)0xad,(BYTE)0x05,(BYTE)0x00,(BYTE)0x00,(BYTE)0x80,(BYTE)0x02,(BYTE)0x00 },
    { 0x3a30,(BYTE)0x00,(BYTE)0x08,(BYTE)0x01,(BYTE)0x01,(BYTE)0x01,(BYTE)0x01,(BYTE)0x01,(BYTE)0x01,(BYTE)0x01,(BYTE)0x01,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x08 },
    { 0x3a40,(BYTE)0x01,(BYTE)0x01,(BYTE)0x01,(BYTE)0x01,(BYTE)0x01,(BYTE)0x01,(BYTE)0x01,(BYTE)0x01,(BYTE)0x00,(BYTE)0x00,(BYTE)0x2e,(BYTE)0xe1,(BYTE)0x00,(BYTE)0x00,(BYTE)0x13,(BYTE)0x88 },
    { 0x3a50,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x2a,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0xc9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x65,(BYTE)0x00,(BYTE)0x00,(BYTE)0x03,(BYTE)0x21 },
    { 0x3a60,(BYTE)0x00,(BYTE)0x0c,(BYTE)0x00,(BYTE)0xff,(BYTE)0xff,(BYTE)0xff,(BYTE)0x24,(BYTE)0x6f,(BYTE)0x00,(BYTE)0x04,(BYTE)0x13,(BYTE)0xea,(BYTE)0x00,(BYTE)0x80,(BYTE)0x00,(BYTE)0x00 },
    { 0x3a70,(BYTE)0x06,(BYTE)0x18,(BYTE)0x20,(BYTE)0x2a,(BYTE)0x00,(BYTE)0x04,(BYTE)0x90,(BYTE)0x92,(BYTE)0x90,(BYTE)0xbc,(BYTE)0x00,(BYTE)0x00,(BYTE)0x01,(BYTE)0x00,(BYTE)0x60,(BYTE)0x00 },
    { 0x3a80,(BYTE)0x21,(BYTE)0xb2,(BYTE)0x50,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0x96,(BYTE)0x90,(BYTE)0xbc,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x1c,(BYTE)0x23,(BYTE)0xca },
    { 0x3a90,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0x9c,(BYTE)0x60,(BYTE)0x00,(BYTE)0x21,(BYTE)0x9c,(BYTE)0x0d,(BYTE)0x41,(BYTE)0x48,(BYTE)0x44,(BYTE)0x49,(BYTE)0x20,(BYTE)0x3a,(BYTE)0x20 },
    { 0x3aa0,(BYTE)0x44,(BYTE)0x65,(BYTE)0x63,(BYTE)0x20,(BYTE)0x30,(BYTE)0x33,(BYTE)0x20,(BYTE)0x20,(BYTE)0x31,(BYTE)0x39,(BYTE)0x39,(BYTE)0x31,(BYTE)0x20,(BYTE)0x76,(BYTE)0x35,(BYTE)0x2e },
    { 0x3ab0,(BYTE)0x30,(BYTE)0x30,(BYTE)0x0d,(BYTE)0x0a,(BYTE)0xbd,(BYTE)0x41,(BYTE)0x74,(BYTE)0x61,(BYTE)0x72,(BYTE)0x69,(BYTE)0x20,(BYTE)0x43,(BYTE)0x6f,(BYTE)0x72,(BYTE)0x70,(BYTE)0x2e },
    { 0x3ac0,(BYTE)0x20,(BYTE)0x31,(BYTE)0x39,(BYTE)0x38,(BYTE)0x35,(BYTE)0x2c,(BYTE)0x20,(BYTE)0x31,(BYTE)0x39,(BYTE)0x38,(BYTE)0x36,(BYTE)0x2c,(BYTE)0x20,(BYTE)0x31,(BYTE)0x39,(BYTE)0x38 },
    { 0x3ad0,(BYTE)0x37,(BYTE)0x2c,(BYTE)0x20,(BYTE)0x31,(BYTE)0x39,(BYTE)0x38,(BYTE)0x38,(BYTE)0x2c,(BYTE)0x20,(BYTE)0x31,(BYTE)0x39,(BYTE)0x38,(BYTE)0x39,(BYTE)0x2c,(BYTE)0x20,(BYTE)0x31 },
    { 0x3ae0,(BYTE)0x39,(BYTE)0x39,(BYTE)0x30,(BYTE)0x2c,(BYTE)0x20,(BYTE)0x31,(BYTE)0x39,(BYTE)0x39,(BYTE)0x31,(BYTE)0x0d,(BYTE)0x0a,(BYTE)0x00,(BYTE)0x1a,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00 },
    { 0x3af0,(BYTE)0xff,(BYTE)0xff,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00 },
    { 0x3b40,(BYTE)0x41,(BYTE)0x48,(BYTE)0x44,(BYTE)0x49,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0x00,(BYTE)0x02,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00 },
    { 0x3b90,(BYTE)0x02,(BYTE)0x02,(BYTE)0x02,(BYTE)0x02,(BYTE)0x02,(BYTE)0x02,(BYTE)0x02,(BYTE)0x02,(BYTE)0x02,(BYTE)0x02,(BYTE)0x02,(BYTE)0x02,(BYTE)0x02,(BYTE)0x02,(BYTE)0x01,(BYTE)0x01 },
    { 0x3ba0,(BYTE)0x01,(BYTE)0x01,(BYTE)0x01,(BYTE)0x01,(BYTE)0x01,(BYTE)0x01,(BYTE)0x01,(BYTE)0x01,(BYTE)0x01,(BYTE)0x01,(BYTE)0x01,(BYTE)0x01,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00 },
    { 0x3be0,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x01,(BYTE)0x01,(BYTE)0x01,(BYTE)0x01 },
    { 0x3bf0,(BYTE)0x01,(BYTE)0x01,(BYTE)0x01,(BYTE)0x01,(BYTE)0x01,(BYTE)0x01,(BYTE)0x01,(BYTE)0x01,(BYTE)0x01,(BYTE)0x01,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00 },
    { 0x3fd0,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x03,(BYTE)0x00,(BYTE)0x01 },
    { 0x3fe0,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x01,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x01,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x01,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00 },
    { 0x4010,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x01,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00 },
    { 0x4020,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x01,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00 },
    { 0x4030,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x30,(BYTE)0x2f,(BYTE)0x00,(BYTE)0x04,(BYTE)0x42,(BYTE)0x41,(BYTE)0x20,(BYTE)0x79,(BYTE)0x00,(BYTE)0x00 },
    { 0x4040,(BYTE)0x05,(BYTE)0xc4,(BYTE)0x43,(BYTE)0xfa,(BYTE)0x00,(BYTE)0x7c,(BYTE)0x60,(BYTE)0x24,(BYTE)0x30,(BYTE)0x2f,(BYTE)0x00,(BYTE)0x0e,(BYTE)0x32,(BYTE)0x2f,(BYTE)0x00,(BYTE)0x04 },
    { 0x4050,(BYTE)0x20,(BYTE)0x79,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xc8,(BYTE)0x43,(BYTE)0xfa,(BYTE)0x06,(BYTE)0x8e,(BYTE)0x60,(BYTE)0x10,(BYTE)0x30,(BYTE)0x2f,(BYTE)0x00,(BYTE)0x04 },
    { 0x4060,(BYTE)0x42,(BYTE)0x41,(BYTE)0x20,(BYTE)0x79,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xcc,(BYTE)0x43,(BYTE)0xfa,(BYTE)0x0c,(BYTE)0x68,(BYTE)0x55,(BYTE)0x40,(BYTE)0x6b,(BYTE)0x4e },
    { 0x4070,(BYTE)0x08,(BYTE)0x01,(BYTE)0x00,(BYTE)0x03,(BYTE)0x67,(BYTE)0x34,(BYTE)0xb0,(BYTE)0x79,(BYTE)0x00,(BYTE)0x00,(BYTE)0x01,(BYTE)0x70,(BYTE)0x6d,(BYTE)0x3e,(BYTE)0x90,(BYTE)0x7c },
    { 0x4080,(BYTE)0x00,(BYTE)0x08,(BYTE)0x6b,(BYTE)0x3a,(BYTE)0xb0,(BYTE)0x79,(BYTE)0x00,(BYTE)0x00,(BYTE)0x01,(BYTE)0x72,(BYTE)0x6d,(BYTE)0x30,(BYTE)0x90,(BYTE)0x7c,(BYTE)0x00,(BYTE)0x08 },
    { 0x4090,(BYTE)0x6b,(BYTE)0x2c,(BYTE)0x32,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0xd2,(BYTE)0x92,(BYTE)0x79,(BYTE)0x00,(BYTE)0x00,(BYTE)0x01,(BYTE)0x70,(BYTE)0x92,(BYTE)0x79 },
    { 0x40a0,(BYTE)0x00,(BYTE)0x00,(BYTE)0x01,(BYTE)0x72,(BYTE)0xb0,(BYTE)0x41,(BYTE)0x6c,(BYTE)0x16,(BYTE)0x60,(BYTE)0x12,(BYTE)0xb0,(BYTE)0x7c,(BYTE)0x00,(BYTE)0x0e,(BYTE)0x6c,(BYTE)0x0e },
    { 0x40b0,(BYTE)0x45,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0xd6,(BYTE)0x4a,(BYTE)0x32,(BYTE)0x00,(BYTE)0x00,(BYTE)0x6b,(BYTE)0x02,(BYTE)0x20,(BYTE)0x49,(BYTE)0x4e,(BYTE)0xd0 },
    { 0x40c0,(BYTE)0x55,(BYTE)0x6f,(BYTE)0x00,(BYTE)0x04,(BYTE)0x32,(BYTE)0x2f,(BYTE)0x00,(BYTE)0x04,(BYTE)0x41,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0xd6,(BYTE)0xd0,(BYTE)0xc1 },
    { 0x40d0,(BYTE)0x74,(BYTE)0x00,(BYTE)0x14,(BYTE)0x10,(BYTE)0x33,(BYTE)0xc2,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xa8,(BYTE)0x43,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x01,(BYTE)0x82 },
    { 0x40e0,(BYTE)0x4a,(BYTE)0x31,(BYTE)0x10,(BYTE)0x00,(BYTE)0x66,(BYTE)0x1c,(BYTE)0x48,(BYTE)0xe7,(BYTE)0x40,(BYTE)0x80,(BYTE)0x3f,(BYTE)0x02,(BYTE)0x61,(BYTE)0x00,(BYTE)0x16,(BYTE)0x50 },
    { 0x40f0,(BYTE)0x54,(BYTE)0x8f,(BYTE)0x4c,(BYTE)0xdf,(BYTE)0x01,(BYTE)0x02,(BYTE)0x4a,(BYTE)0x40,(BYTE)0x67,(BYTE)0x00,(BYTE)0x00,(BYTE)0xcc,(BYTE)0x70,(BYTE)0x01,(BYTE)0x61,(BYTE)0x00 },
    { 0x4100,(BYTE)0x0c,(BYTE)0x3c,(BYTE)0x23,(BYTE)0xf8,(BYTE)0x04,(BYTE)0xc6,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xf4,(BYTE)0x06,(BYTE)0xb9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x02,(BYTE)0x00 },
    { 0x4110,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xf4,(BYTE)0x2f,(BYTE)0x08,(BYTE)0x3f,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xa8,(BYTE)0x54,(BYTE)0x57,(BYTE)0x42,(BYTE)0x67 },
    { 0x4120,(BYTE)0x3f,(BYTE)0x3c,(BYTE)0x00,(BYTE)0x01,(BYTE)0x2f,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xf4,(BYTE)0x3f,(BYTE)0x3c,(BYTE)0x00,(BYTE)0x0a,(BYTE)0x61,(BYTE)0x00 },
    { 0x4130,(BYTE)0x05,(BYTE)0xb6,(BYTE)0xde,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x0c,(BYTE)0x20,(BYTE)0x5f,(BYTE)0x4a,(BYTE)0x40,(BYTE)0x67,(BYTE)0x18,(BYTE)0x2f,(BYTE)0x08,(BYTE)0x32,(BYTE)0x2f },
    { 0x4140,(BYTE)0x00,(BYTE)0x08,(BYTE)0x61,(BYTE)0x00,(BYTE)0x0e,(BYTE)0x18,(BYTE)0x20,(BYTE)0x5f,(BYTE)0x0c,(BYTE)0x80,(BYTE)0x00,(BYTE)0x01,(BYTE)0x00,(BYTE)0x00,(BYTE)0x67,(BYTE)0xc4 },
    { 0x4150,(BYTE)0x60,(BYTE)0x00,(BYTE)0x00,(BYTE)0x74,(BYTE)0x34,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xa8,(BYTE)0x33,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00 },
    { 0x4160,(BYTE)0x06,(BYTE)0x16,(BYTE)0xb4,(BYTE)0x20,(BYTE)0x66,(BYTE)0x08,(BYTE)0x52,(BYTE)0x79,(BYTE)0x00,(BYTE)0x00,(BYTE)0x06,(BYTE)0x16,(BYTE)0x60,(BYTE)0xf4,(BYTE)0x32,(BYTE)0x3c },
    { 0x4170,(BYTE)0x00,(BYTE)0x03,(BYTE)0x20,(BYTE)0x79,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xf4,(BYTE)0x0c,(BYTE)0x68,(BYTE)0x55,(BYTE)0xaa,(BYTE)0x01,(BYTE)0xfe,(BYTE)0x66,(BYTE)0x06 },
    { 0x4180,(BYTE)0x61,(BYTE)0x00,(BYTE)0x00,(BYTE)0x48,(BYTE)0x60,(BYTE)0x0c,(BYTE)0x33,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x01,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xac,(BYTE)0x61,(BYTE)0x00 },
    { 0x4190,(BYTE)0x01,(BYTE)0x72,(BYTE)0x4a,(BYTE)0x40,(BYTE)0x67,(BYTE)0x04,(BYTE)0x6a,(BYTE)0x1e,(BYTE)0x60,(BYTE)0x2c,(BYTE)0x41,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x01,(BYTE)0x82 },
    { 0x41a0,(BYTE)0x30,(BYTE)0x2f,(BYTE)0x00,(BYTE)0x04,(BYTE)0x42,(BYTE)0x30,(BYTE)0x00,(BYTE)0x00,(BYTE)0x41,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x01,(BYTE)0x74,(BYTE)0x11,(BYTE)0xbc },
    { 0x41b0,(BYTE)0x00,(BYTE)0x02,(BYTE)0x00,(BYTE)0x00,(BYTE)0x60,(BYTE)0x10,(BYTE)0x2f,(BYTE)0x01,(BYTE)0x3f,(BYTE)0x2f,(BYTE)0x00,(BYTE)0x08,(BYTE)0x61,(BYTE)0x00,(BYTE)0x02,(BYTE)0xc4 },
    { 0x41c0,(BYTE)0x5c,(BYTE)0x8f,(BYTE)0x4a,(BYTE)0x80,(BYTE)0x6a,(BYTE)0x02,(BYTE)0x70,(BYTE)0x00,(BYTE)0x4e,(BYTE)0x75,(BYTE)0xd0,(BYTE)0xfc,(BYTE)0x01,(BYTE)0xbe,(BYTE)0x48,(BYTE)0xe7 },
    { 0x41d0,(BYTE)0x40,(BYTE)0x80,(BYTE)0x51,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x06,(BYTE)0x0c,(BYTE)0x61,(BYTE)0x00,(BYTE)0x00,(BYTE)0x6c,(BYTE)0x4a,(BYTE)0x00,(BYTE)0x67,(BYTE)0x56 },
    { 0x41e0,(BYTE)0x0c,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0x66,(BYTE)0x26,(BYTE)0x50,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x06,(BYTE)0x0c,(BYTE)0x23,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x00 },
    { 0x41f0,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x06,(BYTE)0x12,(BYTE)0x23,(BYTE)0xc1,(BYTE)0x00,(BYTE)0x00,(BYTE)0x06,(BYTE)0x0e,(BYTE)0x61,(BYTE)0x00,(BYTE)0x00,(BYTE)0x76 },
    { 0x4200,(BYTE)0x4a,(BYTE)0x00,(BYTE)0x67,(BYTE)0x32,(BYTE)0x6b,(BYTE)0x0e,(BYTE)0x0c,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0x67,(BYTE)0xf0,(BYTE)0x53,(BYTE)0x79,(BYTE)0x00,(BYTE)0x00 },
    { 0x4210,(BYTE)0x06,(BYTE)0x16,(BYTE)0x6a,(BYTE)0x1a,(BYTE)0x50,(BYTE)0x8f,(BYTE)0x33,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xac,(BYTE)0x0c,(BYTE)0x00 },
    { 0x4220,(BYTE)0x00,(BYTE)0x01,(BYTE)0x67,(BYTE)0x08,(BYTE)0x33,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x01,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xac,(BYTE)0x60,(BYTE)0x16,(BYTE)0x4a,(BYTE)0x39 },
    { 0x4230,(BYTE)0x00,(BYTE)0x00,(BYTE)0x06,(BYTE)0x0c,(BYTE)0x66,(BYTE)0xc6,(BYTE)0x4c,(BYTE)0xdf,(BYTE)0x01,(BYTE)0x02,(BYTE)0xd0,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x10,(BYTE)0x51,(BYTE)0xc9 },
    { 0x4240,(BYTE)0xff,(BYTE)0x8e,(BYTE)0x70,(BYTE)0x00,(BYTE)0x4e,(BYTE)0x75,(BYTE)0x4a,(BYTE)0xa8,(BYTE)0x00,(BYTE)0x0c,(BYTE)0x67,(BYTE)0x18,(BYTE)0x10,(BYTE)0x28,(BYTE)0x00,(BYTE)0x04 },
    { 0x4250,(BYTE)0x67,(BYTE)0x20,(BYTE)0x0c,(BYTE)0x00,(BYTE)0x00,(BYTE)0x04,(BYTE)0x67,(BYTE)0x10,(BYTE)0x0c,(BYTE)0x00,(BYTE)0x00,(BYTE)0x01,(BYTE)0x67,(BYTE)0x0a,(BYTE)0x0c,(BYTE)0x00 },
    { 0x4260,(BYTE)0x00,(BYTE)0x05,(BYTE)0x67,(BYTE)0x04,(BYTE)0x70,(BYTE)0x00,(BYTE)0x60,(BYTE)0x0a,(BYTE)0x22,(BYTE)0x28,(BYTE)0x00,(BYTE)0x08,(BYTE)0xe0,(BYTE)0x59,(BYTE)0x48,(BYTE)0x41 },
    { 0x4270,(BYTE)0xe0,(BYTE)0x59,(BYTE)0x4e,(BYTE)0x75,(BYTE)0x0c,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0x66,(BYTE)0x3c,(BYTE)0x2f,(BYTE)0x01,(BYTE)0x3f,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00 },
    { 0x4280,(BYTE)0x05,(BYTE)0xa8,(BYTE)0x54,(BYTE)0x57,(BYTE)0x3f,(BYTE)0x3c,(BYTE)0xff,(BYTE)0xff,(BYTE)0x3f,(BYTE)0x3c,(BYTE)0x00,(BYTE)0x01,(BYTE)0x2f,(BYTE)0x38,(BYTE)0x04,(BYTE)0xc6 },
    { 0x4290,(BYTE)0x3f,(BYTE)0x3c,(BYTE)0x00,(BYTE)0x0a,(BYTE)0x61,(BYTE)0x00,(BYTE)0x04,(BYTE)0x50,(BYTE)0xde,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x10,(BYTE)0x4a,(BYTE)0x40,(BYTE)0x66,(BYTE)0x00 },
    { 0x42a0,(BYTE)0x00,(BYTE)0x60,(BYTE)0x20,(BYTE)0x78,(BYTE)0x04,(BYTE)0xc6,(BYTE)0x0c,(BYTE)0x68,(BYTE)0x55,(BYTE)0xaa,(BYTE)0x01,(BYTE)0xfe,(BYTE)0x66,(BYTE)0x52,(BYTE)0xd0,(BYTE)0xfc },
    { 0x42b0,(BYTE)0x01,(BYTE)0xae,(BYTE)0x34,(BYTE)0x3c,(BYTE)0x00,(BYTE)0x04,(BYTE)0x53,(BYTE)0x42,(BYTE)0x6b,(BYTE)0x36,(BYTE)0xd0,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x10,(BYTE)0x4a,(BYTE)0xa8 },
    { 0x42c0,(BYTE)0x00,(BYTE)0x0c,(BYTE)0x67,(BYTE)0xf2,(BYTE)0x10,(BYTE)0x28,(BYTE)0x00,(BYTE)0x04,(BYTE)0x67,(BYTE)0xec,(BYTE)0x22,(BYTE)0x28,(BYTE)0x00,(BYTE)0x08,(BYTE)0x67,(BYTE)0xe6 },
    { 0x42d0,(BYTE)0xe0,(BYTE)0x59,(BYTE)0x48,(BYTE)0x41,(BYTE)0xe0,(BYTE)0x59,(BYTE)0x0c,(BYTE)0x00,(BYTE)0x00,(BYTE)0x04,(BYTE)0x67,(BYTE)0x18,(BYTE)0x0c,(BYTE)0x00,(BYTE)0x00,(BYTE)0x01 },
    { 0x42e0,(BYTE)0x67,(BYTE)0x12,(BYTE)0x0c,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0x66,(BYTE)0xce,(BYTE)0x23,(BYTE)0xc1,(BYTE)0x00,(BYTE)0x00,(BYTE)0x06,(BYTE)0x12,(BYTE)0x60,(BYTE)0x0a },
    { 0x42f0,(BYTE)0x70,(BYTE)0x00,(BYTE)0x60,(BYTE)0x0c,(BYTE)0xd2,(BYTE)0xb9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x06,(BYTE)0x12,(BYTE)0xd2,(BYTE)0xb9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x06,(BYTE)0x0e },
    { 0x4300,(BYTE)0x4e,(BYTE)0x75,(BYTE)0xd0,(BYTE)0xfc,(BYTE)0x01,(BYTE)0xc2,(BYTE)0x4a,(BYTE)0x98,(BYTE)0x67,(BYTE)0x5e,(BYTE)0x48,(BYTE)0xe7,(BYTE)0x40,(BYTE)0x80,(BYTE)0x51,(BYTE)0xf9 },
    { 0x4310,(BYTE)0x00,(BYTE)0x00,(BYTE)0x06,(BYTE)0x0c,(BYTE)0x61,(BYTE)0x00,(BYTE)0x00,(BYTE)0x56,(BYTE)0x4a,(BYTE)0x00,(BYTE)0x67,(BYTE)0x40,(BYTE)0x0c,(BYTE)0x00,(BYTE)0x00,(BYTE)0x58 },
    { 0x4320,(BYTE)0x66,(BYTE)0x26,(BYTE)0x50,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x06,(BYTE)0x0c,(BYTE)0x23,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00 },
    { 0x4330,(BYTE)0x06,(BYTE)0x12,(BYTE)0x23,(BYTE)0xc1,(BYTE)0x00,(BYTE)0x00,(BYTE)0x06,(BYTE)0x0e,(BYTE)0x61,(BYTE)0x00,(BYTE)0x00,(BYTE)0x92,(BYTE)0x4a,(BYTE)0x00,(BYTE)0x67,(BYTE)0x1c },
    { 0x4340,(BYTE)0x6b,(BYTE)0x0e,(BYTE)0x0c,(BYTE)0x00,(BYTE)0x00,(BYTE)0x58,(BYTE)0x67,(BYTE)0xf0,(BYTE)0x53,(BYTE)0x79,(BYTE)0x00,(BYTE)0x00,(BYTE)0x06,(BYTE)0x16,(BYTE)0x6a,(BYTE)0x04 },
    { 0x4350,(BYTE)0x50,(BYTE)0x8f,(BYTE)0x60,(BYTE)0x16,(BYTE)0x4a,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00,(BYTE)0x06,(BYTE)0x0c,(BYTE)0x66,(BYTE)0xdc,(BYTE)0x4c,(BYTE)0xdf,(BYTE)0x01,(BYTE)0x02 },
    { 0x4360,(BYTE)0xd0,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x0c,(BYTE)0x51,(BYTE)0xc9,(BYTE)0xff,(BYTE)0xa4,(BYTE)0x70,(BYTE)0x00,(BYTE)0x4e,(BYTE)0x75,(BYTE)0x4a,(BYTE)0x10,(BYTE)0x67,(BYTE)0x4e },
    { 0x4370,(BYTE)0x4a,(BYTE)0xa8,(BYTE)0x00,(BYTE)0x08,(BYTE)0x67,(BYTE)0x48,(BYTE)0x0c,(BYTE)0x28,(BYTE)0x00,(BYTE)0x47,(BYTE)0x00,(BYTE)0x01,(BYTE)0x66,(BYTE)0x10,(BYTE)0x0c,(BYTE)0x28 },
    { 0x4380,(BYTE)0x00,(BYTE)0x45,(BYTE)0x00,(BYTE)0x02,(BYTE)0x66,(BYTE)0x08,(BYTE)0x0c,(BYTE)0x28,(BYTE)0x00,(BYTE)0x4d,(BYTE)0x00,(BYTE)0x03,(BYTE)0x67,(BYTE)0x34,(BYTE)0x0c,(BYTE)0x28 },
    { 0x4390,(BYTE)0x00,(BYTE)0x42,(BYTE)0x00,(BYTE)0x01,(BYTE)0x66,(BYTE)0x10,(BYTE)0x0c,(BYTE)0x28,(BYTE)0x00,(BYTE)0x47,(BYTE)0x00,(BYTE)0x02,(BYTE)0x66,(BYTE)0x08,(BYTE)0x0c,(BYTE)0x28 },
    { 0x43a0,(BYTE)0x00,(BYTE)0x4d,(BYTE)0x00,(BYTE)0x03,(BYTE)0x67,(BYTE)0x1c,(BYTE)0x0c,(BYTE)0x28,(BYTE)0x00,(BYTE)0x58,(BYTE)0x00,(BYTE)0x01,(BYTE)0x66,(BYTE)0x10,(BYTE)0x0c,(BYTE)0x28 },
    { 0x43b0,(BYTE)0x00,(BYTE)0x47,(BYTE)0x00,(BYTE)0x02,(BYTE)0x66,(BYTE)0x08,(BYTE)0x0c,(BYTE)0x28,(BYTE)0x00,(BYTE)0x4d,(BYTE)0x00,(BYTE)0x03,(BYTE)0x67,(BYTE)0x04,(BYTE)0x70,(BYTE)0x00 },
    { 0x43c0,(BYTE)0x60,(BYTE)0x08,(BYTE)0x22,(BYTE)0x28,(BYTE)0x00,(BYTE)0x04,(BYTE)0x10,(BYTE)0x28,(BYTE)0x00,(BYTE)0x01,(BYTE)0x4e,(BYTE)0x75,(BYTE)0x0c,(BYTE)0x00,(BYTE)0x00,(BYTE)0x58 },
    { 0x43d0,(BYTE)0x66,(BYTE)0x34,(BYTE)0x2f,(BYTE)0x01,(BYTE)0x3f,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xa8,(BYTE)0x54,(BYTE)0x57,(BYTE)0x3f,(BYTE)0x3c,(BYTE)0xff,(BYTE)0xff },
    { 0x43e0,(BYTE)0x3f,(BYTE)0x3c,(BYTE)0x00,(BYTE)0x01,(BYTE)0x2f,(BYTE)0x38,(BYTE)0x04,(BYTE)0xc6,(BYTE)0x3f,(BYTE)0x3c,(BYTE)0x00,(BYTE)0x0a,(BYTE)0x61,(BYTE)0x00,(BYTE)0x02,(BYTE)0xf8 },
    { 0x43f0,(BYTE)0xde,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x10,(BYTE)0x4a,(BYTE)0x40,(BYTE)0x66,(BYTE)0x00,(BYTE)0x00,(BYTE)0x88,(BYTE)0x20,(BYTE)0x78,(BYTE)0x04,(BYTE)0xc6,(BYTE)0xd0,(BYTE)0xfc },
    { 0x4400,(BYTE)0x01,(BYTE)0xba,(BYTE)0x34,(BYTE)0x3c,(BYTE)0x00,(BYTE)0x04,(BYTE)0x53,(BYTE)0x42,(BYTE)0x6b,(BYTE)0x62,(BYTE)0xd0,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x0c,(BYTE)0x4a,(BYTE)0xa8 },
    { 0x4410,(BYTE)0x00,(BYTE)0x08,(BYTE)0x67,(BYTE)0xf2,(BYTE)0x4a,(BYTE)0x10,(BYTE)0x67,(BYTE)0xee,(BYTE)0x22,(BYTE)0x28,(BYTE)0x00,(BYTE)0x04,(BYTE)0x0c,(BYTE)0x28,(BYTE)0x00,(BYTE)0x47 },
    { 0x4420,(BYTE)0x00,(BYTE)0x01,(BYTE)0x66,(BYTE)0x10,(BYTE)0x0c,(BYTE)0x28,(BYTE)0x00,(BYTE)0x45,(BYTE)0x00,(BYTE)0x02,(BYTE)0x66,(BYTE)0x08,(BYTE)0x0c,(BYTE)0x28,(BYTE)0x00,(BYTE)0x4d },
    { 0x4430,(BYTE)0x00,(BYTE)0x03,(BYTE)0x67,(BYTE)0x3c,(BYTE)0x0c,(BYTE)0x28,(BYTE)0x00,(BYTE)0x42,(BYTE)0x00,(BYTE)0x01,(BYTE)0x66,(BYTE)0x10,(BYTE)0x0c,(BYTE)0x28,(BYTE)0x00,(BYTE)0x47 },
    { 0x4440,(BYTE)0x00,(BYTE)0x02,(BYTE)0x66,(BYTE)0x08,(BYTE)0x0c,(BYTE)0x28,(BYTE)0x00,(BYTE)0x4d,(BYTE)0x00,(BYTE)0x03,(BYTE)0x67,(BYTE)0x24,(BYTE)0x0c,(BYTE)0x28,(BYTE)0x00,(BYTE)0x58 },
    { 0x4450,(BYTE)0x00,(BYTE)0x01,(BYTE)0x66,(BYTE)0x18,(BYTE)0x0c,(BYTE)0x28,(BYTE)0x00,(BYTE)0x47,(BYTE)0x00,(BYTE)0x02,(BYTE)0x66,(BYTE)0x10,(BYTE)0x0c,(BYTE)0x28,(BYTE)0x00,(BYTE)0x4d },
    { 0x4460,(BYTE)0x00,(BYTE)0x03,(BYTE)0x66,(BYTE)0xa2,(BYTE)0x23,(BYTE)0xc1,(BYTE)0x00,(BYTE)0x00,(BYTE)0x06,(BYTE)0x12,(BYTE)0x60,(BYTE)0x0a,(BYTE)0x70,(BYTE)0x00,(BYTE)0x60,(BYTE)0x10 },
    { 0x4470,(BYTE)0xd2,(BYTE)0xb9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x06,(BYTE)0x12,(BYTE)0xd2,(BYTE)0xb9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x06,(BYTE)0x0e,(BYTE)0x10,(BYTE)0x28,(BYTE)0x00,(BYTE)0x01 },
    { 0x4480,(BYTE)0x4e,(BYTE)0x75,(BYTE)0x2f,(BYTE)0x2f,(BYTE)0x00,(BYTE)0x06,(BYTE)0x3f,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xa8,(BYTE)0x54,(BYTE)0x57,(BYTE)0x3f,(BYTE)0x3c },
    { 0x4490,(BYTE)0xff,(BYTE)0xff,(BYTE)0x3f,(BYTE)0x3c,(BYTE)0x00,(BYTE)0x01,(BYTE)0x2f,(BYTE)0x38,(BYTE)0x04,(BYTE)0xc6,(BYTE)0x3f,(BYTE)0x3c,(BYTE)0x00,(BYTE)0x0a,(BYTE)0x61,(BYTE)0x00 },
    { 0x44a0,(BYTE)0x02,(BYTE)0x46,(BYTE)0xde,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x10,(BYTE)0x4a,(BYTE)0x40,(BYTE)0x67,(BYTE)0x14,(BYTE)0x32,(BYTE)0x2f,(BYTE)0x00,(BYTE)0x04,(BYTE)0x61,(BYTE)0x00 },
    { 0x44b0,(BYTE)0x0a,(BYTE)0xac,(BYTE)0x0c,(BYTE)0x80,(BYTE)0x00,(BYTE)0x01,(BYTE)0x00,(BYTE)0x00,(BYTE)0x66,(BYTE)0x00,(BYTE)0x01,(BYTE)0xd6,(BYTE)0x60,(BYTE)0xc4,(BYTE)0x20,(BYTE)0x78 },
    { 0x44c0,(BYTE)0x04,(BYTE)0xc6,(BYTE)0x24,(BYTE)0x7c,(BYTE)0x00,(BYTE)0x00,(BYTE)0x01,(BYTE)0x94,(BYTE)0x30,(BYTE)0x3c,(BYTE)0x00,(BYTE)0x0b,(BYTE)0x61,(BYTE)0x00,(BYTE)0x01,(BYTE)0xc6 },
    { 0x44d0,(BYTE)0xb0,(BYTE)0x79,(BYTE)0x00,(BYTE)0x00,(BYTE)0x01,(BYTE)0x2e,(BYTE)0x62,(BYTE)0x00,(BYTE)0x01,(BYTE)0xb8,(BYTE)0x34,(BYTE)0xc0,(BYTE)0x67,(BYTE)0x00,(BYTE)0x01,(BYTE)0xb2 },
    { 0x44e0,(BYTE)0x32,(BYTE)0x00,(BYTE)0x74,(BYTE)0x09,(BYTE)0xe4,(BYTE)0x68,(BYTE)0x33,(BYTE)0xc0,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xfc,(BYTE)0x42,(BYTE)0x40,(BYTE)0x10,(BYTE)0x28 },
    { 0x44f0,(BYTE)0x00,(BYTE)0x0d,(BYTE)0x34,(BYTE)0xc0,(BYTE)0x67,(BYTE)0x00,(BYTE)0x01,(BYTE)0x9a,(BYTE)0xc0,(BYTE)0xc1,(BYTE)0x34,(BYTE)0xc0,(BYTE)0x70,(BYTE)0x11,(BYTE)0x61,(BYTE)0x00 },
    { 0x4500,(BYTE)0x01,(BYTE)0x94,(BYTE)0x4a,(BYTE)0x40,(BYTE)0x67,(BYTE)0x00,(BYTE)0x01,(BYTE)0x8a,(BYTE)0xeb,(BYTE)0x88,(BYTE)0x80,(BYTE)0xc1,(BYTE)0x22,(BYTE)0x00,(BYTE)0x48,(BYTE)0x41 },
    { 0x4510,(BYTE)0x4a,(BYTE)0x41,(BYTE)0x67,(BYTE)0x02,(BYTE)0x52,(BYTE)0x40,(BYTE)0x34,(BYTE)0xc0,(BYTE)0x34,(BYTE)0x00,(BYTE)0x30,(BYTE)0x3c,(BYTE)0x00,(BYTE)0x16,(BYTE)0x61,(BYTE)0x00 },
    { 0x4520,(BYTE)0x01,(BYTE)0x74,(BYTE)0x34,(BYTE)0xc0,(BYTE)0x67,(BYTE)0x00,(BYTE)0x01,(BYTE)0x6a,(BYTE)0x32,(BYTE)0x00,(BYTE)0x33,(BYTE)0xc0,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xf8 },
    { 0x4530,(BYTE)0x30,(BYTE)0x3c,(BYTE)0x00,(BYTE)0x0e,(BYTE)0x61,(BYTE)0x00,(BYTE)0x01,(BYTE)0x5e,(BYTE)0xd0,(BYTE)0x41,(BYTE)0x34,(BYTE)0xc0,(BYTE)0x33,(BYTE)0xc0,(BYTE)0x00,(BYTE)0x00 },
    { 0x4540,(BYTE)0x05,(BYTE)0xfa,(BYTE)0xd0,(BYTE)0x41,(BYTE)0xd0,(BYTE)0x42,(BYTE)0x34,(BYTE)0xc0,(BYTE)0x34,(BYTE)0x00,(BYTE)0x30,(BYTE)0x3c,(BYTE)0x00,(BYTE)0x13,(BYTE)0x61,(BYTE)0x00 },
    { 0x4550,(BYTE)0x01,(BYTE)0x44,(BYTE)0x90,(BYTE)0x42,(BYTE)0x67,(BYTE)0x00,(BYTE)0x01,(BYTE)0x3a,(BYTE)0x42,(BYTE)0x81,(BYTE)0x32,(BYTE)0x00,(BYTE)0x42,(BYTE)0x40,(BYTE)0x10,(BYTE)0x28 },
    { 0x4560,(BYTE)0x00,(BYTE)0x0d,(BYTE)0x82,(BYTE)0xc0,(BYTE)0x34,(BYTE)0xc1,(BYTE)0x34,(BYTE)0xb9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xac,(BYTE)0x34,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00 },
    { 0x4570,(BYTE)0x05,(BYTE)0xfc,(BYTE)0x43,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x01,(BYTE)0xd0,(BYTE)0x30,(BYTE)0x2f,(BYTE)0x00,(BYTE)0x04,(BYTE)0x13,(BYTE)0x82,(BYTE)0x00,(BYTE)0x00 },
    { 0x4580,(BYTE)0x08,(BYTE)0x39,(BYTE)0x00,(BYTE)0x06,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xa9,(BYTE)0x67,(BYTE)0x00,(BYTE)0x00,(BYTE)0xae,(BYTE)0x43,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00 },
    { 0x4590,(BYTE)0x01,(BYTE)0xa6,(BYTE)0xc0,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x03,(BYTE)0xd3,(BYTE)0xc0,(BYTE)0x32,(BYTE)0x3c,(BYTE)0x00,(BYTE)0x02,(BYTE)0x13,(BYTE)0xb0,(BYTE)0x10,(BYTE)0x08 },
    { 0x45a0,(BYTE)0x10,(BYTE)0x00,(BYTE)0x51,(BYTE)0xc9,(BYTE)0xff,(BYTE)0xf8,(BYTE)0x45,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x01,(BYTE)0xde,(BYTE)0x70,(BYTE)0x00,(BYTE)0x30,(BYTE)0x2f },
    { 0x45b0,(BYTE)0x00,(BYTE)0x04,(BYTE)0xed,(BYTE)0x88,(BYTE)0xd5,(BYTE)0xc0,(BYTE)0x30,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xfa,(BYTE)0xc0,(BYTE)0xc2,(BYTE)0x22,(BYTE)0x6f },
    { 0x45c0,(BYTE)0x00,(BYTE)0x06,(BYTE)0xd3,(BYTE)0xc0,(BYTE)0x32,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xf8,(BYTE)0x53,(BYTE)0x81,(BYTE)0x34,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00 },
    { 0x45d0,(BYTE)0x05,(BYTE)0xfc,(BYTE)0x53,(BYTE)0x42,(BYTE)0x42,(BYTE)0xb9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x06,(BYTE)0x02,(BYTE)0x48,(BYTE)0xe7,(BYTE)0x60,(BYTE)0xe0,(BYTE)0x2f,(BYTE)0x09 },
    { 0x45e0,(BYTE)0x3f,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xa8,(BYTE)0x54,(BYTE)0x57,(BYTE)0x3f,(BYTE)0x3c,(BYTE)0xff,(BYTE)0xff,(BYTE)0x3f,(BYTE)0x3c,(BYTE)0x00,(BYTE)0x01 },
    { 0x45f0,(BYTE)0x2f,(BYTE)0x08,(BYTE)0x3f,(BYTE)0x3c,(BYTE)0x00,(BYTE)0x0a,(BYTE)0x61,(BYTE)0x00,(BYTE)0x00,(BYTE)0xee,(BYTE)0xde,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x10,(BYTE)0x4c,(BYTE)0xdf },
    { 0x4600,(BYTE)0x07,(BYTE)0x06,(BYTE)0x4a,(BYTE)0x40,(BYTE)0x67,(BYTE)0x00,(BYTE)0x00,(BYTE)0x1e,(BYTE)0x48,(BYTE)0xe7,(BYTE)0x60,(BYTE)0xe0,(BYTE)0x32,(BYTE)0x2f,(BYTE)0x00,(BYTE)0x18 },
    { 0x4610,(BYTE)0x61,(BYTE)0x00,(BYTE)0x09,(BYTE)0x4a,(BYTE)0x4c,(BYTE)0xdf,(BYTE)0x07,(BYTE)0x06,(BYTE)0x0c,(BYTE)0x80,(BYTE)0x00,(BYTE)0x01,(BYTE)0x00,(BYTE)0x00,(BYTE)0x67,(BYTE)0xba },
    { 0x4620,(BYTE)0x60,(BYTE)0x00,(BYTE)0x00,(BYTE)0x6e,(BYTE)0x61,(BYTE)0x00,(BYTE)0x00,(BYTE)0x80,(BYTE)0x52,(BYTE)0x49,(BYTE)0x51,(BYTE)0xca,(BYTE)0xff,(BYTE)0xae,(BYTE)0x61,(BYTE)0x00 },
    { 0x4630,(BYTE)0x00,(BYTE)0x96,(BYTE)0x14,(BYTE)0xc0,(BYTE)0x51,(BYTE)0xc9,(BYTE)0xff,(BYTE)0x96,(BYTE)0x30,(BYTE)0x2f,(BYTE)0x00,(BYTE)0x04,(BYTE)0x41,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00 },
    { 0x4640,(BYTE)0x01,(BYTE)0x74,(BYTE)0x42,(BYTE)0x30,(BYTE)0x00,(BYTE)0x00,(BYTE)0x41,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x01,(BYTE)0x82,(BYTE)0x11,(BYTE)0xbc,(BYTE)0x00,(BYTE)0x02 },
    { 0x4650,(BYTE)0x00,(BYTE)0x00,(BYTE)0x41,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0x5e,(BYTE)0xe3,(BYTE)0x40,(BYTE)0x31,(BYTE)0xb9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xfa },
    { 0x4660,(BYTE)0x00,(BYTE)0x00,(BYTE)0x41,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0x7a,(BYTE)0x32,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xfa,(BYTE)0xd2,(BYTE)0x79 },
    { 0x4670,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xf8,(BYTE)0x53,(BYTE)0x41,(BYTE)0x31,(BYTE)0x81,(BYTE)0x00,(BYTE)0x00,(BYTE)0x41,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0xec },
    { 0x4680,(BYTE)0xe3,(BYTE)0x40,(BYTE)0x21,(BYTE)0xaf,(BYTE)0x00,(BYTE)0x06,(BYTE)0x00,(BYTE)0x00,(BYTE)0x20,(BYTE)0x3c,(BYTE)0x00,(BYTE)0x00,(BYTE)0x01,(BYTE)0x94,(BYTE)0x60,(BYTE)0x02 },
    { 0x4690,(BYTE)0x70,(BYTE)0xff,(BYTE)0x4e,(BYTE)0x75,(BYTE)0x3f,(BYTE)0x01,(BYTE)0x12,(BYTE)0x30,(BYTE)0x00,(BYTE)0x01,(BYTE)0xe1,(BYTE)0x49,(BYTE)0x12,(BYTE)0x30,(BYTE)0x00,(BYTE)0x00 },
    { 0x46a0,(BYTE)0x30,(BYTE)0x01,(BYTE)0x32,(BYTE)0x1f,(BYTE)0x4e,(BYTE)0x75,(BYTE)0x48,(BYTE)0xe7,(BYTE)0x40,(BYTE)0x80,(BYTE)0x20,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00,(BYTE)0x06,(BYTE)0x02 },
    { 0x46b0,(BYTE)0x32,(BYTE)0x3c,(BYTE)0x00,(BYTE)0x7f,(BYTE)0xd0,(BYTE)0x98,(BYTE)0x51,(BYTE)0xc9,(BYTE)0xff,(BYTE)0xfc,(BYTE)0x23,(BYTE)0xc0,(BYTE)0x00,(BYTE)0x00,(BYTE)0x06,(BYTE)0x02 },
    { 0x46c0,(BYTE)0x4c,(BYTE)0xdf,(BYTE)0x01,(BYTE)0x02,(BYTE)0x4e,(BYTE)0x75,(BYTE)0x30,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00,(BYTE)0x06,(BYTE)0x04,(BYTE)0xb1,(BYTE)0x79,(BYTE)0x00,(BYTE)0x00 },
    { 0x46d0,(BYTE)0x06,(BYTE)0x02,(BYTE)0x10,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00,(BYTE)0x06,(BYTE)0x03,(BYTE)0xb1,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00,(BYTE)0x06,(BYTE)0x02,(BYTE)0x10,(BYTE)0x39 },
    { 0x46e0,(BYTE)0x00,(BYTE)0x00,(BYTE)0x06,(BYTE)0x02,(BYTE)0x4e,(BYTE)0x75,(BYTE)0x4e,(BYTE)0x56,(BYTE)0x00,(BYTE)0x00,(BYTE)0x55,(BYTE)0x6e,(BYTE)0x00,(BYTE)0x12,(BYTE)0x30,(BYTE)0x2e },
    { 0x46f0,(BYTE)0x00,(BYTE)0x12,(BYTE)0x08,(BYTE)0x2e,(BYTE)0x00,(BYTE)0x03,(BYTE)0x00,(BYTE)0x09,(BYTE)0x67,(BYTE)0x0a,(BYTE)0x33,(BYTE)0xc0,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xa8 },
    { 0x4700,(BYTE)0x60,(BYTE)0x00,(BYTE)0x01,(BYTE)0x76,(BYTE)0x41,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0xd6,(BYTE)0x13,(BYTE)0xf0,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00 },
    { 0x4710,(BYTE)0x05,(BYTE)0xa9,(BYTE)0x41,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x01,(BYTE)0xd0,(BYTE)0x13,(BYTE)0xf0,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xfd },
    { 0x4720,(BYTE)0x41,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0xec,(BYTE)0x32,(BYTE)0x00,(BYTE)0xd2,(BYTE)0x41,(BYTE)0xd2,(BYTE)0x41,(BYTE)0x23,(BYTE)0xf0,(BYTE)0x10,(BYTE)0x00 },
    { 0x4730,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xfe,(BYTE)0x08,(BYTE)0x2e,(BYTE)0x00,(BYTE)0x01,(BYTE)0x00,(BYTE)0x09,(BYTE)0x66,(BYTE)0x00,(BYTE)0x01,(BYTE)0x3c,(BYTE)0x41,(BYTE)0xf9 },
    { 0x4740,(BYTE)0x00,(BYTE)0x00,(BYTE)0x01,(BYTE)0x74,(BYTE)0x10,(BYTE)0x30,(BYTE)0x00,(BYTE)0x00,(BYTE)0x67,(BYTE)0x00,(BYTE)0x01,(BYTE)0x2e,(BYTE)0x0c,(BYTE)0x00,(BYTE)0x00,(BYTE)0x02 },
    { 0x4750,(BYTE)0x67,(BYTE)0x00,(BYTE)0x02,(BYTE)0xa4,(BYTE)0x2f,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xfe,(BYTE)0x3f,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xa8 },
    { 0x4760,(BYTE)0x54,(BYTE)0x57,(BYTE)0x3f,(BYTE)0x3c,(BYTE)0xff,(BYTE)0xff,(BYTE)0x3f,(BYTE)0x3c,(BYTE)0x00,(BYTE)0x01,(BYTE)0x2f,(BYTE)0x38,(BYTE)0x04,(BYTE)0xc6,(BYTE)0x3f,(BYTE)0x3c },
    { 0x4770,(BYTE)0x00,(BYTE)0x0a,(BYTE)0x61,(BYTE)0x00,(BYTE)0xff,(BYTE)0x72,(BYTE)0xde,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x10,(BYTE)0x4a,(BYTE)0x40,(BYTE)0x67,(BYTE)0x14,(BYTE)0x32,(BYTE)0x2e },
    { 0x4780,(BYTE)0x00,(BYTE)0x12,(BYTE)0x61,(BYTE)0x00,(BYTE)0x07,(BYTE)0xd8,(BYTE)0x0c,(BYTE)0x80,(BYTE)0x00,(BYTE)0x01,(BYTE)0x00,(BYTE)0x00,(BYTE)0x67,(BYTE)0xc6,(BYTE)0x60,(BYTE)0x00 },
    { 0x4790,(BYTE)0x02,(BYTE)0x68,(BYTE)0x43,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x01,(BYTE)0xa6,(BYTE)0x30,(BYTE)0x2e,(BYTE)0x00,(BYTE)0x12,(BYTE)0xc0,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x03 },
    { 0x47a0,(BYTE)0xd3,(BYTE)0xc0,(BYTE)0x24,(BYTE)0x78,(BYTE)0x04,(BYTE)0xc6,(BYTE)0x50,(BYTE)0x4a,(BYTE)0x30,(BYTE)0x3c,(BYTE)0x00,(BYTE)0x02,(BYTE)0xb3,(BYTE)0x0a,(BYTE)0x66,(BYTE)0x00 },
    { 0x47b0,(BYTE)0x02,(BYTE)0x2a,(BYTE)0x51,(BYTE)0xc8,(BYTE)0xff,(BYTE)0xf8,(BYTE)0x72,(BYTE)0x00,(BYTE)0x32,(BYTE)0x2e,(BYTE)0x00,(BYTE)0x12,(BYTE)0x43,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00 },
    { 0x47c0,(BYTE)0x01,(BYTE)0xde,(BYTE)0xed,(BYTE)0x89,(BYTE)0xd3,(BYTE)0xc1,(BYTE)0xd0,(BYTE)0x40,(BYTE)0x45,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0x5e,(BYTE)0x33,(BYTE)0xf2 },
    { 0x47d0,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xfa,(BYTE)0x45,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0x7a,(BYTE)0x32,(BYTE)0x32,(BYTE)0x00,(BYTE)0x00 },
    { 0x47e0,(BYTE)0x92,(BYTE)0x79,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xfa,(BYTE)0x24,(BYTE)0x79,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xfe,(BYTE)0x34,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00 },
    { 0x47f0,(BYTE)0x05,(BYTE)0xfa,(BYTE)0xc4,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xfc,(BYTE)0xd5,(BYTE)0xc2,(BYTE)0x20,(BYTE)0x78,(BYTE)0x04,(BYTE)0xc6,(BYTE)0x34,(BYTE)0x39 },
    { 0x4800,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xfc,(BYTE)0x53,(BYTE)0x42,(BYTE)0x42,(BYTE)0xb9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x06,(BYTE)0x02,(BYTE)0x48,(BYTE)0xe7,(BYTE)0x60,(BYTE)0xe0 },
    { 0x4810,(BYTE)0x2f,(BYTE)0x0a,(BYTE)0x3f,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xa8,(BYTE)0x54,(BYTE)0x57,(BYTE)0x3f,(BYTE)0x3c,(BYTE)0xff,(BYTE)0xff,(BYTE)0x3f,(BYTE)0x3c },
    { 0x4820,(BYTE)0x00,(BYTE)0x01,(BYTE)0x2f,(BYTE)0x08,(BYTE)0x3f,(BYTE)0x3c,(BYTE)0x00,(BYTE)0x0a,(BYTE)0x61,(BYTE)0x00,(BYTE)0xfe,(BYTE)0xbc,(BYTE)0xde,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x10 },
    { 0x4830,(BYTE)0x4c,(BYTE)0xdf,(BYTE)0x07,(BYTE)0x06,(BYTE)0x4a,(BYTE)0x40,(BYTE)0x67,(BYTE)0x1c,(BYTE)0x48,(BYTE)0xe7,(BYTE)0x60,(BYTE)0xe0,(BYTE)0x32,(BYTE)0x2e,(BYTE)0x00,(BYTE)0x12 },
    { 0x4840,(BYTE)0x61,(BYTE)0x00,(BYTE)0x07,(BYTE)0x1a,(BYTE)0x4c,(BYTE)0xdf,(BYTE)0x07,(BYTE)0x06,(BYTE)0x0c,(BYTE)0x80,(BYTE)0x00,(BYTE)0x01,(BYTE)0x00,(BYTE)0x00,(BYTE)0x67,(BYTE)0xbc },
    { 0x4850,(BYTE)0x60,(BYTE)0x00,(BYTE)0x01,(BYTE)0xa6,(BYTE)0x61,(BYTE)0x00,(BYTE)0xfe,(BYTE)0x50,(BYTE)0x52,(BYTE)0x4a,(BYTE)0x51,(BYTE)0xca,(BYTE)0xff,(BYTE)0xb0,(BYTE)0x61,(BYTE)0x00 },
    { 0x4860,(BYTE)0xfe,(BYTE)0x66,(BYTE)0xb0,(BYTE)0x19,(BYTE)0x66,(BYTE)0x00,(BYTE)0x01,(BYTE)0x74,(BYTE)0x51,(BYTE)0xc9,(BYTE)0xff,(BYTE)0x94,(BYTE)0x41,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00 },
    { 0x4870,(BYTE)0x01,(BYTE)0x74,(BYTE)0xd0,(BYTE)0xee,(BYTE)0x00,(BYTE)0x12,(BYTE)0x42,(BYTE)0x10,(BYTE)0x4a,(BYTE)0x6e,(BYTE)0x00,(BYTE)0x0e,(BYTE)0x67,(BYTE)0x00,(BYTE)0x01,(BYTE)0x58 },
    { 0x4880,(BYTE)0x0c,(BYTE)0x6e,(BYTE)0xff,(BYTE)0xff,(BYTE)0x00,(BYTE)0x10,(BYTE)0x66,(BYTE)0x06,(BYTE)0x22,(BYTE)0x6e,(BYTE)0x00,(BYTE)0x14,(BYTE)0x60,(BYTE)0x08,(BYTE)0x70,(BYTE)0x00 },
    { 0x4890,(BYTE)0x30,(BYTE)0x2e,(BYTE)0x00,(BYTE)0x10,(BYTE)0x22,(BYTE)0x40,(BYTE)0x23,(BYTE)0xc9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xae,(BYTE)0x22,(BYTE)0x09,(BYTE)0x74,(BYTE)0x00 },
    { 0x48a0,(BYTE)0x34,(BYTE)0x2e,(BYTE)0x00,(BYTE)0x0e,(BYTE)0xd3,(BYTE)0xc2,(BYTE)0x53,(BYTE)0x89,(BYTE)0x23,(BYTE)0xc9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xb2,(BYTE)0x23,(BYTE)0xee },
    { 0x48b0,(BYTE)0x00,(BYTE)0x0a,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xb6,(BYTE)0x08,(BYTE)0x2e,(BYTE)0x00,(BYTE)0x03,(BYTE)0x00,(BYTE)0x09,(BYTE)0x66,(BYTE)0x12,(BYTE)0xc2,(BYTE)0xf9 },
    { 0x48c0,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xfc,(BYTE)0xd2,(BYTE)0xb9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xfe,(BYTE)0xc4,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xfc },
    { 0x48d0,(BYTE)0x3f,(BYTE)0x2e,(BYTE)0x00,(BYTE)0x12,(BYTE)0x2f,(BYTE)0x01,(BYTE)0x2f,(BYTE)0x02,(BYTE)0x2f,(BYTE)0x2e,(BYTE)0x00,(BYTE)0x0a,(BYTE)0x3f,(BYTE)0x2e,(BYTE)0x00,(BYTE)0x08 },
    { 0x48e0,(BYTE)0x61,(BYTE)0x00,(BYTE)0x01,(BYTE)0x62,(BYTE)0xde,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x10,(BYTE)0x4a,(BYTE)0x80,(BYTE)0x67,(BYTE)0x18,(BYTE)0x0c,(BYTE)0x80,(BYTE)0xff,(BYTE)0xff },
    { 0x48f0,(BYTE)0xff,(BYTE)0xf2,(BYTE)0x66,(BYTE)0x00,(BYTE)0x01,(BYTE)0x04,(BYTE)0x08,(BYTE)0x2e,(BYTE)0x00,(BYTE)0x03,(BYTE)0x00,(BYTE)0x09,(BYTE)0x66,(BYTE)0x00,(BYTE)0x00,(BYTE)0xfa },
    { 0x4900,(BYTE)0x60,(BYTE)0x00,(BYTE)0xfe,(BYTE)0x52,(BYTE)0x30,(BYTE)0x2e,(BYTE)0x00,(BYTE)0x08,(BYTE)0x08,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x67,(BYTE)0x00,(BYTE)0x00,(BYTE)0xc8 },
    { 0x4910,(BYTE)0x08,(BYTE)0x39,(BYTE)0x00,(BYTE)0x06,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xa9,(BYTE)0x67,(BYTE)0x00,(BYTE)0x00,(BYTE)0xbc,(BYTE)0x08,(BYTE)0x00,(BYTE)0x00,(BYTE)0x03 },
    { 0x4920,(BYTE)0x66,(BYTE)0x00,(BYTE)0x00,(BYTE)0xb4,(BYTE)0x4a,(BYTE)0xb9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xae,(BYTE)0x66,(BYTE)0x0e,(BYTE)0x41,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00 },
    { 0x4930,(BYTE)0x01,(BYTE)0x74,(BYTE)0xd0,(BYTE)0xee,(BYTE)0x00,(BYTE)0x12,(BYTE)0x10,(BYTE)0xbc,(BYTE)0x00,(BYTE)0x02,(BYTE)0x41,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0x7a },
    { 0x4940,(BYTE)0x30,(BYTE)0x2e,(BYTE)0x00,(BYTE)0x12,(BYTE)0xd0,(BYTE)0x40,(BYTE)0x72,(BYTE)0x00,(BYTE)0x32,(BYTE)0x30,(BYTE)0x00,(BYTE)0x00,(BYTE)0xb2,(BYTE)0xb9,(BYTE)0x00,(BYTE)0x00 },
    { 0x4950,(BYTE)0x05,(BYTE)0xae,(BYTE)0x6d,(BYTE)0x00,(BYTE)0x00,(BYTE)0x82,(BYTE)0x41,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0x5e,(BYTE)0x74,(BYTE)0x00,(BYTE)0x34,(BYTE)0x30 },
    { 0x4960,(BYTE)0x00,(BYTE)0x00,(BYTE)0xb4,(BYTE)0xb9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xb2,(BYTE)0x6e,(BYTE)0x00,(BYTE)0x00,(BYTE)0x6c,(BYTE)0x20,(BYTE)0x79,(BYTE)0x00,(BYTE)0x00 },
    { 0x4970,(BYTE)0x05,(BYTE)0xb6,(BYTE)0x43,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x01,(BYTE)0xde,(BYTE)0x70,(BYTE)0x00,(BYTE)0x30,(BYTE)0x2e,(BYTE)0x00,(BYTE)0x12,(BYTE)0xed,(BYTE)0x88 },
    { 0x4980,(BYTE)0xd3,(BYTE)0xc0,(BYTE)0x20,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xae,(BYTE)0x90,(BYTE)0x82,(BYTE)0x67,(BYTE)0x14,(BYTE)0x6d,(BYTE)0x0a,(BYTE)0x24,(BYTE)0x39 },
    { 0x4990,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xae,(BYTE)0xd3,(BYTE)0xc0,(BYTE)0x60,(BYTE)0x08,(BYTE)0x44,(BYTE)0x80,(BYTE)0xe1,(BYTE)0x80,(BYTE)0xd0,(BYTE)0x80,(BYTE)0xd1,(BYTE)0xc0 },
    { 0x49a0,(BYTE)0xb2,(BYTE)0xb9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xb2,(BYTE)0x6f,(BYTE)0x06,(BYTE)0x22,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xb2,(BYTE)0x92,(BYTE)0x42 },
    { 0x49b0,(BYTE)0x34,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xfc,(BYTE)0x53,(BYTE)0x42,(BYTE)0x42,(BYTE)0xb9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x06,(BYTE)0x02,(BYTE)0x61,(BYTE)0x00 },
    { 0x49c0,(BYTE)0xfc,(BYTE)0xe6,(BYTE)0xd1,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x00,(BYTE)0x02,(BYTE)0x00,(BYTE)0x51,(BYTE)0xca,(BYTE)0xff,(BYTE)0xf4,(BYTE)0x61,(BYTE)0x00,(BYTE)0xfc,(BYTE)0xf8 },
    { 0x49d0,(BYTE)0x12,(BYTE)0xc0,(BYTE)0x51,(BYTE)0xc9,(BYTE)0xff,(BYTE)0xdc,(BYTE)0x42,(BYTE)0x80,(BYTE)0x60,(BYTE)0x1e,(BYTE)0x41,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x01,(BYTE)0x74 },
    { 0x49e0,(BYTE)0xd0,(BYTE)0xee,(BYTE)0x00,(BYTE)0x12,(BYTE)0x10,(BYTE)0xbc,(BYTE)0x00,(BYTE)0x02,(BYTE)0x41,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x01,(BYTE)0x82,(BYTE)0xd0,(BYTE)0xee },
    { 0x49f0,(BYTE)0x00,(BYTE)0x12,(BYTE)0x10,(BYTE)0xbc,(BYTE)0x00,(BYTE)0x02,(BYTE)0x70,(BYTE)0xf2,(BYTE)0x08,(BYTE)0x2e,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x09,(BYTE)0x66,(BYTE)0x20 },
    { 0x4a00,(BYTE)0x4a,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00,(BYTE)0x06,(BYTE)0x0a,(BYTE)0x67,(BYTE)0x18,(BYTE)0x2f,(BYTE)0x00,(BYTE)0x40,(BYTE)0xe7,(BYTE)0x00,(BYTE)0x7c,(BYTE)0x07,(BYTE)0x00 },
    { 0x4a10,(BYTE)0x4e,(BYTE)0x7a,(BYTE)0x00,(BYTE)0x02,(BYTE)0x00,(BYTE)0x40,(BYTE)0x08,(BYTE)0x08,(BYTE)0x4e,(BYTE)0x7b,(BYTE)0x00,(BYTE)0x02,(BYTE)0x46,(BYTE)0xdf,(BYTE)0x20,(BYTE)0x1f },
    { 0x4a20,(BYTE)0x4e,(BYTE)0x5e,(BYTE)0x4e,(BYTE)0x75,(BYTE)0x3f,(BYTE)0x01,(BYTE)0x44,(BYTE)0x41,(BYTE)0x06,(BYTE)0x41,(BYTE)0x00,(BYTE)0x09,(BYTE)0xe3,(BYTE)0x60,(BYTE)0x53,(BYTE)0x40 },
    { 0x4a30,(BYTE)0x32,(BYTE)0x1f,(BYTE)0x66,(BYTE)0x08,(BYTE)0x12,(BYTE)0xda,(BYTE)0x51,(BYTE)0xc8,(BYTE)0xff,(BYTE)0xfc,(BYTE)0x4e,(BYTE)0x75,(BYTE)0x22,(BYTE)0xda,(BYTE)0x51,(BYTE)0xc8 },
    { 0x4a40,(BYTE)0xff,(BYTE)0xfc,(BYTE)0x4e,(BYTE)0x75,(BYTE)0x4e,(BYTE)0x56,(BYTE)0x00,(BYTE)0x00,(BYTE)0x24,(BYTE)0x2e,(BYTE)0x00,(BYTE)0x0e,(BYTE)0x22,(BYTE)0x6e,(BYTE)0x00,(BYTE)0x0a },
    { 0x4a50,(BYTE)0x08,(BYTE)0x39,(BYTE)0x00,(BYTE)0x04,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xa9,(BYTE)0x66,(BYTE)0x00,(BYTE)0x01,(BYTE)0x04,(BYTE)0x08,(BYTE)0x39,(BYTE)0x00,(BYTE)0x03 },
    { 0x4a60,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xa9,(BYTE)0x66,(BYTE)0x00,(BYTE)0x00,(BYTE)0xc4,(BYTE)0xb4,(BYTE)0xbc,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0xfe,(BYTE)0x63,(BYTE)0x06 },
    { 0x4a70,(BYTE)0x24,(BYTE)0x3c,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0xfe,(BYTE)0x08,(BYTE)0x2e,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x0d,(BYTE)0x66,(BYTE)0x46,(BYTE)0x0c,(BYTE)0x79 },
    { 0x4a80,(BYTE)0x00,(BYTE)0x02,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xa4,(BYTE)0x66,(BYTE)0x20,(BYTE)0xb3,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x9f,(BYTE)0xff,(BYTE)0xff,(BYTE)0x63,(BYTE)0x00 },
    { 0x4a90,(BYTE)0x00,(BYTE)0xea,(BYTE)0xb3,(BYTE)0xfc,(BYTE)0xff,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x65,(BYTE)0x2a,(BYTE)0xb3,(BYTE)0xfc,(BYTE)0xff,(BYTE)0x9f,(BYTE)0xff,(BYTE)0xff },
    { 0x4aa0,(BYTE)0x63,(BYTE)0x00,(BYTE)0x00,(BYTE)0xd8,(BYTE)0x60,(BYTE)0x00,(BYTE)0x00,(BYTE)0x1e,(BYTE)0xb3,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x3f,(BYTE)0xff,(BYTE)0xff,(BYTE)0x63,(BYTE)0x00 },
    { 0x4ab0,(BYTE)0x00,(BYTE)0xca,(BYTE)0xb3,(BYTE)0xfc,(BYTE)0xff,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x65,(BYTE)0x0a,(BYTE)0xb3,(BYTE)0xfc,(BYTE)0xff,(BYTE)0x3f,(BYTE)0xff,(BYTE)0xff },
    { 0x4ac0,(BYTE)0x63,(BYTE)0x00,(BYTE)0x00,(BYTE)0xb8,(BYTE)0x72,(BYTE)0x00,(BYTE)0x08,(BYTE)0x2e,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x0d,(BYTE)0x66,(BYTE)0x02,(BYTE)0x72,(BYTE)0x02 },
    { 0x4ad0,(BYTE)0x4a,(BYTE)0xb9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xbc,(BYTE)0x66,(BYTE)0x2a,(BYTE)0x2f,(BYTE)0x01,(BYTE)0x2f,(BYTE)0x3c,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xbc },
    { 0x4ae0,(BYTE)0x2f,(BYTE)0x3c,(BYTE)0x5f,(BYTE)0x46,(BYTE)0x52,(BYTE)0x42,(BYTE)0x61,(BYTE)0x00,(BYTE)0x0d,(BYTE)0xb0,(BYTE)0x50,(BYTE)0x4f,(BYTE)0x22,(BYTE)0x1f,(BYTE)0x4a,(BYTE)0x40 },
    { 0x4af0,(BYTE)0x66,(BYTE)0x10,(BYTE)0x22,(BYTE)0x78,(BYTE)0x04,(BYTE)0xc6,(BYTE)0x0c,(BYTE)0x42,(BYTE)0x00,(BYTE)0x02,(BYTE)0x63,(BYTE)0x16,(BYTE)0x34,(BYTE)0x3c,(BYTE)0x00,(BYTE)0x02 },
    { 0x4b00,(BYTE)0x60,(BYTE)0x10,(BYTE)0x22,(BYTE)0x79,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xbc,(BYTE)0x0c,(BYTE)0x42,(BYTE)0x00,(BYTE)0x80,(BYTE)0x63,(BYTE)0x04,(BYTE)0x34,(BYTE)0x3c },
    { 0x4b10,(BYTE)0x00,(BYTE)0x80,(BYTE)0x08,(BYTE)0x2e,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x09,(BYTE)0x67,(BYTE)0x60,(BYTE)0x2f,(BYTE)0x09,(BYTE)0x24,(BYTE)0x6e,(BYTE)0x00,(BYTE)0x0a },
    { 0x4b20,(BYTE)0x30,(BYTE)0x02,(BYTE)0x61,(BYTE)0x00,(BYTE)0xff,(BYTE)0x00,(BYTE)0x22,(BYTE)0x5f,(BYTE)0x60,(BYTE)0x50,(BYTE)0x0c,(BYTE)0x82,(BYTE)0x00,(BYTE)0x00,(BYTE)0xff,(BYTE)0xfe },
    { 0x4b30,(BYTE)0x63,(BYTE)0x06,(BYTE)0x24,(BYTE)0x3c,(BYTE)0x00,(BYTE)0x00,(BYTE)0xff,(BYTE)0xfe,(BYTE)0xb3,(BYTE)0xfc,(BYTE)0xc0,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x65,(BYTE)0x3a },
    { 0x4b40,(BYTE)0xb3,(BYTE)0xfc,(BYTE)0xfc,(BYTE)0xff,(BYTE)0xff,(BYTE)0xff,(BYTE)0x63,(BYTE)0x00,(BYTE)0xff,(BYTE)0x7c,(BYTE)0xb3,(BYTE)0xfc,(BYTE)0xfe,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00 },
    { 0x4b50,(BYTE)0x65,(BYTE)0x28,(BYTE)0xb3,(BYTE)0xfc,(BYTE)0xfe,(BYTE)0xff,(BYTE)0xff,(BYTE)0xff,(BYTE)0x63,(BYTE)0x00,(BYTE)0xff,(BYTE)0x6a,(BYTE)0x60,(BYTE)0x1c,(BYTE)0xb4,(BYTE)0xbc },
    { 0x4b60,(BYTE)0x00,(BYTE)0x00,(BYTE)0x01,(BYTE)0x00,(BYTE)0x63,(BYTE)0x06,(BYTE)0x24,(BYTE)0x3c,(BYTE)0x00,(BYTE)0x00,(BYTE)0x01,(BYTE)0x00,(BYTE)0x08,(BYTE)0x2e,(BYTE)0x00,(BYTE)0x00 },
    { 0x4b70,(BYTE)0x00,(BYTE)0x0d,(BYTE)0x66,(BYTE)0x00,(BYTE)0xff,(BYTE)0x50,(BYTE)0x22,(BYTE)0x6e,(BYTE)0x00,(BYTE)0x0a,(BYTE)0x33,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xc0 },
    { 0x4b80,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xc2,(BYTE)0x08,(BYTE)0x2e,(BYTE)0x00,(BYTE)0x02,(BYTE)0x00,(BYTE)0x09,(BYTE)0x67,(BYTE)0x08,(BYTE)0x33,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x00 },
    { 0x4b90,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xc2,(BYTE)0x48,(BYTE)0xe7,(BYTE)0x60,(BYTE)0x40,(BYTE)0x3f,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xa8,(BYTE)0x2f,(BYTE)0x09 },
    { 0x4ba0,(BYTE)0x3f,(BYTE)0x02,(BYTE)0x2f,(BYTE)0x2e,(BYTE)0x00,(BYTE)0x12,(BYTE)0x08,(BYTE)0x39,(BYTE)0x00,(BYTE)0x04,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xa9,(BYTE)0x66,(BYTE)0x32 },
    { 0x4bb0,(BYTE)0x0c,(BYTE)0x82,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0xfe,(BYTE)0x62,(BYTE)0x16,(BYTE)0x08,(BYTE)0x2e,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x09,(BYTE)0x66,(BYTE)0x08 },
    { 0x4bc0,(BYTE)0x61,(BYTE)0x00,(BYTE)0x0b,(BYTE)0xc6,(BYTE)0x60,(BYTE)0x00,(BYTE)0x00,(BYTE)0x3c,(BYTE)0x61,(BYTE)0x00,(BYTE)0x0b,(BYTE)0xf4,(BYTE)0x60,(BYTE)0x34,(BYTE)0x08,(BYTE)0x2e },
    { 0x4bd0,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x09,(BYTE)0x66,(BYTE)0x06,(BYTE)0x61,(BYTE)0x00,(BYTE)0x0c,(BYTE)0x46,(BYTE)0x60,(BYTE)0x26,(BYTE)0x61,(BYTE)0x00,(BYTE)0x0c,(BYTE)0x76 },
    { 0x4be0,(BYTE)0x60,(BYTE)0x20,(BYTE)0x41,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x01,(BYTE)0x90,(BYTE)0x3f,(BYTE)0x28,(BYTE)0x00,(BYTE)0x02,(BYTE)0x3f,(BYTE)0x10,(BYTE)0x08,(BYTE)0x2e },
    { 0x4bf0,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x09,(BYTE)0x66,(BYTE)0x06,(BYTE)0x61,(BYTE)0x00,(BYTE)0x08,(BYTE)0x94,(BYTE)0x60,(BYTE)0x04,(BYTE)0x61,(BYTE)0x00,(BYTE)0x08,(BYTE)0xda },
    { 0x4c00,(BYTE)0x58,(BYTE)0x4f,(BYTE)0xde,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x0c,(BYTE)0x4c,(BYTE)0xdf,(BYTE)0x02,(BYTE)0x06,(BYTE)0x4a,(BYTE)0x80,(BYTE)0x67,(BYTE)0x00,(BYTE)0x00,(BYTE)0x74 },
    { 0x4c10,(BYTE)0x6b,(BYTE)0x40,(BYTE)0x48,(BYTE)0xe7,(BYTE)0x60,(BYTE)0x40,(BYTE)0x3f,(BYTE)0x2e,(BYTE)0x00,(BYTE)0x08,(BYTE)0x61,(BYTE)0x00,(BYTE)0x03,(BYTE)0x50,(BYTE)0x54,(BYTE)0x4f },
    { 0x4c20,(BYTE)0x4c,(BYTE)0xdf,(BYTE)0x02,(BYTE)0x06,(BYTE)0x0c,(BYTE)0x80,(BYTE)0xff,(BYTE)0xff,(BYTE)0xff,(BYTE)0xf2,(BYTE)0x66,(BYTE)0x16,(BYTE)0x10,(BYTE)0x3c,(BYTE)0x00,(BYTE)0x01 },
    { 0x4c30,(BYTE)0x61,(BYTE)0x00,(BYTE)0x01,(BYTE)0x0a,(BYTE)0x08,(BYTE)0x2e,(BYTE)0x00,(BYTE)0x01,(BYTE)0x00,(BYTE)0x09,(BYTE)0x66,(BYTE)0x00,(BYTE)0xff,(BYTE)0x58,(BYTE)0x60,(BYTE)0x00 },
    { 0x4c40,(BYTE)0x00,(BYTE)0x8e,(BYTE)0x0c,(BYTE)0x80,(BYTE)0xff,(BYTE)0xff,(BYTE)0xff,(BYTE)0xf3,(BYTE)0x67,(BYTE)0x12,(BYTE)0x0c,(BYTE)0x80,(BYTE)0xff,(BYTE)0xff,(BYTE)0xff,(BYTE)0xfe },
    { 0x4c50,(BYTE)0x67,(BYTE)0x0a,(BYTE)0x53,(BYTE)0x79,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xc2,(BYTE)0x6a,(BYTE)0x00,(BYTE)0xff,(BYTE)0x3a,(BYTE)0x08,(BYTE)0x2e,(BYTE)0x00,(BYTE)0x03 },
    { 0x4c60,(BYTE)0x00,(BYTE)0x09,(BYTE)0x66,(BYTE)0x00,(BYTE)0x00,(BYTE)0x6a,(BYTE)0x48,(BYTE)0xe7,(BYTE)0x60,(BYTE)0x40,(BYTE)0x32,(BYTE)0x2e,(BYTE)0x00,(BYTE)0x16,(BYTE)0x61,(BYTE)0x00 },
    { 0x4c70,(BYTE)0x02,(BYTE)0xec,(BYTE)0x4c,(BYTE)0xdf,(BYTE)0x02,(BYTE)0x06,(BYTE)0x0c,(BYTE)0x80,(BYTE)0x00,(BYTE)0x01,(BYTE)0x00,(BYTE)0x00,(BYTE)0x67,(BYTE)0x00,(BYTE)0xff,(BYTE)0x16 },
    { 0x4c80,(BYTE)0x60,(BYTE)0x4c,(BYTE)0x41,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x11,(BYTE)0x5a,(BYTE)0xb3,(BYTE)0xee,(BYTE)0x00,(BYTE)0x0a,(BYTE)0x67,(BYTE)0x1e,(BYTE)0x41,(BYTE)0xf9 },
    { 0x4c90,(BYTE)0x00,(BYTE)0x00,(BYTE)0x10,(BYTE)0xf6,(BYTE)0x08,(BYTE)0x2e,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x09,(BYTE)0x66,(BYTE)0x10,(BYTE)0x2f,(BYTE)0x09,(BYTE)0x24,(BYTE)0x49 },
    { 0x4ca0,(BYTE)0x22,(BYTE)0x6e,(BYTE)0x00,(BYTE)0x0a,(BYTE)0x30,(BYTE)0x02,(BYTE)0x61,(BYTE)0x00,(BYTE)0xfd,(BYTE)0x7c,(BYTE)0x22,(BYTE)0x5f,(BYTE)0x95,(BYTE)0xae,(BYTE)0x00,(BYTE)0x0e },
    { 0x4cb0,(BYTE)0x67,(BYTE)0x1a,(BYTE)0xd5,(BYTE)0xae,(BYTE)0x00,(BYTE)0x12,(BYTE)0x20,(BYTE)0x02,(BYTE)0xe1,(BYTE)0x80,(BYTE)0xd0,(BYTE)0x80,(BYTE)0xd1,(BYTE)0xae,(BYTE)0x00,(BYTE)0x0a },
    { 0x4cc0,(BYTE)0xb4,(BYTE)0xae,(BYTE)0x00,(BYTE)0x0e,(BYTE)0x63,(BYTE)0x04,(BYTE)0x24,(BYTE)0x2e,(BYTE)0x00,(BYTE)0x0e,(BYTE)0x4e,(BYTE)0xd0,(BYTE)0x70,(BYTE)0x00,(BYTE)0x4e,(BYTE)0x5e },
    { 0x4cd0,(BYTE)0x4e,(BYTE)0x75,(BYTE)0x55,(BYTE)0x6f,(BYTE)0x00,(BYTE)0x04,(BYTE)0x32,(BYTE)0x2f,(BYTE)0x00,(BYTE)0x04,(BYTE)0x41,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x01,(BYTE)0x74 },
    { 0x4ce0,(BYTE)0x70,(BYTE)0x00,(BYTE)0x10,(BYTE)0x30,(BYTE)0x10,(BYTE)0x00,(BYTE)0x4a,(BYTE)0x00,(BYTE)0x66,(BYTE)0x50,(BYTE)0x24,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xf0 },
    { 0x4cf0,(BYTE)0xb4,(BYTE)0xb8,(BYTE)0x04,(BYTE)0xba,(BYTE)0x64,(BYTE)0x44,(BYTE)0x43,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0xd6,(BYTE)0x13,(BYTE)0xf1,(BYTE)0x10,(BYTE)0x00 },
    { 0x4d00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xa9,(BYTE)0x08,(BYTE)0x39,(BYTE)0x00,(BYTE)0x06,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xa9,(BYTE)0x67,(BYTE)0x2a,(BYTE)0x3f,(BYTE)0x39 },
    { 0x4d10,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xa8,(BYTE)0x61,(BYTE)0x00,(BYTE)0x0a,(BYTE)0x28,(BYTE)0x54,(BYTE)0x8f,(BYTE)0x23,(BYTE)0xf8,(BYTE)0x04,(BYTE)0xba,(BYTE)0x00,(BYTE)0x00 },
    { 0x4d20,(BYTE)0x05,(BYTE)0xf0,(BYTE)0x06,(BYTE)0xb9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0xc8,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xf0,(BYTE)0x4a,(BYTE)0x40,(BYTE)0x67,(BYTE)0x08 },
    { 0x4d30,(BYTE)0x70,(BYTE)0x01,(BYTE)0x61,(BYTE)0x00,(BYTE)0x00,(BYTE)0x08,(BYTE)0x60,(BYTE)0x02,(BYTE)0x70,(BYTE)0x00,(BYTE)0x4e,(BYTE)0x75,(BYTE)0x48,(BYTE)0xe7,(BYTE)0x60,(BYTE)0xe0 },
    { 0x4d40,(BYTE)0x41,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0xd6,(BYTE)0x43,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x01,(BYTE)0x74,(BYTE)0x45,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00 },
    { 0x4d50,(BYTE)0x01,(BYTE)0x82,(BYTE)0x32,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xa8,(BYTE)0x74,(BYTE)0x00,(BYTE)0x08,(BYTE)0x30,(BYTE)0x00,(BYTE)0x07,(BYTE)0x20,(BYTE)0x00 },
    { 0x4d60,(BYTE)0x66,(BYTE)0x16,(BYTE)0xb2,(BYTE)0x30,(BYTE)0x20,(BYTE)0x00,(BYTE)0x66,(BYTE)0x08,(BYTE)0x13,(BYTE)0x80,(BYTE)0x20,(BYTE)0x00,(BYTE)0x15,(BYTE)0x80,(BYTE)0x20,(BYTE)0x00 },
    { 0x4d70,(BYTE)0x52,(BYTE)0x42,(BYTE)0xb4,(BYTE)0x7c,(BYTE)0x00,(BYTE)0x0e,(BYTE)0x6d,(BYTE)0xe2,(BYTE)0x4c,(BYTE)0xdf,(BYTE)0x07,(BYTE)0x06,(BYTE)0x4e,(BYTE)0x75,(BYTE)0x4a,(BYTE)0x39 },
    { 0x4d80,(BYTE)0x00,(BYTE)0x00,(BYTE)0x22,(BYTE)0x14,(BYTE)0x67,(BYTE)0x1a,(BYTE)0x33,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x44,(BYTE)0x00,(BYTE)0x00,(BYTE)0x01,(BYTE)0x34 },
    { 0x4d90,(BYTE)0x3f,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x44,(BYTE)0x3f,(BYTE)0x3c,(BYTE)0x00,(BYTE)0x10,(BYTE)0x61,(BYTE)0x00,(BYTE)0x08,(BYTE)0x1c,(BYTE)0x58,(BYTE)0x4f },
    { 0x4da0,(BYTE)0x23,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x00,(BYTE)0x22,(BYTE)0x16,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xa0,(BYTE)0x04,(BYTE)0xb9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00 },
    { 0x4db0,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xa0,(BYTE)0x0c,(BYTE)0x79,(BYTE)0x02,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x01,(BYTE)0x2e,(BYTE)0x63,(BYTE)0x4a,(BYTE)0x61,(BYTE)0x00 },
    { 0x4dc0,(BYTE)0x01,(BYTE)0x7e,(BYTE)0x4a,(BYTE)0x80,(BYTE)0x6a,(BYTE)0x20,(BYTE)0x30,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x12,(BYTE)0xb0,(BYTE)0x79,(BYTE)0x00,(BYTE)0x00 },
    { 0x4dd0,(BYTE)0x01,(BYTE)0x2e,(BYTE)0x64,(BYTE)0x08,(BYTE)0x33,(BYTE)0xc0,(BYTE)0x00,(BYTE)0x00,(BYTE)0x01,(BYTE)0x2e,(BYTE)0x60,(BYTE)0xe2,(BYTE)0x33,(BYTE)0xfc,(BYTE)0x02,(BYTE)0x00 },
    { 0x4de0,(BYTE)0x00,(BYTE)0x00,(BYTE)0x01,(BYTE)0x2e,(BYTE)0x60,(BYTE)0x22,(BYTE)0x23,(BYTE)0xc1,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xa0,(BYTE)0x41,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00 },
    { 0x4df0,(BYTE)0x22,(BYTE)0x16,(BYTE)0x72,(BYTE)0x03,(BYTE)0x61,(BYTE)0x00,(BYTE)0x01,(BYTE)0x24,(BYTE)0x42,(BYTE)0xb0,(BYTE)0x00,(BYTE)0x00,(BYTE)0x21,(BYTE)0xc8,(BYTE)0x04,(BYTE)0xb2 },
    { 0x4e00,(BYTE)0xd0,(BYTE)0x80,(BYTE)0xd1,(BYTE)0xc0,(BYTE)0x21,(BYTE)0xc8,(BYTE)0x04,(BYTE)0xb6,(BYTE)0x61,(BYTE)0x00,(BYTE)0x01,(BYTE)0xf6,(BYTE)0xd1,(BYTE)0xb9,(BYTE)0x00,(BYTE)0x00 },
    { 0x4e10,(BYTE)0x05,(BYTE)0xa0,(BYTE)0x23,(BYTE)0xf8,(BYTE)0x04,(BYTE)0xba,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xf0,(BYTE)0x4a,(BYTE)0x79,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0x96 },
    { 0x4e20,(BYTE)0x66,(BYTE)0x36,(BYTE)0x2f,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00,(BYTE)0x06,(BYTE)0x06,(BYTE)0x3f,(BYTE)0x3c,(BYTE)0x00,(BYTE)0x20,(BYTE)0x4e,(BYTE)0x41,(BYTE)0x5c,(BYTE)0x4f },
    { 0x4e30,(BYTE)0x20,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xa0,(BYTE)0xd0,(BYTE)0xbc,(BYTE)0x00,(BYTE)0x00,(BYTE)0x01,(BYTE)0x00,(BYTE)0x2f,(BYTE)0x00,(BYTE)0x20,(BYTE)0x7c },
    { 0x4e40,(BYTE)0x00,(BYTE)0x00,(BYTE)0x14,(BYTE)0xd4,(BYTE)0x61,(BYTE)0x00,(BYTE)0x15,(BYTE)0x0a,(BYTE)0x20,(BYTE)0x1f,(BYTE)0x3f,(BYTE)0x3c,(BYTE)0x00,(BYTE)0x00,(BYTE)0x2f,(BYTE)0x00 },
    { 0x4e50,(BYTE)0x3f,(BYTE)0x3c,(BYTE)0x00,(BYTE)0x31,(BYTE)0x4e,(BYTE)0x41,(BYTE)0x4a,(BYTE)0xfc,(BYTE)0x20,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xa0,(BYTE)0xd0,(BYTE)0xbc },
    { 0x4e60,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x1c,(BYTE)0x2f,(BYTE)0x00,(BYTE)0x20,(BYTE)0x7c,(BYTE)0x00,(BYTE)0x00,(BYTE)0x14,(BYTE)0xae,(BYTE)0x61,(BYTE)0x00,(BYTE)0x14,(BYTE)0xe2 },
    { 0x4e70,(BYTE)0x32,(BYTE)0x04,(BYTE)0x6a,(BYTE)0x04,(BYTE)0x12,(BYTE)0x07,(BYTE)0xea,(BYTE)0x09,(BYTE)0x41,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0xd6,(BYTE)0x74,(BYTE)0x00 },
    { 0x4e80,(BYTE)0x10,(BYTE)0x30,(BYTE)0x20,(BYTE)0x00,(BYTE)0x02,(BYTE)0x00,(BYTE)0x00,(BYTE)0x1f,(BYTE)0xb2,(BYTE)0x00,(BYTE)0x67,(BYTE)0x04,(BYTE)0x52,(BYTE)0x42,(BYTE)0x60,(BYTE)0xf0 },
    { 0x4e90,(BYTE)0x54,(BYTE)0x42,(BYTE)0x31,(BYTE)0xc2,(BYTE)0x04,(BYTE)0x46,(BYTE)0x2f,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0x9c,(BYTE)0x42,(BYTE)0x67,(BYTE)0x3f,(BYTE)0x3c },
    { 0x4ea0,(BYTE)0x00,(BYTE)0x4a,(BYTE)0x4e,(BYTE)0x41,(BYTE)0xde,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x0c,(BYTE)0x3f,(BYTE)0x38,(BYTE)0x04,(BYTE)0x46,(BYTE)0x3f,(BYTE)0x3c,(BYTE)0x00,(BYTE)0x0e },
    { 0x4eb0,(BYTE)0x4e,(BYTE)0x41,(BYTE)0x58,(BYTE)0x4f,(BYTE)0x2f,(BYTE)0x3c,(BYTE)0x00,(BYTE)0x00,(BYTE)0x14,(BYTE)0xac,(BYTE)0x3f,(BYTE)0x3c,(BYTE)0x00,(BYTE)0x3b,(BYTE)0x4e,(BYTE)0x41 },
    { 0x4ec0,(BYTE)0x5c,(BYTE)0x4f,(BYTE)0x1e,(BYTE)0x3c,(BYTE)0x00,(BYTE)0xe0,(BYTE)0x4e,(BYTE)0x75,(BYTE)0x5c,(BYTE)0x00,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d },
    { 0x4ed0,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d },
    { 0x4ee0,(BYTE)0x2d,(BYTE)0x0d,(BYTE)0x0a,(BYTE)0x42,(BYTE)0x4f,(BYTE)0x4f,(BYTE)0x54,(BYTE)0x4c,(BYTE)0x4f,(BYTE)0x41,(BYTE)0x44,(BYTE)0x45,(BYTE)0x44,(BYTE)0x0d,(BYTE)0x0a,(BYTE)0x00 },
    { 0x4ef0,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d },
    { 0x4f00,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x0d,(BYTE)0x0a,(BYTE)0x4e,(BYTE)0x4f,(BYTE)0x54,(BYTE)0x20,(BYTE)0x42,(BYTE)0x6f,(BYTE)0x6f },
    { 0x4f10,(BYTE)0x74,(BYTE)0x6c,(BYTE)0x6f,(BYTE)0x61,(BYTE)0x64,(BYTE)0x65,(BYTE)0x64,(BYTE)0x0d,(BYTE)0x0a,(BYTE)0x00,(BYTE)0x2f,(BYTE)0x08,(BYTE)0x22,(BYTE)0x48,(BYTE)0xd3,(BYTE)0xc0 },
    { 0x4f20,(BYTE)0x20,(BYTE)0x89,(BYTE)0x31,(BYTE)0x7c,(BYTE)0xff,(BYTE)0xff,(BYTE)0x00,(BYTE)0x04,(BYTE)0xd0,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x14,(BYTE)0x21,(BYTE)0x48,(BYTE)0xff,(BYTE)0xfc },
    { 0x4f30,(BYTE)0x20,(BYTE)0x49,(BYTE)0x51,(BYTE)0xc9,(BYTE)0xff,(BYTE)0xe8,(BYTE)0x91,(BYTE)0xc0,(BYTE)0x42,(BYTE)0x90,(BYTE)0x20,(BYTE)0x5f,(BYTE)0x4e,(BYTE)0x75,(BYTE)0x70,(BYTE)0x14 },
    { 0x4f40,(BYTE)0xd0,(BYTE)0x79,(BYTE)0x00,(BYTE)0x00,(BYTE)0x01,(BYTE)0x2e,(BYTE)0x22,(BYTE)0x00,(BYTE)0xe5,(BYTE)0x89,(BYTE)0xd2,(BYTE)0xb9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xa0 },
    { 0x4f50,(BYTE)0xb2,(BYTE)0xb9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0x98,(BYTE)0x63,(BYTE)0x02,(BYTE)0x70,(BYTE)0xff,(BYTE)0x4e,(BYTE)0x75,(BYTE)0x54,(BYTE)0x41,(BYTE)0x3f,(BYTE)0x01 },
    { 0x4f60,(BYTE)0x3f,(BYTE)0x00,(BYTE)0x20,(BYTE)0x78,(BYTE)0x04,(BYTE)0x04,(BYTE)0x4e,(BYTE)0x90,(BYTE)0x58,(BYTE)0x8f,(BYTE)0x4e,(BYTE)0x75,(BYTE)0x08,(BYTE)0x39,(BYTE)0x00,(BYTE)0x04 },
    { 0x4f70,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xa9,(BYTE)0x66,(BYTE)0x7a,(BYTE)0x72,(BYTE)0x16,(BYTE)0x30,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xa8,(BYTE)0x08,(BYTE)0x00 },
    { 0x4f80,(BYTE)0x00,(BYTE)0x03,(BYTE)0x66,(BYTE)0x0e,(BYTE)0x02,(BYTE)0x40,(BYTE)0x00,(BYTE)0x07,(BYTE)0x01,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xba,(BYTE)0x66,(BYTE)0x02 },
    { 0x4f90,(BYTE)0x72,(BYTE)0x00,(BYTE)0x41,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xd0,(BYTE)0x48,(BYTE)0xe7,(BYTE)0x40,(BYTE)0x80,(BYTE)0x2f,(BYTE)0x08,(BYTE)0x3f,(BYTE)0x01 },
    { 0x4fa0,(BYTE)0x3f,(BYTE)0x00,(BYTE)0x61,(BYTE)0x00,(BYTE)0x07,(BYTE)0xb4,(BYTE)0x50,(BYTE)0x8f,(BYTE)0x4c,(BYTE)0xdf,(BYTE)0x01,(BYTE)0x02,(BYTE)0x4a,(BYTE)0x40,(BYTE)0x67,(BYTE)0x04 },
    { 0x4fb0,(BYTE)0x70,(BYTE)0xff,(BYTE)0x4e,(BYTE)0x75,(BYTE)0x0c,(BYTE)0x41,(BYTE)0x00,(BYTE)0x04,(BYTE)0x6e,(BYTE)0x08,(BYTE)0x12,(BYTE)0x10,(BYTE)0x02,(BYTE)0x01,(BYTE)0x00,(BYTE)0x7f },
    { 0x4fc0,(BYTE)0x60,(BYTE)0x04,(BYTE)0x12,(BYTE)0x28,(BYTE)0x00,(BYTE)0x0c,(BYTE)0x0c,(BYTE)0x01,(BYTE)0x00,(BYTE)0x28,(BYTE)0x66,(BYTE)0x04,(BYTE)0x70,(BYTE)0xf2,(BYTE)0x60,(BYTE)0x2e },
    { 0x4fd0,(BYTE)0x0c,(BYTE)0x01,(BYTE)0x00,(BYTE)0x27,(BYTE)0x66,(BYTE)0x04,(BYTE)0x70,(BYTE)0xf3,(BYTE)0x60,(BYTE)0x24,(BYTE)0x0c,(BYTE)0x01,(BYTE)0x00,(BYTE)0x04,(BYTE)0x66,(BYTE)0x04 },
    { 0x4fe0,(BYTE)0x70,(BYTE)0xfe,(BYTE)0x60,(BYTE)0x1a,(BYTE)0x08,(BYTE)0x2f,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0x66,(BYTE)0x10,(BYTE)0x70,(BYTE)0xf5,(BYTE)0x60,(BYTE)0x0e },
    { 0x4ff0,(BYTE)0x08,(BYTE)0x00,(BYTE)0x00,(BYTE)0x06,(BYTE)0x67,(BYTE)0xea,(BYTE)0x08,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0x66,(BYTE)0xe8,(BYTE)0x70,(BYTE)0xf6,(BYTE)0x4e,(BYTE)0x75 },
    { 0x5000,(BYTE)0x26,(BYTE)0x78,(BYTE)0x04,(BYTE)0xf2,(BYTE)0xb7,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x80,(BYTE)0x00,(BYTE)0x00,(BYTE)0x6d,(BYTE)0x00,(BYTE)0x00,(BYTE)0x64,(BYTE)0x41,(BYTE)0xfa },
    { 0x5010,(BYTE)0x02,(BYTE)0x1a,(BYTE)0x22,(BYTE)0x18,(BYTE)0x67,(BYTE)0x00,(BYTE)0x00,(BYTE)0x60,(BYTE)0x24,(BYTE)0x58,(BYTE)0xb2,(BYTE)0xab,(BYTE)0x00,(BYTE)0x18,(BYTE)0x66,(BYTE)0xf2 },
    { 0x5020,(BYTE)0x30,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x10,(BYTE)0xc0,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x42,(BYTE)0x22,(BYTE)0x00,(BYTE)0xd2,(BYTE)0xb9,(BYTE)0x00,(BYTE)0x00 },
    { 0x5030,(BYTE)0x05,(BYTE)0xa0,(BYTE)0xb2,(BYTE)0xb9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0x98,(BYTE)0x6e,(BYTE)0x72,(BYTE)0x20,(BYTE)0x7c,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x02 },
    { 0x5040,(BYTE)0xd1,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xa0,(BYTE)0x2f,(BYTE)0x08,(BYTE)0x32,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x10,(BYTE)0x53,(BYTE)0x41 },
    { 0x5050,(BYTE)0x43,(BYTE)0xe8,(BYTE)0x00,(BYTE)0x42,(BYTE)0x20,(BYTE)0x89,(BYTE)0x31,(BYTE)0x7c,(BYTE)0x00,(BYTE)0x04,(BYTE)0xff,(BYTE)0xfe,(BYTE)0x20,(BYTE)0x49,(BYTE)0x51,(BYTE)0xc9 },
    { 0x5060,(BYTE)0xff,(BYTE)0xf0,(BYTE)0x90,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x42,(BYTE)0x20,(BYTE)0xaa,(BYTE)0x00,(BYTE)0x10,(BYTE)0x25,(BYTE)0x5f,(BYTE)0x00,(BYTE)0x10,(BYTE)0x4e,(BYTE)0x75 },
    { 0x5070,(BYTE)0x41,(BYTE)0xfa,(BYTE)0x00,(BYTE)0x7a,(BYTE)0x60,(BYTE)0x04,(BYTE)0x41,(BYTE)0xfa,(BYTE)0x00,(BYTE)0xfd,(BYTE)0x0c,(BYTE)0x6b,(BYTE)0x0c,(BYTE)0xa1,(BYTE)0x00,(BYTE)0x1e },
    { 0x5080,(BYTE)0x64,(BYTE)0x00,(BYTE)0x00,(BYTE)0x2a,(BYTE)0x2f,(BYTE)0x08,(BYTE)0x3f,(BYTE)0x3c,(BYTE)0x00,(BYTE)0x09,(BYTE)0x4e,(BYTE)0x41,(BYTE)0x5c,(BYTE)0x8f,(BYTE)0x48,(BYTE)0x7a },
    { 0x5090,(BYTE)0x00,(BYTE)0x20,(BYTE)0x3f,(BYTE)0x3c,(BYTE)0x00,(BYTE)0x09,(BYTE)0x4e,(BYTE)0x41,(BYTE)0x5c,(BYTE)0x8f,(BYTE)0x3f,(BYTE)0x3c,(BYTE)0x00,(BYTE)0x02,(BYTE)0x3f,(BYTE)0x3c },
    { 0x50a0,(BYTE)0x00,(BYTE)0x02,(BYTE)0x4e,(BYTE)0x4d,(BYTE)0x58,(BYTE)0x8f,(BYTE)0xb0,(BYTE)0x7c,(BYTE)0x00,(BYTE)0x0d,(BYTE)0x66,(BYTE)0xee,(BYTE)0x70,(BYTE)0x00,(BYTE)0x4e,(BYTE)0x75 },
    { 0x50b0,(BYTE)0x48,(BYTE)0x61,(BYTE)0x72,(BYTE)0x64,(BYTE)0x20,(BYTE)0x64,(BYTE)0x69,(BYTE)0x73,(BYTE)0x6b,(BYTE)0x20,(BYTE)0x64,(BYTE)0x72,(BYTE)0x69,(BYTE)0x76,(BYTE)0x65,(BYTE)0x72 },
    { 0x50c0,(BYTE)0x20,(BYTE)0x6e,(BYTE)0x6f,(BYTE)0x74,(BYTE)0x20,(BYTE)0x6c,(BYTE)0x6f,(BYTE)0x61,(BYTE)0x64,(BYTE)0x65,(BYTE)0x64,(BYTE)0x3b,(BYTE)0x20,(BYTE)0x68,(BYTE)0x69,(BYTE)0x74 },
    { 0x50d0,(BYTE)0x20,(BYTE)0x52,(BYTE)0x45,(BYTE)0x54,(BYTE)0x55,(BYTE)0x52,(BYTE)0x4e,(BYTE)0x0d,(BYTE)0x0a,(BYTE)0x6b,(BYTE)0x65,(BYTE)0x79,(BYTE)0x20,(BYTE)0x74,(BYTE)0x6f,(BYTE)0x20 },
    { 0x50e0,(BYTE)0x63,(BYTE)0x6f,(BYTE)0x6e,(BYTE)0x74,(BYTE)0x69,(BYTE)0x6e,(BYTE)0x75,(BYTE)0x65,(BYTE)0x3a,(BYTE)0x0d,(BYTE)0x0a,(BYTE)0x00,(BYTE)0x2a,(BYTE)0x2a,(BYTE)0x2a,(BYTE)0x20 },
    { 0x50f0,(BYTE)0x57,(BYTE)0x41,(BYTE)0x52,(BYTE)0x4e,(BYTE)0x49,(BYTE)0x4e,(BYTE)0x47,(BYTE)0x20,(BYTE)0x2a,(BYTE)0x2a,(BYTE)0x2a,(BYTE)0x0d,(BYTE)0x0a,(BYTE)0x07,(BYTE)0x54,(BYTE)0x68 },
    { 0x5100,(BYTE)0x69,(BYTE)0x73,(BYTE)0x20,(BYTE)0x68,(BYTE)0x61,(BYTE)0x72,(BYTE)0x64,(BYTE)0x20,(BYTE)0x64,(BYTE)0x69,(BYTE)0x73,(BYTE)0x6b,(BYTE)0x20,(BYTE)0x64,(BYTE)0x72,(BYTE)0x69 },
    { 0x5110,(BYTE)0x76,(BYTE)0x65,(BYTE)0x72,(BYTE)0x20,(BYTE)0x6d,(BYTE)0x61,(BYTE)0x79,(BYTE)0x20,(BYTE)0x6e,(BYTE)0x6f,(BYTE)0x74,(BYTE)0x20,(BYTE)0x77,(BYTE)0x6f,(BYTE)0x72,(BYTE)0x6b },
    { 0x5120,(BYTE)0x20,(BYTE)0x77,(BYTE)0x69,(BYTE)0x74,(BYTE)0x68,(BYTE)0x0d,(BYTE)0x0a,(BYTE)0x07,(BYTE)0x61,(BYTE)0x20,(BYTE)0x64,(BYTE)0x69,(BYTE)0x73,(BYTE)0x6b,(BYTE)0x2d,(BYTE)0x62 },
    { 0x5130,(BYTE)0x61,(BYTE)0x73,(BYTE)0x65,(BYTE)0x64,(BYTE)0x20,(BYTE)0x76,(BYTE)0x65,(BYTE)0x72,(BYTE)0x73,(BYTE)0x69,(BYTE)0x6f,(BYTE)0x6e,(BYTE)0x20,(BYTE)0x6f,(BYTE)0x66,(BYTE)0x20 },
    { 0x5140,(BYTE)0x54,(BYTE)0x4f,(BYTE)0x53,(BYTE)0x3b,(BYTE)0x20,(BYTE)0x66,(BYTE)0x69,(BYTE)0x6c,(BYTE)0x65,(BYTE)0x73,(BYTE)0x20,(BYTE)0x6f,(BYTE)0x6e,(BYTE)0x0d,(BYTE)0x0a,(BYTE)0x07 },
    { 0x5150,(BYTE)0x79,(BYTE)0x6f,(BYTE)0x75,(BYTE)0x72,(BYTE)0x20,(BYTE)0x68,(BYTE)0x61,(BYTE)0x72,(BYTE)0x64,(BYTE)0x20,(BYTE)0x64,(BYTE)0x69,(BYTE)0x73,(BYTE)0x6b,(BYTE)0x20,(BYTE)0x6d },
    { 0x5160,(BYTE)0x61,(BYTE)0x79,(BYTE)0x20,(BYTE)0x62,(BYTE)0x65,(BYTE)0x20,(BYTE)0x64,(BYTE)0x61,(BYTE)0x6d,(BYTE)0x61,(BYTE)0x67,(BYTE)0x65,(BYTE)0x64,(BYTE)0x2e,(BYTE)0x0d,(BYTE)0x0a },
    { 0x5170,(BYTE)0x07,(BYTE)0x0d,(BYTE)0x0a,(BYTE)0x07,(BYTE)0x00,(BYTE)0x2a,(BYTE)0x2a,(BYTE)0x2a,(BYTE)0x20,(BYTE)0x57,(BYTE)0x41,(BYTE)0x52,(BYTE)0x4e,(BYTE)0x49,(BYTE)0x4e,(BYTE)0x47 },
    { 0x5180,(BYTE)0x20,(BYTE)0x2a,(BYTE)0x2a,(BYTE)0x2a,(BYTE)0x0d,(BYTE)0x0a,(BYTE)0x07,(BYTE)0x59,(BYTE)0x6f,(BYTE)0x75,(BYTE)0x20,(BYTE)0x61,(BYTE)0x72,(BYTE)0x65,(BYTE)0x20,(BYTE)0x75 },
    { 0x5190,(BYTE)0x73,(BYTE)0x69,(BYTE)0x6e,(BYTE)0x67,(BYTE)0x20,(BYTE)0x61,(BYTE)0x6e,(BYTE)0x20,(BYTE)0x75,(BYTE)0x6e,(BYTE)0x6f,(BYTE)0x66,(BYTE)0x66,(BYTE)0x69,(BYTE)0x63,(BYTE)0x69 },
    { 0x51a0,(BYTE)0x61,(BYTE)0x6c,(BYTE)0x20,(BYTE)0x52,(BYTE)0x4f,(BYTE)0x4d,(BYTE)0x20,(BYTE)0x72,(BYTE)0x65,(BYTE)0x6c,(BYTE)0x65,(BYTE)0x61,(BYTE)0x73,(BYTE)0x65,(BYTE)0x0d,(BYTE)0x0a },
    { 0x51b0,(BYTE)0x07,(BYTE)0x6f,(BYTE)0x66,(BYTE)0x20,(BYTE)0x74,(BYTE)0x68,(BYTE)0x65,(BYTE)0x20,(BYTE)0x6f,(BYTE)0x70,(BYTE)0x65,(BYTE)0x72,(BYTE)0x61,(BYTE)0x74,(BYTE)0x69,(BYTE)0x6e },
    { 0x51c0,(BYTE)0x67,(BYTE)0x20,(BYTE)0x73,(BYTE)0x79,(BYTE)0x73,(BYTE)0x74,(BYTE)0x65,(BYTE)0x6d,(BYTE)0x2e,(BYTE)0x20,(BYTE)0x20,(BYTE)0x54,(BYTE)0x68,(BYTE)0x69,(BYTE)0x73,(BYTE)0x20 },
    { 0x51d0,(BYTE)0x64,(BYTE)0x72,(BYTE)0x69,(BYTE)0x76,(BYTE)0x65,(BYTE)0x72,(BYTE)0x0d,(BYTE)0x0a,(BYTE)0x07,(BYTE)0x6d,(BYTE)0x61,(BYTE)0x79,(BYTE)0x20,(BYTE)0x6e,(BYTE)0x6f,(BYTE)0x74 },
    { 0x51e0,(BYTE)0x20,(BYTE)0x77,(BYTE)0x6f,(BYTE)0x72,(BYTE)0x6b,(BYTE)0x20,(BYTE)0x63,(BYTE)0x6f,(BYTE)0x72,(BYTE)0x72,(BYTE)0x65,(BYTE)0x63,(BYTE)0x74,(BYTE)0x6c,(BYTE)0x79,(BYTE)0x20 },
    { 0x51f0,(BYTE)0x77,(BYTE)0x69,(BYTE)0x74,(BYTE)0x68,(BYTE)0x20,(BYTE)0x69,(BYTE)0x74,(BYTE)0x2e,(BYTE)0x20,(BYTE)0x20,(BYTE)0x46,(BYTE)0x69,(BYTE)0x6c,(BYTE)0x65,(BYTE)0x73,(BYTE)0x0d },
    { 0x5200,(BYTE)0x0a,(BYTE)0x07,(BYTE)0x6f,(BYTE)0x6e,(BYTE)0x20,(BYTE)0x79,(BYTE)0x6f,(BYTE)0x75,(BYTE)0x72,(BYTE)0x20,(BYTE)0x68,(BYTE)0x61,(BYTE)0x72,(BYTE)0x64,(BYTE)0x20,(BYTE)0x64 },
    { 0x5210,(BYTE)0x69,(BYTE)0x73,(BYTE)0x6b,(BYTE)0x20,(BYTE)0x6d,(BYTE)0x61,(BYTE)0x79,(BYTE)0x20,(BYTE)0x62,(BYTE)0x65,(BYTE)0x20,(BYTE)0x64,(BYTE)0x61,(BYTE)0x6d,(BYTE)0x61,(BYTE)0x67 },
    { 0x5220,(BYTE)0x65,(BYTE)0x64,(BYTE)0x2e,(BYTE)0x0d,(BYTE)0x0a,(BYTE)0x07,(BYTE)0x0d,(BYTE)0x0a,(BYTE)0x07,(BYTE)0x00,(BYTE)0x11,(BYTE)0x20,(BYTE)0x19,(BYTE)0x85,(BYTE)0x00,(BYTE)0x00 },
    { 0x5230,(BYTE)0x56,(BYTE)0xfa,(BYTE)0x02,(BYTE)0x06,(BYTE)0x19,(BYTE)0x86,(BYTE)0x00,(BYTE)0x00,(BYTE)0x56,(BYTE)0xfa,(BYTE)0x04,(BYTE)0x24,(BYTE)0x19,(BYTE)0x86,(BYTE)0x00,(BYTE)0x00 },
    { 0x5240,(BYTE)0x56,(BYTE)0xfa,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x20,(BYTE)0x3c,(BYTE)0x00,(BYTE)0x00 },
    { 0x5250,(BYTE)0x02,(BYTE)0x58,(BYTE)0x60,(BYTE)0x02,(BYTE)0x70,(BYTE)0x14,(BYTE)0x4a,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00,(BYTE)0x18,(BYTE)0x2f,(BYTE)0x67,(BYTE)0x10,(BYTE)0x2f,(BYTE)0x00 },
    { 0x5260,(BYTE)0x70,(BYTE)0x02,(BYTE)0xd0,(BYTE)0xb8,(BYTE)0x04,(BYTE)0xba,(BYTE)0xb0,(BYTE)0xb8,(BYTE)0x04,(BYTE)0xba,(BYTE)0x6c,(BYTE)0xfa,(BYTE)0x20,(BYTE)0x1f,(BYTE)0xd0,(BYTE)0xb8 },
    { 0x5270,(BYTE)0x04,(BYTE)0xba,(BYTE)0xb0,(BYTE)0xb8,(BYTE)0x04,(BYTE)0xba,(BYTE)0x65,(BYTE)0x0c,(BYTE)0x08,(BYTE)0x38,(BYTE)0x00,(BYTE)0x05,(BYTE)0xfa,(BYTE)0x01,(BYTE)0x66,(BYTE)0xf2 },
    { 0x5280,(BYTE)0x70,(BYTE)0x00,(BYTE)0x4e,(BYTE)0x75,(BYTE)0x70,(BYTE)0xff,(BYTE)0x4e,(BYTE)0x75,(BYTE)0x61,(BYTE)0xc2,(BYTE)0x6b,(BYTE)0x28,(BYTE)0x4a,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00 },
    { 0x5290,(BYTE)0x06,(BYTE)0x0a,(BYTE)0x67,(BYTE)0x14,(BYTE)0x40,(BYTE)0xe7,(BYTE)0x00,(BYTE)0x7c,(BYTE)0x07,(BYTE)0x00,(BYTE)0x4e,(BYTE)0x7a,(BYTE)0x00,(BYTE)0x02,(BYTE)0x00,(BYTE)0x40 },
    { 0x52a0,(BYTE)0x08,(BYTE)0x00,(BYTE)0x4e,(BYTE)0x7b,(BYTE)0x00,(BYTE)0x02,(BYTE)0x46,(BYTE)0xdf,(BYTE)0x31,(BYTE)0xc1,(BYTE)0x86,(BYTE)0x06,(BYTE)0x30,(BYTE)0x38,(BYTE)0x86,(BYTE)0x04 },
    { 0x52b0,(BYTE)0xc0,(BYTE)0x7c,(BYTE)0x00,(BYTE)0xff,(BYTE)0x23,(BYTE)0xf8,(BYTE)0x04,(BYTE)0xba,(BYTE)0x00,(BYTE)0x00,(BYTE)0x18,(BYTE)0x2a,(BYTE)0x54,(BYTE)0xb9,(BYTE)0x00,(BYTE)0x00 },
    { 0x52c0,(BYTE)0x18,(BYTE)0x2a,(BYTE)0x4e,(BYTE)0x75,(BYTE)0x31,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x80,(BYTE)0x86,(BYTE)0x06,(BYTE)0x42,(BYTE)0x78,(BYTE)0x04,(BYTE)0x3e,(BYTE)0x4e,(BYTE)0x75 },
    { 0x52d0,(BYTE)0x2f,(BYTE)0x00,(BYTE)0x20,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00,(BYTE)0x18,(BYTE)0x2a,(BYTE)0xb0,(BYTE)0xb8,(BYTE)0x04,(BYTE)0xba,(BYTE)0x64,(BYTE)0xfa,(BYTE)0x20,(BYTE)0x1f },
    { 0x52e0,(BYTE)0x4e,(BYTE)0x75,(BYTE)0x50,(BYTE)0xf8,(BYTE)0x04,(BYTE)0x3e,(BYTE)0x61,(BYTE)0xe8,(BYTE)0x22,(BYTE)0x7c,(BYTE)0xff,(BYTE)0xff,(BYTE)0x86,(BYTE)0x04,(BYTE)0x02,(BYTE)0x40 },
    { 0x52f0,(BYTE)0x00,(BYTE)0x07,(BYTE)0x72,(BYTE)0x00,(BYTE)0x61,(BYTE)0x00,(BYTE)0x00,(BYTE)0xd6,(BYTE)0x60,(BYTE)0xca,(BYTE)0x50,(BYTE)0xf8,(BYTE)0x04,(BYTE)0x3e,(BYTE)0x3f,(BYTE)0x03 },
    { 0x5300,(BYTE)0x61,(BYTE)0xce,(BYTE)0x22,(BYTE)0x7c,(BYTE)0xff,(BYTE)0xff,(BYTE)0x86,(BYTE)0x04,(BYTE)0x61,(BYTE)0x00,(BYTE)0x01,(BYTE)0x0a,(BYTE)0x33,(BYTE)0x7c,(BYTE)0x01,(BYTE)0x90 },
    { 0x5310,(BYTE)0x00,(BYTE)0x02,(BYTE)0x61,(BYTE)0x00,(BYTE)0x01,(BYTE)0x32,(BYTE)0x33,(BYTE)0x7c,(BYTE)0x00,(BYTE)0x90,(BYTE)0x00,(BYTE)0x02,(BYTE)0x61,(BYTE)0x00,(BYTE)0x01,(BYTE)0x28 },
    { 0x5320,(BYTE)0x61,(BYTE)0x00,(BYTE)0x01,(BYTE)0x0a,(BYTE)0x02,(BYTE)0x40,(BYTE)0x00,(BYTE)0x07,(BYTE)0x0c,(BYTE)0x81,(BYTE)0x00,(BYTE)0x00,(BYTE)0x02,(BYTE)0x00,(BYTE)0x65,(BYTE)0x04 },
    { 0x5330,(BYTE)0x72,(BYTE)0x00,(BYTE)0x60,(BYTE)0x26,(BYTE)0x36,(BYTE)0x01,(BYTE)0x0c,(BYTE)0x41,(BYTE)0x00,(BYTE)0x10,(BYTE)0x65,(BYTE)0x0a,(BYTE)0x02,(BYTE)0x41,(BYTE)0x00,(BYTE)0x0f },
    { 0x5340,(BYTE)0x67,(BYTE)0x18,(BYTE)0x72,(BYTE)0x01,(BYTE)0x60,(BYTE)0x14,(BYTE)0x72,(BYTE)0x10,(BYTE)0x82,(BYTE)0xc3,(BYTE)0x53,(BYTE)0x41,(BYTE)0x48,(BYTE)0x41,(BYTE)0x4a,(BYTE)0x41 },
    { 0x5350,(BYTE)0x66,(BYTE)0x04,(BYTE)0x48,(BYTE)0x41,(BYTE)0x60,(BYTE)0x04,(BYTE)0x48,(BYTE)0x41,(BYTE)0x52,(BYTE)0x41,(BYTE)0x41,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x1d,(BYTE)0x18 },
    { 0x5360,(BYTE)0x48,(BYTE)0xe7,(BYTE)0xe0,(BYTE)0x80,(BYTE)0x72,(BYTE)0x00,(BYTE)0x61,(BYTE)0x00,(BYTE)0xff,(BYTE)0x68,(BYTE)0x61,(BYTE)0x00,(BYTE)0x00,(BYTE)0x60,(BYTE)0x4a,(BYTE)0x40 },
    { 0x5370,(BYTE)0x66,(BYTE)0x14,(BYTE)0x4c,(BYTE)0xdf,(BYTE)0x01,(BYTE)0x07,(BYTE)0x51,(BYTE)0xc9,(BYTE)0x00,(BYTE)0x06,(BYTE)0x70,(BYTE)0x00,(BYTE)0x60,(BYTE)0x0c,(BYTE)0x70,(BYTE)0xff },
    { 0x5380,(BYTE)0x11,(BYTE)0x43,(BYTE)0x00,(BYTE)0x04,(BYTE)0x60,(BYTE)0xda,(BYTE)0xde,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x10,(BYTE)0x36,(BYTE)0x1f,(BYTE)0x60,(BYTE)0x00,(BYTE)0xff,(BYTE)0x36 },
    { 0x5390,(BYTE)0x50,(BYTE)0xf8,(BYTE)0x04,(BYTE)0x3e,(BYTE)0x61,(BYTE)0x00,(BYTE)0xff,(BYTE)0x3a,(BYTE)0x22,(BYTE)0x7c,(BYTE)0xff,(BYTE)0xff,(BYTE)0x86,(BYTE)0x04,(BYTE)0x61,(BYTE)0x00 },
    { 0x53a0,(BYTE)0x00,(BYTE)0x74,(BYTE)0x33,(BYTE)0x7c,(BYTE)0x00,(BYTE)0x90,(BYTE)0x00,(BYTE)0x02,(BYTE)0x61,(BYTE)0x00,(BYTE)0x00,(BYTE)0x9c,(BYTE)0x33,(BYTE)0x7c,(BYTE)0x01,(BYTE)0x90 },
    { 0x53b0,(BYTE)0x00,(BYTE)0x02,(BYTE)0x61,(BYTE)0x00,(BYTE)0x00,(BYTE)0x92,(BYTE)0x61,(BYTE)0x00,(BYTE)0x00,(BYTE)0x74,(BYTE)0x02,(BYTE)0x40,(BYTE)0x00,(BYTE)0x07,(BYTE)0x22,(BYTE)0x3c },
    { 0x53c0,(BYTE)0x00,(BYTE)0x00,(BYTE)0x01,(BYTE)0x00,(BYTE)0x61,(BYTE)0x00,(BYTE)0x00,(BYTE)0x06,(BYTE)0x60,(BYTE)0x00,(BYTE)0xfe,(BYTE)0xfa,(BYTE)0x12,(BYTE)0x3c,(BYTE)0x00,(BYTE)0x88 },
    { 0x53d0,(BYTE)0x33,(BYTE)0x41,(BYTE)0x00,(BYTE)0x02,(BYTE)0x12,(BYTE)0x3c,(BYTE)0x00,(BYTE)0x8a,(BYTE)0x45,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x1d,(BYTE)0x18,(BYTE)0x4a,(BYTE)0x40 },
    { 0x53e0,(BYTE)0x6b,(BYTE)0x04,(BYTE)0xeb,(BYTE)0x08,(BYTE)0x81,(BYTE)0x12,(BYTE)0x55,(BYTE)0x42,(BYTE)0x48,(BYTE)0x41,(BYTE)0x12,(BYTE)0x1a,(BYTE)0x48,(BYTE)0x41,(BYTE)0x22,(BYTE)0x81 },
    { 0x53f0,(BYTE)0x61,(BYTE)0x00,(BYTE)0xfe,(BYTE)0x62,(BYTE)0x6b,(BYTE)0x1c,(BYTE)0x51,(BYTE)0xca,(BYTE)0xff,(BYTE)0xf0,(BYTE)0x33,(BYTE)0x41,(BYTE)0x00,(BYTE)0x02,(BYTE)0x12,(BYTE)0x3c },
    { 0x5400,(BYTE)0x00,(BYTE)0x00,(BYTE)0x48,(BYTE)0x41,(BYTE)0x12,(BYTE)0x12,(BYTE)0x48,(BYTE)0x41,(BYTE)0x22,(BYTE)0x81,(BYTE)0x12,(BYTE)0x3c,(BYTE)0x00,(BYTE)0x8a,(BYTE)0x61,(BYTE)0x00 },
    { 0x5410,(BYTE)0xfe,(BYTE)0x78,(BYTE)0x4e,(BYTE)0x75,(BYTE)0x2f,(BYTE)0x08,(BYTE)0x11,(BYTE)0xef,(BYTE)0x00,(BYTE)0x03,(BYTE)0x86,(BYTE)0x0d,(BYTE)0x11,(BYTE)0xef,(BYTE)0x00,(BYTE)0x02 },
    { 0x5420,(BYTE)0x86,(BYTE)0x0b,(BYTE)0x11,(BYTE)0xef,(BYTE)0x00,(BYTE)0x01,(BYTE)0x86,(BYTE)0x09,(BYTE)0x58,(BYTE)0x8f,(BYTE)0x4e,(BYTE)0x75,(BYTE)0x0c,(BYTE)0x81,(BYTE)0x00,(BYTE)0x00 },
    { 0x5430,(BYTE)0x02,(BYTE)0x00,(BYTE)0x62,(BYTE)0x06,(BYTE)0x32,(BYTE)0xbc,(BYTE)0x00,(BYTE)0x01,(BYTE)0x60,(BYTE)0x0a,(BYTE)0x2f,(BYTE)0x01,(BYTE)0xe0,(BYTE)0x89,(BYTE)0xe2,(BYTE)0x89 },
    { 0x5440,(BYTE)0x32,(BYTE)0x81,(BYTE)0x22,(BYTE)0x1f,(BYTE)0x4e,(BYTE)0x75,(BYTE)0x4a,(BYTE)0x38,(BYTE)0xfa,(BYTE)0x01,(BYTE)0x4a,(BYTE)0x38,(BYTE)0xfa,(BYTE)0x01,(BYTE)0x4a,(BYTE)0x38 },
    { 0x5450,(BYTE)0xfa,(BYTE)0x01,(BYTE)0x4a,(BYTE)0x38,(BYTE)0xfa,(BYTE)0x01,(BYTE)0x4e,(BYTE)0x75,(BYTE)0x20,(BYTE)0x3c,(BYTE)0x00,(BYTE)0x00,(BYTE)0x07,(BYTE)0xd0,(BYTE)0xd0,(BYTE)0xb8 },
    { 0x5460,(BYTE)0x04,(BYTE)0xba,(BYTE)0x08,(BYTE)0x38,(BYTE)0x00,(BYTE)0x05,(BYTE)0xfa,(BYTE)0x01,(BYTE)0x67,(BYTE)0x0a,(BYTE)0xb0,(BYTE)0xb8,(BYTE)0x04,(BYTE)0xba,(BYTE)0x62,(BYTE)0xf2 },
    { 0x5470,(BYTE)0x70,(BYTE)0xff,(BYTE)0x60,(BYTE)0x16,(BYTE)0x70,(BYTE)0x00,(BYTE)0x10,(BYTE)0x39,(BYTE)0xff,(BYTE)0xf0,(BYTE)0x00,(BYTE)0x1d,(BYTE)0x08,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00 },
    { 0x5480,(BYTE)0x66,(BYTE)0x08,(BYTE)0x08,(BYTE)0x00,(BYTE)0x00,(BYTE)0x03,(BYTE)0x66,(BYTE)0x02,(BYTE)0x70,(BYTE)0x00,(BYTE)0x4e,(BYTE)0x75,(BYTE)0x61,(BYTE)0x00,(BYTE)0x00,(BYTE)0xa2 },
    { 0x5490,(BYTE)0x20,(BYTE)0x6f,(BYTE)0x00,(BYTE)0x0e,(BYTE)0x13,(BYTE)0xef,(BYTE)0x00,(BYTE)0x0d,(BYTE)0xff,(BYTE)0xf0,(BYTE)0x00,(BYTE)0x09,(BYTE)0x32,(BYTE)0x2f,(BYTE)0x00,(BYTE)0x0c },
    { 0x54a0,(BYTE)0x53,(BYTE)0x41,(BYTE)0x4a,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00,(BYTE)0x06,(BYTE)0x0b,(BYTE)0x67,(BYTE)0x06,(BYTE)0x70,(BYTE)0x00,(BYTE)0x61,(BYTE)0x00,(BYTE)0x01,(BYTE)0xa6 },
    { 0x54b0,(BYTE)0x13,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x00,(BYTE)0xff,(BYTE)0xf0,(BYTE)0x00,(BYTE)0x39,(BYTE)0x13,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x20,(BYTE)0xff,(BYTE)0xf0,(BYTE)0x00,(BYTE)0x1d },
    { 0x54c0,(BYTE)0x61,(BYTE)0x96,(BYTE)0x4a,(BYTE)0x40,(BYTE)0x6b,(BYTE)0x10,(BYTE)0x08,(BYTE)0x00,(BYTE)0x00,(BYTE)0x03,(BYTE)0x67,(BYTE)0x0a,(BYTE)0x61,(BYTE)0x00,(BYTE)0x01,(BYTE)0x0e },
    { 0x54d0,(BYTE)0x51,(BYTE)0xc9,(BYTE)0xff,(BYTE)0xee,(BYTE)0x70,(BYTE)0x00,(BYTE)0x4e,(BYTE)0x75,(BYTE)0x61,(BYTE)0x00,(BYTE)0x00,(BYTE)0x56,(BYTE)0x20,(BYTE)0x6f,(BYTE)0x00,(BYTE)0x0e },
    { 0x54e0,(BYTE)0x13,(BYTE)0xef,(BYTE)0x00,(BYTE)0x0d,(BYTE)0xff,(BYTE)0xf0,(BYTE)0x00,(BYTE)0x09,(BYTE)0x32,(BYTE)0x2f,(BYTE)0x00,(BYTE)0x0c,(BYTE)0x53,(BYTE)0x41,(BYTE)0x4a,(BYTE)0x39 },
    { 0x54f0,(BYTE)0x00,(BYTE)0x00,(BYTE)0x06,(BYTE)0x0b,(BYTE)0x67,(BYTE)0x06,(BYTE)0x70,(BYTE)0x01,(BYTE)0x61,(BYTE)0x00,(BYTE)0x01,(BYTE)0x5a,(BYTE)0x13,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x00 },
    { 0x5500,(BYTE)0xff,(BYTE)0xf0,(BYTE)0x00,(BYTE)0x39,(BYTE)0x13,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x30,(BYTE)0xff,(BYTE)0xf0,(BYTE)0x00,(BYTE)0x1d,(BYTE)0x08,(BYTE)0x39,(BYTE)0x00,(BYTE)0x03 },
    { 0x5510,(BYTE)0xff,(BYTE)0xf0,(BYTE)0x00,(BYTE)0x39,(BYTE)0x67,(BYTE)0xf6,(BYTE)0x61,(BYTE)0x00,(BYTE)0x01,(BYTE)0x00,(BYTE)0x61,(BYTE)0x00,(BYTE)0xff,(BYTE)0x3c,(BYTE)0x4a,(BYTE)0x40 },
    { 0x5520,(BYTE)0x6b,(BYTE)0x0c,(BYTE)0x08,(BYTE)0x00,(BYTE)0x00,(BYTE)0x03,(BYTE)0x67,(BYTE)0x06,(BYTE)0x51,(BYTE)0xc9,(BYTE)0xff,(BYTE)0xec,(BYTE)0x70,(BYTE)0x00,(BYTE)0x4e,(BYTE)0x75 },
    { 0x5530,(BYTE)0x22,(BYTE)0x2f,(BYTE)0x00,(BYTE)0x0c,(BYTE)0x34,(BYTE)0x2f,(BYTE)0x00,(BYTE)0x08,(BYTE)0x30,(BYTE)0x2f,(BYTE)0x00,(BYTE)0x0a,(BYTE)0xc4,(BYTE)0xc0,(BYTE)0x82,(BYTE)0xc2 },
    { 0x5540,(BYTE)0x13,(BYTE)0xc1,(BYTE)0xff,(BYTE)0xf0,(BYTE)0x00,(BYTE)0x11,(BYTE)0xe0,(BYTE)0x89,(BYTE)0x13,(BYTE)0xc1,(BYTE)0xff,(BYTE)0xf0,(BYTE)0x00,(BYTE)0x15,(BYTE)0xe0,(BYTE)0x89 },
    { 0x5550,(BYTE)0x82,(BYTE)0xc0,(BYTE)0x30,(BYTE)0x2f,(BYTE)0x00,(BYTE)0x16,(BYTE)0x02,(BYTE)0x00,(BYTE)0x00,(BYTE)0x07,(BYTE)0xe9,(BYTE)0x08,(BYTE)0x82,(BYTE)0x00,(BYTE)0x13,(BYTE)0xc1 },
    { 0x5560,(BYTE)0xff,(BYTE)0xf0,(BYTE)0x00,(BYTE)0x19,(BYTE)0x48,(BYTE)0x41,(BYTE)0x52,(BYTE)0x41,(BYTE)0x13,(BYTE)0xc1,(BYTE)0xff,(BYTE)0xf0,(BYTE)0x00,(BYTE)0x0d,(BYTE)0x4e,(BYTE)0x75 },
    { 0x5570,(BYTE)0x30,(BYTE)0x2f,(BYTE)0x00,(BYTE)0x04,(BYTE)0x02,(BYTE)0x00,(BYTE)0x00,(BYTE)0x07,(BYTE)0xe9,(BYTE)0x08,(BYTE)0x13,(BYTE)0xc0,(BYTE)0xff,(BYTE)0xf0,(BYTE)0x00,(BYTE)0x19 },
    { 0x5580,(BYTE)0x20,(BYTE)0x6f,(BYTE)0x00,(BYTE)0x06,(BYTE)0x4a,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00,(BYTE)0x06,(BYTE)0x0b,(BYTE)0x67,(BYTE)0x06,(BYTE)0x70,(BYTE)0x00,(BYTE)0x61,(BYTE)0x00 },
    { 0x5590,(BYTE)0x00,(BYTE)0xc4,(BYTE)0x13,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x00,(BYTE)0xff,(BYTE)0xf0,(BYTE)0x00,(BYTE)0x39,(BYTE)0x13,(BYTE)0xfc,(BYTE)0x00,(BYTE)0xec,(BYTE)0xff,(BYTE)0xf0 },
    { 0x55a0,(BYTE)0x00,(BYTE)0x1d,(BYTE)0x61,(BYTE)0x00,(BYTE)0xfe,(BYTE)0xb4,(BYTE)0x4a,(BYTE)0x40,(BYTE)0x6b,(BYTE)0x0c,(BYTE)0x08,(BYTE)0x00,(BYTE)0x00,(BYTE)0x03,(BYTE)0x67,(BYTE)0x06 },
    { 0x55b0,(BYTE)0x61,(BYTE)0x00,(BYTE)0x00,(BYTE)0x2a,(BYTE)0x70,(BYTE)0x00,(BYTE)0x4e,(BYTE)0x75,(BYTE)0x30,(BYTE)0x2f,(BYTE)0x00,(BYTE)0x04,(BYTE)0x02,(BYTE)0x00,(BYTE)0x00,(BYTE)0x07 },
    { 0x55c0,(BYTE)0xe9,(BYTE)0x08,(BYTE)0x13,(BYTE)0xc0,(BYTE)0xff,(BYTE)0xf0,(BYTE)0x00,(BYTE)0x19,(BYTE)0x13,(BYTE)0xef,(BYTE)0x00,(BYTE)0x07,(BYTE)0xff,(BYTE)0xf0,(BYTE)0x00,(BYTE)0x09 },
    { 0x55d0,(BYTE)0x13,(BYTE)0xfc,(BYTE)0x00,(BYTE)0xe3,(BYTE)0xff,(BYTE)0xf0,(BYTE)0x00,(BYTE)0x1d,(BYTE)0x60,(BYTE)0x00,(BYTE)0xfe,(BYTE)0x7e,(BYTE)0x4a,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00 },
    { 0x55e0,(BYTE)0x06,(BYTE)0x0b,(BYTE)0x67,(BYTE)0x16,(BYTE)0x33,(BYTE)0x7c,(BYTE)0x00,(BYTE)0x01,(BYTE)0x00,(BYTE)0x38,(BYTE)0x13,(BYTE)0x7c,(BYTE)0x00,(BYTE)0x80,(BYTE)0x00,(BYTE)0x3c },
    { 0x55f0,(BYTE)0x61,(BYTE)0x00,(BYTE)0x00,(BYTE)0xc2,(BYTE)0x54,(BYTE)0xa9,(BYTE)0x00,(BYTE)0x32,(BYTE)0x4e,(BYTE)0x75,(BYTE)0x70,(BYTE)0x0f,(BYTE)0x43,(BYTE)0xf9,(BYTE)0xff,(BYTE)0xf0 },
    { 0x5600,(BYTE)0x00,(BYTE)0x00,(BYTE)0x20,(BYTE)0xd1,(BYTE)0x20,(BYTE)0xd1,(BYTE)0x20,(BYTE)0xd1,(BYTE)0x20,(BYTE)0xd1,(BYTE)0x20,(BYTE)0xd1,(BYTE)0x20,(BYTE)0xd1,(BYTE)0x20,(BYTE)0xd1 },
    { 0x5610,(BYTE)0x20,(BYTE)0xd1,(BYTE)0x51,(BYTE)0xc8,(BYTE)0xff,(BYTE)0xee,(BYTE)0x4e,(BYTE)0x75,(BYTE)0x4a,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00,(BYTE)0x06,(BYTE)0x0b,(BYTE)0x67,(BYTE)0x16 },
    { 0x5620,(BYTE)0x33,(BYTE)0x7c,(BYTE)0x00,(BYTE)0x01,(BYTE)0x00,(BYTE)0x38,(BYTE)0x13,(BYTE)0x7c,(BYTE)0x00,(BYTE)0x80,(BYTE)0x00,(BYTE)0x3c,(BYTE)0x61,(BYTE)0x00,(BYTE)0x00,(BYTE)0x86 },
    { 0x5630,(BYTE)0x54,(BYTE)0xa9,(BYTE)0x00,(BYTE)0x24,(BYTE)0x4e,(BYTE)0x75,(BYTE)0x70,(BYTE)0x0f,(BYTE)0x43,(BYTE)0xf9,(BYTE)0xff,(BYTE)0xf0,(BYTE)0x00,(BYTE)0x00,(BYTE)0x22,(BYTE)0x98 },
    { 0x5640,(BYTE)0x22,(BYTE)0x98,(BYTE)0x22,(BYTE)0x98,(BYTE)0x22,(BYTE)0x98,(BYTE)0x22,(BYTE)0x98,(BYTE)0x22,(BYTE)0x98,(BYTE)0x22,(BYTE)0x98,(BYTE)0x22,(BYTE)0x98,(BYTE)0x51,(BYTE)0xc8 },
    { 0x5650,(BYTE)0xff,(BYTE)0xee,(BYTE)0x4e,(BYTE)0x75,(BYTE)0x43,(BYTE)0xf9,(BYTE)0x00,(BYTE)0xff,(BYTE)0x8a,(BYTE)0x00,(BYTE)0x4a,(BYTE)0x00,(BYTE)0x66,(BYTE)0x1a,(BYTE)0x23,(BYTE)0x7c },
    { 0x5660,(BYTE)0xff,(BYTE)0xf0,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x24,(BYTE)0x23,(BYTE)0x48,(BYTE)0x00,(BYTE)0x32,(BYTE)0x33,(BYTE)0x7c,(BYTE)0x00,(BYTE)0x02,(BYTE)0x00,(BYTE)0x2e },
    { 0x5670,(BYTE)0x70,(BYTE)0x00,(BYTE)0x33,(BYTE)0x40,(BYTE)0x00,(BYTE)0x20,(BYTE)0x60,(BYTE)0x18,(BYTE)0x23,(BYTE)0x48,(BYTE)0x00,(BYTE)0x24,(BYTE)0x23,(BYTE)0x7c,(BYTE)0xff,(BYTE)0xf0 },
    { 0x5680,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x32,(BYTE)0x33,(BYTE)0x7c,(BYTE)0x00,(BYTE)0x02,(BYTE)0x00,(BYTE)0x20,(BYTE)0x70,(BYTE)0x00,(BYTE)0x33,(BYTE)0x40,(BYTE)0x00,(BYTE)0x2e },
    { 0x5690,(BYTE)0x33,(BYTE)0x40,(BYTE)0x00,(BYTE)0x22,(BYTE)0x33,(BYTE)0x40,(BYTE)0x00,(BYTE)0x30,(BYTE)0x13,(BYTE)0x40,(BYTE)0x00,(BYTE)0x3d,(BYTE)0x70,(BYTE)0xff,(BYTE)0x23,(BYTE)0x40 },
    { 0x56a0,(BYTE)0x00,(BYTE)0x28,(BYTE)0x33,(BYTE)0x40,(BYTE)0x00,(BYTE)0x2c,(BYTE)0x33,(BYTE)0x7c,(BYTE)0x02,(BYTE)0x03,(BYTE)0x00,(BYTE)0x3a,(BYTE)0x33,(BYTE)0x7c,(BYTE)0x01,(BYTE)0x00 },
    { 0x56b0,(BYTE)0x00,(BYTE)0x36,(BYTE)0x4e,(BYTE)0x75,(BYTE)0x4e,(BYTE)0x71,(BYTE)0x4a,(BYTE)0xe9,(BYTE)0x00,(BYTE)0x3c,(BYTE)0x6b,(BYTE)0xf8,(BYTE)0x4e,(BYTE)0x75,(BYTE)0x20,(BYTE)0x6f },
    { 0x56c0,(BYTE)0x00,(BYTE)0x04,(BYTE)0xd1,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x50,(BYTE)0x2f,(BYTE)0x08,(BYTE)0x48,(BYTE)0x79,(BYTE)0x00,(BYTE)0x00,(BYTE)0x1c,(BYTE)0xec },
    { 0x56d0,(BYTE)0x3f,(BYTE)0x3c,(BYTE)0x00,(BYTE)0x06,(BYTE)0x61,(BYTE)0x00,(BYTE)0x00,(BYTE)0x3a,(BYTE)0xde,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x0a,(BYTE)0x4a,(BYTE)0x40,(BYTE)0x66,(BYTE)0x0e },
    { 0x56e0,(BYTE)0x30,(BYTE)0x3c,(BYTE)0x02,(BYTE)0x67,(BYTE)0x32,(BYTE)0x3c,(BYTE)0x00,(BYTE)0x04,(BYTE)0x34,(BYTE)0x3c,(BYTE)0x00,(BYTE)0x11,(BYTE)0x60,(BYTE)0x10,(BYTE)0x20,(BYTE)0x6f },
    { 0x56f0,(BYTE)0x00,(BYTE)0x04,(BYTE)0x30,(BYTE)0x28,(BYTE)0x00,(BYTE)0x02,(BYTE)0x32,(BYTE)0x28,(BYTE)0x00,(BYTE)0x06,(BYTE)0x34,(BYTE)0x28,(BYTE)0x00,(BYTE)0x0c,(BYTE)0x4e,(BYTE)0x75 },
    { 0x5700,(BYTE)0x43,(BYTE)0x6f,(BYTE)0x6e,(BYTE)0x6e,(BYTE)0x65,(BYTE)0x72,(BYTE)0x00,(BYTE)0x00,(BYTE)0x43,(BYTE)0x50,(BYTE)0x32,(BYTE)0x30,(BYTE)0x32,(BYTE)0x34,(BYTE)0x00,(BYTE)0x00 },
    { 0x5710,(BYTE)0x48,(BYTE)0xe7,(BYTE)0x40,(BYTE)0xc0,(BYTE)0x32,(BYTE)0x2f,(BYTE)0x00,(BYTE)0x10,(BYTE)0x53,(BYTE)0x41,(BYTE)0x20,(BYTE)0x6f,(BYTE)0x00,(BYTE)0x12,(BYTE)0x22,(BYTE)0x6f },
    { 0x5720,(BYTE)0x00,(BYTE)0x16,(BYTE)0x70,(BYTE)0x01,(BYTE)0xb3,(BYTE)0x08,(BYTE)0x66,(BYTE)0x06,(BYTE)0x51,(BYTE)0xc9,(BYTE)0xff,(BYTE)0xfa,(BYTE)0x70,(BYTE)0x00,(BYTE)0x4c,(BYTE)0xdf },
    { 0x5730,(BYTE)0x03,(BYTE)0x02,(BYTE)0x4e,(BYTE)0x75,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x61,(BYTE)0x00 },
    { 0x5740,(BYTE)0x01,(BYTE)0x4a,(BYTE)0x41,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x1d,(BYTE)0x18,(BYTE)0x10,(BYTE)0xbc,(BYTE)0x00,(BYTE)0x00,(BYTE)0x30,(BYTE)0x2f,(BYTE)0x00,(BYTE)0x04 },
    { 0x5750,(BYTE)0x74,(BYTE)0x06,(BYTE)0x61,(BYTE)0x00,(BYTE)0x01,(BYTE)0xba,(BYTE)0x4e,(BYTE)0x75,(BYTE)0x61,(BYTE)0x00,(BYTE)0x01,(BYTE)0x30,(BYTE)0x41,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00 },
    { 0x5760,(BYTE)0x1d,(BYTE)0x18,(BYTE)0x10,(BYTE)0xbc,(BYTE)0x00,(BYTE)0x03,(BYTE)0x11,(BYTE)0x6f,(BYTE)0x00,(BYTE)0x07,(BYTE)0x00,(BYTE)0x04,(BYTE)0x72,(BYTE)0x04,(BYTE)0xb2,(BYTE)0x6f },
    { 0x5770,(BYTE)0x00,(BYTE)0x06,(BYTE)0x64,(BYTE)0x04,(BYTE)0x32,(BYTE)0x2f,(BYTE)0x00,(BYTE)0x06,(BYTE)0x30,(BYTE)0x2f,(BYTE)0x00,(BYTE)0x04,(BYTE)0x74,(BYTE)0x06,(BYTE)0x20,(BYTE)0x6f },
    { 0x5780,(BYTE)0x00,(BYTE)0x08,(BYTE)0x61,(BYTE)0x00,(BYTE)0x01,(BYTE)0x9c,(BYTE)0x4e,(BYTE)0x75,(BYTE)0x41,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x1d,(BYTE)0x18,(BYTE)0x10,(BYTE)0xfc },
    { 0x5790,(BYTE)0x00,(BYTE)0x08,(BYTE)0x10,(BYTE)0xef,(BYTE)0x00,(BYTE)0x05,(BYTE)0x10,(BYTE)0xef,(BYTE)0x00,(BYTE)0x06,(BYTE)0x10,(BYTE)0xef,(BYTE)0x00,(BYTE)0x07,(BYTE)0x10,(BYTE)0xef },
    { 0x57a0,(BYTE)0x00,(BYTE)0x09,(BYTE)0x42,(BYTE)0x10,(BYTE)0x30,(BYTE)0x2f,(BYTE)0x00,(BYTE)0x0e,(BYTE)0x72,(BYTE)0x00,(BYTE)0x32,(BYTE)0x2f,(BYTE)0x00,(BYTE)0x08,(BYTE)0xe1,(BYTE)0x89 },
    { 0x57b0,(BYTE)0xd2,(BYTE)0x81,(BYTE)0x74,(BYTE)0x06,(BYTE)0x20,(BYTE)0x6f,(BYTE)0x00,(BYTE)0x0a,(BYTE)0x61,(BYTE)0x00,(BYTE)0x01,(BYTE)0x66,(BYTE)0x4e,(BYTE)0x75,(BYTE)0x41,(BYTE)0xf9 },
    { 0x57c0,(BYTE)0x00,(BYTE)0x00,(BYTE)0x1d,(BYTE)0x18,(BYTE)0x10,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x0a,(BYTE)0x10,(BYTE)0xef,(BYTE)0x00,(BYTE)0x05,(BYTE)0x10,(BYTE)0xef,(BYTE)0x00,(BYTE)0x06 },
    { 0x57d0,(BYTE)0x10,(BYTE)0xef,(BYTE)0x00,(BYTE)0x07,(BYTE)0x10,(BYTE)0xef,(BYTE)0x00,(BYTE)0x09,(BYTE)0x42,(BYTE)0x10,(BYTE)0x30,(BYTE)0x2f,(BYTE)0x00,(BYTE)0x0e,(BYTE)0x72,(BYTE)0x00 },
    { 0x57e0,(BYTE)0x32,(BYTE)0x2f,(BYTE)0x00,(BYTE)0x08,(BYTE)0xe1,(BYTE)0x89,(BYTE)0xd2,(BYTE)0x81,(BYTE)0x74,(BYTE)0x06,(BYTE)0x20,(BYTE)0x6f,(BYTE)0x00,(BYTE)0x0a,(BYTE)0x61,(BYTE)0x00 },
    { 0x57f0,(BYTE)0x01,(BYTE)0x42,(BYTE)0x4e,(BYTE)0x75,(BYTE)0x61,(BYTE)0x00,(BYTE)0x00,(BYTE)0x94,(BYTE)0x41,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x1d,(BYTE)0x18,(BYTE)0x10,(BYTE)0xbc },
    { 0x5800,(BYTE)0x00,(BYTE)0x12,(BYTE)0x11,(BYTE)0x6f,(BYTE)0x00,(BYTE)0x07,(BYTE)0x00,(BYTE)0x04,(BYTE)0x30,(BYTE)0x2f,(BYTE)0x00,(BYTE)0x04,(BYTE)0x72,(BYTE)0x00,(BYTE)0x32,(BYTE)0x2f },
    { 0x5810,(BYTE)0x00,(BYTE)0x06,(BYTE)0x74,(BYTE)0x06,(BYTE)0x20,(BYTE)0x6f,(BYTE)0x00,(BYTE)0x08,(BYTE)0x61,(BYTE)0x00,(BYTE)0x01,(BYTE)0x06,(BYTE)0x4e,(BYTE)0x75,(BYTE)0x41,(BYTE)0xf9 },
    { 0x5820,(BYTE)0x00,(BYTE)0x00,(BYTE)0x1d,(BYTE)0x18,(BYTE)0x10,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x28,(BYTE)0x42,(BYTE)0x18,(BYTE)0x20,(BYTE)0xef,(BYTE)0x00,(BYTE)0x04,(BYTE)0x42,(BYTE)0x18 },
    { 0x5830,(BYTE)0x10,(BYTE)0xef,(BYTE)0x00,(BYTE)0x08,(BYTE)0x10,(BYTE)0xef,(BYTE)0x00,(BYTE)0x09,(BYTE)0x42,(BYTE)0x10,(BYTE)0x30,(BYTE)0x2f,(BYTE)0x00,(BYTE)0x0e,(BYTE)0x72,(BYTE)0x00 },
    { 0x5840,(BYTE)0x32,(BYTE)0x2f,(BYTE)0x00,(BYTE)0x08,(BYTE)0xe1,(BYTE)0x89,(BYTE)0xd2,(BYTE)0x81,(BYTE)0x74,(BYTE)0x0a,(BYTE)0x20,(BYTE)0x6f,(BYTE)0x00,(BYTE)0x0a,(BYTE)0x61,(BYTE)0x00 },
    { 0x5850,(BYTE)0x00,(BYTE)0xd0,(BYTE)0x4e,(BYTE)0x75,(BYTE)0x41,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x1d,(BYTE)0x18,(BYTE)0x10,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x2a,(BYTE)0x42,(BYTE)0x18 },
    { 0x5860,(BYTE)0x20,(BYTE)0xef,(BYTE)0x00,(BYTE)0x04,(BYTE)0x42,(BYTE)0x18,(BYTE)0x10,(BYTE)0xef,(BYTE)0x00,(BYTE)0x08,(BYTE)0x10,(BYTE)0xef,(BYTE)0x00,(BYTE)0x09,(BYTE)0x42,(BYTE)0x10 },
    { 0x5870,(BYTE)0x30,(BYTE)0x2f,(BYTE)0x00,(BYTE)0x0e,(BYTE)0x72,(BYTE)0x00,(BYTE)0x32,(BYTE)0x2f,(BYTE)0x00,(BYTE)0x08,(BYTE)0xe1,(BYTE)0x89,(BYTE)0xd2,(BYTE)0x81,(BYTE)0x74,(BYTE)0x0a },
    { 0x5880,(BYTE)0x20,(BYTE)0x6f,(BYTE)0x00,(BYTE)0x0a,(BYTE)0x61,(BYTE)0x00,(BYTE)0x00,(BYTE)0xac,(BYTE)0x4e,(BYTE)0x75,(BYTE)0x41,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x1d,(BYTE)0x18 },
    { 0x5890,(BYTE)0x42,(BYTE)0x98,(BYTE)0x42,(BYTE)0x98,(BYTE)0x42,(BYTE)0x50,(BYTE)0x4e,(BYTE)0x75,(BYTE)0x2f,(BYTE)0x03,(BYTE)0x42,(BYTE)0xb9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x1e,(BYTE)0xee },
    { 0x58a0,(BYTE)0x2f,(BYTE)0x3c,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x01,(BYTE)0x3f,(BYTE)0x3c,(BYTE)0x00,(BYTE)0x20,(BYTE)0x4e,(BYTE)0x41,(BYTE)0x5c,(BYTE)0x4f,(BYTE)0x4a,(BYTE)0x40 },
    { 0x58b0,(BYTE)0x66,(BYTE)0x10,(BYTE)0x42,(BYTE)0xa7,(BYTE)0x3f,(BYTE)0x3c,(BYTE)0x00,(BYTE)0x20,(BYTE)0x4e,(BYTE)0x41,(BYTE)0x5c,(BYTE)0x4f,(BYTE)0x23,(BYTE)0xc0,(BYTE)0x00,(BYTE)0x00 },
    { 0x58c0,(BYTE)0x1e,(BYTE)0xee,(BYTE)0x76,(BYTE)0x00,(BYTE)0x4a,(BYTE)0xb8,(BYTE)0x05,(BYTE)0xa0,(BYTE)0x67,(BYTE)0x24,(BYTE)0x20,(BYTE)0x2f,(BYTE)0x00,(BYTE)0x08,(BYTE)0x20,(BYTE)0x78 },
    { 0x58d0,(BYTE)0x05,(BYTE)0xa0,(BYTE)0x22,(BYTE)0x10,(BYTE)0x67,(BYTE)0x18,(BYTE)0xb0,(BYTE)0x81,(BYTE)0x67,(BYTE)0x04,(BYTE)0x50,(BYTE)0x48,(BYTE)0x60,(BYTE)0xf4,(BYTE)0x76,(BYTE)0x01 },
    { 0x58e0,(BYTE)0x4a,(BYTE)0xaf,(BYTE)0x00,(BYTE)0x0c,(BYTE)0x67,(BYTE)0x08,(BYTE)0x22,(BYTE)0x6f,(BYTE)0x00,(BYTE)0x0c,(BYTE)0x22,(BYTE)0xa8,(BYTE)0x00,(BYTE)0x04,(BYTE)0x4a,(BYTE)0xb9 },
    { 0x58f0,(BYTE)0x00,(BYTE)0x00,(BYTE)0x1e,(BYTE)0xee,(BYTE)0x67,(BYTE)0x0e,(BYTE)0x2f,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00,(BYTE)0x1e,(BYTE)0xee,(BYTE)0x3f,(BYTE)0x3c,(BYTE)0x00,(BYTE)0x20 },
    { 0x5900,(BYTE)0x4e,(BYTE)0x41,(BYTE)0x5c,(BYTE)0x4f,(BYTE)0x20,(BYTE)0x03,(BYTE)0x26,(BYTE)0x1f,(BYTE)0x4e,(BYTE)0x75,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x08,(BYTE)0x00 },
    { 0x5910,(BYTE)0x00,(BYTE)0x03,(BYTE)0x66,(BYTE)0x06,(BYTE)0x61,(BYTE)0x00,(BYTE)0xf9,(BYTE)0xcc,(BYTE)0x60,(BYTE)0x04,(BYTE)0x61,(BYTE)0x00,(BYTE)0x00,(BYTE)0x28,(BYTE)0x4e,(BYTE)0x75 },
    { 0x5920,(BYTE)0x08,(BYTE)0x00,(BYTE)0x00,(BYTE)0x03,(BYTE)0x66,(BYTE)0x06,(BYTE)0x61,(BYTE)0x00,(BYTE)0xf9,(BYTE)0xd2,(BYTE)0x60,(BYTE)0x04,(BYTE)0x61,(BYTE)0x00,(BYTE)0x00,(BYTE)0x26 },
    { 0x5930,(BYTE)0x4e,(BYTE)0x75,(BYTE)0x08,(BYTE)0x00,(BYTE)0x00,(BYTE)0x03,(BYTE)0x66,(BYTE)0x06,(BYTE)0x61,(BYTE)0x00,(BYTE)0xfa,(BYTE)0x56,(BYTE)0x60,(BYTE)0x04,(BYTE)0x61,(BYTE)0x00 },
    { 0x5940,(BYTE)0x00,(BYTE)0xb4,(BYTE)0x4e,(BYTE)0x75,(BYTE)0x02,(BYTE)0x40,(BYTE)0x00,(BYTE)0x07,(BYTE)0x61,(BYTE)0x00,(BYTE)0x00,(BYTE)0xfa,(BYTE)0x6a,(BYTE)0x02,(BYTE)0x4e,(BYTE)0x75 },
    { 0x5950,(BYTE)0x60,(BYTE)0x00,(BYTE)0x02,(BYTE)0x1a,(BYTE)0x02,(BYTE)0x40,(BYTE)0x00,(BYTE)0x07,(BYTE)0x3f,(BYTE)0x02,(BYTE)0x61,(BYTE)0x00,(BYTE)0x00,(BYTE)0xe8,(BYTE)0x34,(BYTE)0x1f },
    { 0x5960,(BYTE)0x4a,(BYTE)0x40,(BYTE)0x6b,(BYTE)0x00,(BYTE)0x00,(BYTE)0x8e,(BYTE)0x11,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x00,(BYTE)0x87,(BYTE)0x83,(BYTE)0x11,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x01 },
    { 0x5970,(BYTE)0x87,(BYTE)0x87,(BYTE)0x10,(BYTE)0x38,(BYTE)0x87,(BYTE)0x8f,(BYTE)0x11,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x02,(BYTE)0x87,(BYTE)0x85,(BYTE)0x11,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x00 },
    { 0x5980,(BYTE)0x87,(BYTE)0x8f,(BYTE)0x31,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x00,(BYTE)0x87,(BYTE)0x14,(BYTE)0x31,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x02,(BYTE)0x87,(BYTE)0x14,(BYTE)0x61,(BYTE)0x00 },
    { 0x5990,(BYTE)0x02,(BYTE)0x6e,(BYTE)0x0c,(BYTE)0x42,(BYTE)0x00,(BYTE)0x06,(BYTE)0x67,(BYTE)0x04,(BYTE)0x61,(BYTE)0x00,(BYTE)0x02,(BYTE)0x84,(BYTE)0x61,(BYTE)0x00,(BYTE)0x01,(BYTE)0x7c },
    { 0x59a0,(BYTE)0x4a,(BYTE)0x40,(BYTE)0x66,(BYTE)0x4e,(BYTE)0x61,(BYTE)0x00,(BYTE)0x01,(BYTE)0xc6,(BYTE)0x4a,(BYTE)0x40,(BYTE)0x66,(BYTE)0x46,(BYTE)0x2f,(BYTE)0x00,(BYTE)0x40,(BYTE)0xe7 },
    { 0x59b0,(BYTE)0x00,(BYTE)0x7c,(BYTE)0x07,(BYTE)0x00,(BYTE)0x4e,(BYTE)0x7a,(BYTE)0x00,(BYTE)0x02,(BYTE)0x00,(BYTE)0x40,(BYTE)0x08,(BYTE)0x08,(BYTE)0x4e,(BYTE)0x7b,(BYTE)0x00,(BYTE)0x02 },
    { 0x59c0,(BYTE)0x46,(BYTE)0xdf,(BYTE)0x10,(BYTE)0x38,(BYTE)0x87,(BYTE)0x07,(BYTE)0x02,(BYTE)0x40,(BYTE)0x00,(BYTE)0x03,(BYTE)0x67,(BYTE)0x24,(BYTE)0x22,(BYTE)0x7c,(BYTE)0xff,(BYTE)0xff },
    { 0x59d0,(BYTE)0x87,(BYTE)0x01,(BYTE)0x01,(BYTE)0x49,(BYTE)0x00,(BYTE)0x00,(BYTE)0x22,(BYTE)0x00,(BYTE)0x02,(BYTE)0x40,(BYTE)0x00,(BYTE)0x03,(BYTE)0x02,(BYTE)0x41,(BYTE)0xff,(BYTE)0xfc },
    { 0x59e0,(BYTE)0x20,(BYTE)0x41,(BYTE)0x22,(BYTE)0x38,(BYTE)0x87,(BYTE)0x10,(BYTE)0x53,(BYTE)0x40,(BYTE)0xe1,(BYTE)0x99,(BYTE)0x10,(BYTE)0xc1,(BYTE)0x51,(BYTE)0xc8,(BYTE)0xff,(BYTE)0xfa },
    { 0x59f0,(BYTE)0x20,(BYTE)0x1f,(BYTE)0x4e,(BYTE)0x75,(BYTE)0x02,(BYTE)0x40,(BYTE)0x00,(BYTE)0x07,(BYTE)0x3f,(BYTE)0x02,(BYTE)0x61,(BYTE)0x00,(BYTE)0x00,(BYTE)0x48,(BYTE)0x34,(BYTE)0x1f },
    { 0x5a00,(BYTE)0x4a,(BYTE)0x40,(BYTE)0x6a,(BYTE)0x02,(BYTE)0x4e,(BYTE)0x75,(BYTE)0x11,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x00,(BYTE)0x87,(BYTE)0x87,(BYTE)0x10,(BYTE)0x38,(BYTE)0x87,(BYTE)0x8f },
    { 0x5a10,(BYTE)0x11,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x02,(BYTE)0x87,(BYTE)0x85,(BYTE)0x11,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x00,(BYTE)0x87,(BYTE)0x8b,(BYTE)0x31,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x01 },
    { 0x5a20,(BYTE)0x87,(BYTE)0x14,(BYTE)0x31,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x03,(BYTE)0x87,(BYTE)0x14,(BYTE)0x61,(BYTE)0x00,(BYTE)0x01,(BYTE)0xd4,(BYTE)0x0c,(BYTE)0x42,(BYTE)0x00,(BYTE)0x06 },
    { 0x5a30,(BYTE)0x67,(BYTE)0x04,(BYTE)0x61,(BYTE)0x00,(BYTE)0x01,(BYTE)0xea,(BYTE)0x61,(BYTE)0x00,(BYTE)0x00,(BYTE)0xe2,(BYTE)0x4a,(BYTE)0x40,(BYTE)0x66,(BYTE)0x04,(BYTE)0x61,(BYTE)0x00 },
    { 0x5a40,(BYTE)0x01,(BYTE)0x2c,(BYTE)0x4e,(BYTE)0x75,(BYTE)0x48,(BYTE)0xe7,(BYTE)0x60,(BYTE)0x80,(BYTE)0x3f,(BYTE)0x00,(BYTE)0x61,(BYTE)0x00,(BYTE)0x00,(BYTE)0x4c,(BYTE)0x54,(BYTE)0x8f },
    { 0x5a50,(BYTE)0x4c,(BYTE)0xdf,(BYTE)0x01,(BYTE)0x06,(BYTE)0x4a,(BYTE)0x40,(BYTE)0x6b,(BYTE)0x3e,(BYTE)0x20,(BYTE)0x08,(BYTE)0x22,(BYTE)0x7c,(BYTE)0xff,(BYTE)0xff,(BYTE)0x87,(BYTE)0x01 },
    { 0x5a60,(BYTE)0x01,(BYTE)0xc9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x22,(BYTE)0x7c,(BYTE)0xff,(BYTE)0xff,(BYTE)0x87,(BYTE)0x09,(BYTE)0x03,(BYTE)0xc9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x11,(BYTE)0xfc },
    { 0x5a70,(BYTE)0x00,(BYTE)0x02,(BYTE)0x87,(BYTE)0x87,(BYTE)0x11,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x01,(BYTE)0x87,(BYTE)0x83,(BYTE)0x61,(BYTE)0x00,(BYTE)0x01,(BYTE)0x82,(BYTE)0x43,(BYTE)0xf9 },
    { 0x5a80,(BYTE)0x00,(BYTE)0x00,(BYTE)0x1d,(BYTE)0x18,(BYTE)0x53,(BYTE)0x42,(BYTE)0x10,(BYTE)0x19,(BYTE)0x61,(BYTE)0x00,(BYTE)0x01,(BYTE)0x62,(BYTE)0x4a,(BYTE)0x40,(BYTE)0x6b,(BYTE)0x06 },
    { 0x5a90,(BYTE)0x51,(BYTE)0xca,(BYTE)0xff,(BYTE)0xf4,(BYTE)0x70,(BYTE)0x00,(BYTE)0x4e,(BYTE)0x75,(BYTE)0x61,(BYTE)0x00,(BYTE)0x01,(BYTE)0x64,(BYTE)0x08,(BYTE)0x38,(BYTE)0x00,(BYTE)0x06 },
    { 0x5aa0,(BYTE)0x87,(BYTE)0x89,(BYTE)0x67,(BYTE)0x06,(BYTE)0xb2,(BYTE)0x90,(BYTE)0x62,(BYTE)0xf4,(BYTE)0x60,(BYTE)0x44,(BYTE)0x11,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x00,(BYTE)0x87,(BYTE)0x87 },
    { 0x5ab0,(BYTE)0x11,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x00,(BYTE)0x87,(BYTE)0x89,(BYTE)0x11,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x0c,(BYTE)0x87,(BYTE)0x83,(BYTE)0x42,(BYTE)0x40,(BYTE)0x32,(BYTE)0x2f },
    { 0x5ac0,(BYTE)0x00,(BYTE)0x04,(BYTE)0x03,(BYTE)0xc0,(BYTE)0x11,(BYTE)0xc0,(BYTE)0x87,(BYTE)0x81,(BYTE)0x11,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x0d,(BYTE)0x87,(BYTE)0x83,(BYTE)0x02,(BYTE)0x38 },
    { 0x5ad0,(BYTE)0x00,(BYTE)0xfe,(BYTE)0x87,(BYTE)0x85,(BYTE)0x02,(BYTE)0x38,(BYTE)0x00,(BYTE)0xf7,(BYTE)0x87,(BYTE)0x83,(BYTE)0x4e,(BYTE)0x71,(BYTE)0x4e,(BYTE)0x71,(BYTE)0x61,(BYTE)0x00 },
    { 0x5ae0,(BYTE)0x01,(BYTE)0x1e,(BYTE)0x08,(BYTE)0x38,(BYTE)0x00,(BYTE)0x06,(BYTE)0x87,(BYTE)0x89,(BYTE)0x66,(BYTE)0x08,(BYTE)0xb2,(BYTE)0x90,(BYTE)0x62,(BYTE)0xf4,(BYTE)0x70,(BYTE)0xff },
    { 0x5af0,(BYTE)0x60,(BYTE)0x02,(BYTE)0x42,(BYTE)0x40,(BYTE)0x11,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x00,(BYTE)0x87,(BYTE)0x83,(BYTE)0x4e,(BYTE)0x75,(BYTE)0x11,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x80 },
    { 0x5b00,(BYTE)0x87,(BYTE)0x83,(BYTE)0x61,(BYTE)0x00,(BYTE)0x00,(BYTE)0xfa,(BYTE)0xb2,(BYTE)0x90,(BYTE)0x62,(BYTE)0xfc,(BYTE)0x11,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x00,(BYTE)0x87,(BYTE)0x83 },
    { 0x5b10,(BYTE)0x61,(BYTE)0x00,(BYTE)0x00,(BYTE)0xfc,(BYTE)0xb2,(BYTE)0x90,(BYTE)0x62,(BYTE)0xfc,(BYTE)0x4e,(BYTE)0x75,(BYTE)0xd2,(BYTE)0xb9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x40 },
    { 0x5b20,(BYTE)0x08,(BYTE)0x38,(BYTE)0x00,(BYTE)0x07,(BYTE)0xfa,(BYTE)0x81,(BYTE)0x66,(BYTE)0x2a,(BYTE)0xb2,(BYTE)0x90,(BYTE)0x63,(BYTE)0x0c,(BYTE)0x08,(BYTE)0x38,(BYTE)0x00,(BYTE)0x05 },
    { 0x5b30,(BYTE)0xfa,(BYTE)0x81,(BYTE)0x67,(BYTE)0x0a,(BYTE)0xb2,(BYTE)0x90,(BYTE)0x62,(BYTE)0xe8,(BYTE)0x61,(BYTE)0xc2,(BYTE)0x70,(BYTE)0xff,(BYTE)0x60,(BYTE)0x2c,(BYTE)0x30,(BYTE)0x38 },
    { 0x5b40,(BYTE)0x87,(BYTE)0x14,(BYTE)0x02,(BYTE)0x80,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x80,(BYTE)0x67,(BYTE)0xd6,(BYTE)0x3f,(BYTE)0x00,(BYTE)0x61,(BYTE)0xae,(BYTE)0x30,(BYTE)0x1f },
    { 0x5b50,(BYTE)0x60,(BYTE)0x18,(BYTE)0x70,(BYTE)0x00,(BYTE)0x12,(BYTE)0x38,(BYTE)0x87,(BYTE)0x8f,(BYTE)0x31,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x00,(BYTE)0x87,(BYTE)0x14,(BYTE)0x11,(BYTE)0xfc },
    { 0x5b60,(BYTE)0x00,(BYTE)0x00,(BYTE)0x87,(BYTE)0x85,(BYTE)0x11,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x00,(BYTE)0x87,(BYTE)0x83,(BYTE)0x4e,(BYTE)0x75,(BYTE)0x61,(BYTE)0x00,(BYTE)0x00,(BYTE)0x90 },
    { 0x5b70,(BYTE)0x11,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x03,(BYTE)0x87,(BYTE)0x87,(BYTE)0x10,(BYTE)0x38,(BYTE)0x87,(BYTE)0x8f,(BYTE)0x61,(BYTE)0x00,(BYTE)0x00,(BYTE)0x3c,(BYTE)0x6b,(BYTE)0x30 },
    { 0x5b80,(BYTE)0x70,(BYTE)0x00,(BYTE)0x10,(BYTE)0x38,(BYTE)0x87,(BYTE)0x81,(BYTE)0x2f,(BYTE)0x00,(BYTE)0x61,(BYTE)0x00,(BYTE)0x00,(BYTE)0x74,(BYTE)0x61,(BYTE)0x00,(BYTE)0x00,(BYTE)0x3e },
    { 0x5b90,(BYTE)0x4a,(BYTE)0x40,(BYTE)0x67,(BYTE)0x04,(BYTE)0x58,(BYTE)0x8f,(BYTE)0x60,(BYTE)0x18,(BYTE)0x61,(BYTE)0x00,(BYTE)0x00,(BYTE)0x64,(BYTE)0x61,(BYTE)0x00,(BYTE)0x00,(BYTE)0x1a },
    { 0x5ba0,(BYTE)0x6b,(BYTE)0xf2,(BYTE)0x10,(BYTE)0x38,(BYTE)0x87,(BYTE)0x81,(BYTE)0x61,(BYTE)0x00,(BYTE)0x00,(BYTE)0x24,(BYTE)0x4a,(BYTE)0x40,(BYTE)0x6b,(BYTE)0xe6,(BYTE)0x20,(BYTE)0x1f },
    { 0x5bb0,(BYTE)0x4e,(BYTE)0x75,(BYTE)0xd2,(BYTE)0xb9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x40,(BYTE)0x08,(BYTE)0x38,(BYTE)0x00,(BYTE)0x05,(BYTE)0x87,(BYTE)0x89,(BYTE)0x66,(BYTE)0x08 },
    { 0x5bc0,(BYTE)0xb2,(BYTE)0x90,(BYTE)0x62,(BYTE)0xf4,(BYTE)0x70,(BYTE)0xff,(BYTE)0x60,(BYTE)0x02,(BYTE)0x70,(BYTE)0x00,(BYTE)0x4e,(BYTE)0x75,(BYTE)0x00,(BYTE)0x38,(BYTE)0x00,(BYTE)0x11 },
    { 0x5bd0,(BYTE)0x87,(BYTE)0x83,(BYTE)0x08,(BYTE)0x38,(BYTE)0x00,(BYTE)0x05,(BYTE)0x87,(BYTE)0x89,(BYTE)0x67,(BYTE)0x08,(BYTE)0xb2,(BYTE)0x90,(BYTE)0x62,(BYTE)0xf4,(BYTE)0x70,(BYTE)0xff },
    { 0x5be0,(BYTE)0x60,(BYTE)0x02,(BYTE)0x70,(BYTE)0x00,(BYTE)0x02,(BYTE)0x38,(BYTE)0x00,(BYTE)0xef,(BYTE)0x87,(BYTE)0x83,(BYTE)0x4e,(BYTE)0x75,(BYTE)0x3f,(BYTE)0x00,(BYTE)0x61,(BYTE)0xc8 },
    { 0x5bf0,(BYTE)0x6b,(BYTE)0x08,(BYTE)0x11,(BYTE)0xef,(BYTE)0x00,(BYTE)0x01,(BYTE)0x87,(BYTE)0x81,(BYTE)0x61,(BYTE)0xd2,(BYTE)0x54,(BYTE)0x8f,(BYTE)0x4e,(BYTE)0x75,(BYTE)0x20,(BYTE)0x7c },
    { 0x5c00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x04,(BYTE)0xba,(BYTE)0x22,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x3c,(BYTE)0xd2,(BYTE)0x90,(BYTE)0x4e,(BYTE)0x75,(BYTE)0x20,(BYTE)0x7c },
    { 0x5c10,(BYTE)0x00,(BYTE)0x00,(BYTE)0x04,(BYTE)0xba,(BYTE)0x22,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x38,(BYTE)0xd2,(BYTE)0x90,(BYTE)0x4e,(BYTE)0x75,(BYTE)0x20,(BYTE)0x7c },
    { 0x5c20,(BYTE)0x00,(BYTE)0x00,(BYTE)0x04,(BYTE)0xba,(BYTE)0x22,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x2c,(BYTE)0xd2,(BYTE)0x90,(BYTE)0x4e,(BYTE)0x75,(BYTE)0x00,(BYTE)0x00 },
    { 0x5c30,(BYTE)0x00,(BYTE)0x00,(BYTE)0x23,(BYTE)0xc0,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0x98,(BYTE)0x20,(BYTE)0x7c,(BYTE)0x00,(BYTE)0x00,(BYTE)0x29,(BYTE)0x84,(BYTE)0x61,(BYTE)0x00 },
    { 0x5c40,(BYTE)0x07,(BYTE)0x10,(BYTE)0x4a,(BYTE)0x79,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0x96,(BYTE)0x66,(BYTE)0x10,(BYTE)0x42,(BYTE)0xa7,(BYTE)0x3f,(BYTE)0x3c,(BYTE)0x00,(BYTE)0x20 },
    { 0x5c50,(BYTE)0x4e,(BYTE)0x41,(BYTE)0x5c,(BYTE)0x8f,(BYTE)0x23,(BYTE)0xc0,(BYTE)0x00,(BYTE)0x00,(BYTE)0x06,(BYTE)0x06,(BYTE)0x2f,(BYTE)0x3c,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xa4 },
    { 0x5c60,(BYTE)0x2f,(BYTE)0x3c,(BYTE)0x5f,(BYTE)0x4d,(BYTE)0x43,(BYTE)0x48,(BYTE)0x61,(BYTE)0x00,(BYTE)0xfc,(BYTE)0x30,(BYTE)0x50,(BYTE)0x4f,(BYTE)0x51,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00 },
    { 0x5c70,(BYTE)0x06,(BYTE)0x0a,(BYTE)0x20,(BYTE)0x78,(BYTE)0x00,(BYTE)0x10,(BYTE)0x21,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x00,(BYTE)0x22,(BYTE)0x6e,(BYTE)0x00,(BYTE)0x10,(BYTE)0x22,(BYTE)0x4f },
    { 0x5c80,(BYTE)0x4e,(BYTE)0x7a,(BYTE)0x00,(BYTE)0x02,(BYTE)0x50,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x06,(BYTE)0x0a,(BYTE)0x21,(BYTE)0xc8,(BYTE)0x00,(BYTE)0x10,(BYTE)0x2e,(BYTE)0x49 },
    { 0x5c90,(BYTE)0x51,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x22,(BYTE)0x14,(BYTE)0x22,(BYTE)0x78,(BYTE)0x00,(BYTE)0x08,(BYTE)0x21,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x00,(BYTE)0x22,(BYTE)0xd6 },
    { 0x5ca0,(BYTE)0x00,(BYTE)0x08,(BYTE)0x24,(BYTE)0x4f,(BYTE)0x10,(BYTE)0x39,(BYTE)0xff,(BYTE)0xf0,(BYTE)0x00,(BYTE)0x01,(BYTE)0x50,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x22,(BYTE)0x14 },
    { 0x5cb0,(BYTE)0x51,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x06,(BYTE)0x0b,(BYTE)0x41,(BYTE)0xfa,(BYTE)0x06,(BYTE)0xc8,(BYTE)0x48,(BYTE)0xe7,(BYTE)0x00,(BYTE)0xe0,(BYTE)0x3f,(BYTE)0x3c },
    { 0x5cc0,(BYTE)0xff,(BYTE)0xff,(BYTE)0x3f,(BYTE)0x3c,(BYTE)0x00,(BYTE)0x40,(BYTE)0x4e,(BYTE)0x4e,(BYTE)0x58,(BYTE)0x4f,(BYTE)0x4c,(BYTE)0xdf,(BYTE)0x07,(BYTE)0x00,(BYTE)0x08,(BYTE)0x00 },
    { 0x5cd0,(BYTE)0x00,(BYTE)0x01,(BYTE)0x67,(BYTE)0x12,(BYTE)0x61,(BYTE)0x00,(BYTE)0x06,(BYTE)0x44,(BYTE)0x4a,(BYTE)0x40,(BYTE)0x66,(BYTE)0x0a,(BYTE)0x50,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00 },
    { 0x5ce0,(BYTE)0x06,(BYTE)0x0b,(BYTE)0x41,(BYTE)0xfa,(BYTE)0x06,(BYTE)0x80,(BYTE)0x48,(BYTE)0xe7,(BYTE)0x00,(BYTE)0x60,(BYTE)0x61,(BYTE)0x00,(BYTE)0x06,(BYTE)0x64,(BYTE)0x4c,(BYTE)0xdf },
    { 0x5cf0,(BYTE)0x06,(BYTE)0x00,(BYTE)0x21,(BYTE)0xc9,(BYTE)0x00,(BYTE)0x08,(BYTE)0x2e,(BYTE)0x4a,(BYTE)0x51,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x22,(BYTE)0x12,(BYTE)0x20,(BYTE)0x78 },
    { 0x5d00,(BYTE)0x00,(BYTE)0x08,(BYTE)0x21,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x00,(BYTE)0x22,(BYTE)0xfa,(BYTE)0x00,(BYTE)0x08,(BYTE)0x22,(BYTE)0x4f,(BYTE)0x10,(BYTE)0x38,(BYTE)0x87,(BYTE)0x81 },
    { 0x5d10,(BYTE)0x50,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x22,(BYTE)0x12,(BYTE)0x21,(BYTE)0xc8,(BYTE)0x00,(BYTE)0x08,(BYTE)0x2e,(BYTE)0x49,(BYTE)0x51,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00 },
    { 0x5d20,(BYTE)0x22,(BYTE)0x13,(BYTE)0x20,(BYTE)0x78,(BYTE)0x00,(BYTE)0x08,(BYTE)0x21,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x00,(BYTE)0x23,(BYTE)0x1e,(BYTE)0x00,(BYTE)0x08,(BYTE)0x22,(BYTE)0x4f },
    { 0x5d30,(BYTE)0x30,(BYTE)0x38,(BYTE)0x86,(BYTE)0x04,(BYTE)0x50,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x22,(BYTE)0x13,(BYTE)0x21,(BYTE)0xc8,(BYTE)0x00,(BYTE)0x08,(BYTE)0x2e,(BYTE)0x49 },
    { 0x5d40,(BYTE)0x32,(BYTE)0x3c,(BYTE)0x00,(BYTE)0x0d,(BYTE)0x70,(BYTE)0xff,(BYTE)0x41,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0xd6,(BYTE)0x10,(BYTE)0xc0,(BYTE)0x51,(BYTE)0xc9 },
    { 0x5d50,(BYTE)0xff,(BYTE)0xfc,(BYTE)0x70,(BYTE)0x00,(BYTE)0x33,(BYTE)0xc0,(BYTE)0x00,(BYTE)0x00,(BYTE)0x2a,(BYTE)0x62,(BYTE)0x23,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x04 },
    { 0x5d60,(BYTE)0x00,(BYTE)0x00,(BYTE)0x2a,(BYTE)0x64,(BYTE)0x33,(BYTE)0xc0,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0xd2,(BYTE)0x33,(BYTE)0xc0,(BYTE)0x00,(BYTE)0x00,(BYTE)0x01,(BYTE)0x70 },
    { 0x5d70,(BYTE)0x33,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x12,(BYTE)0x00,(BYTE)0x00,(BYTE)0x01,(BYTE)0x2e,(BYTE)0x23,(BYTE)0xf8,(BYTE)0x04,(BYTE)0xc6,(BYTE)0x00,(BYTE)0x00 },
    { 0x5d80,(BYTE)0x05,(BYTE)0xf4,(BYTE)0x06,(BYTE)0xb9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x02,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xf4,(BYTE)0x4a,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00 },
    { 0x5d90,(BYTE)0x22,(BYTE)0x14,(BYTE)0x67,(BYTE)0x28,(BYTE)0x4a,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x46,(BYTE)0x67,(BYTE)0x0a,(BYTE)0x08,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00 },
    { 0x5da0,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x49,(BYTE)0x67,(BYTE)0x16,(BYTE)0x20,(BYTE)0x7c,(BYTE)0x00,(BYTE)0x00,(BYTE)0x29,(BYTE)0xf6,(BYTE)0x61,(BYTE)0x00,(BYTE)0x05,(BYTE)0xa2 },
    { 0x5db0,(BYTE)0x33,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x10,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xa8,(BYTE)0x60,(BYTE)0x00,(BYTE)0x01,(BYTE)0x04,(BYTE)0x4a,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00 },
    { 0x5dc0,(BYTE)0x22,(BYTE)0x12,(BYTE)0x67,(BYTE)0x00,(BYTE)0x00,(BYTE)0x7a,(BYTE)0x33,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x07,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xa8,(BYTE)0x13,(BYTE)0xfc },
    { 0x5dd0,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x2c,(BYTE)0x6a,(BYTE)0x61,(BYTE)0x00,(BYTE)0xfd,(BYTE)0x24,(BYTE)0x4a,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x46 },
    { 0x5de0,(BYTE)0x67,(BYTE)0x0a,(BYTE)0x4a,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x48,(BYTE)0x67,(BYTE)0x00,(BYTE)0x00,(BYTE)0x54,(BYTE)0x20,(BYTE)0x7c,(BYTE)0x00,(BYTE)0x00 },
    { 0x5df0,(BYTE)0x2a,(BYTE)0x06,(BYTE)0x61,(BYTE)0x00,(BYTE)0x05,(BYTE)0x5c,(BYTE)0x30,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xa8,(BYTE)0x52,(BYTE)0x40,(BYTE)0x33,(BYTE)0xc0 },
    { 0x5e00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xa8,(BYTE)0x0c,(BYTE)0x40,(BYTE)0x00,(BYTE)0x0f,(BYTE)0x62,(BYTE)0x00,(BYTE)0x00,(BYTE)0x34,(BYTE)0x4a,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00 },
    { 0x5e10,(BYTE)0x00,(BYTE)0x46,(BYTE)0x67,(BYTE)0x20,(BYTE)0x02,(BYTE)0x00,(BYTE)0x00,(BYTE)0x07,(BYTE)0x01,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x48,(BYTE)0x66,(BYTE)0x00 },
    { 0x5e20,(BYTE)0x00,(BYTE)0x9c,(BYTE)0x52,(BYTE)0x40,(BYTE)0x0c,(BYTE)0x40,(BYTE)0x00,(BYTE)0x08,(BYTE)0x64,(BYTE)0x00,(BYTE)0x00,(BYTE)0x14,(BYTE)0x52,(BYTE)0x79,(BYTE)0x00,(BYTE)0x00 },
    { 0x5e30,(BYTE)0x05,(BYTE)0xa8,(BYTE)0x60,(BYTE)0xe4,(BYTE)0x4a,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00,(BYTE)0x2c,(BYTE)0x6a,(BYTE)0x67,(BYTE)0x00,(BYTE)0x00,(BYTE)0x80,(BYTE)0x4a,(BYTE)0x39 },
    { 0x5e40,(BYTE)0x00,(BYTE)0x00,(BYTE)0x22,(BYTE)0x13,(BYTE)0x67,(BYTE)0x00,(BYTE)0x02,(BYTE)0x2e,(BYTE)0x33,(BYTE)0xfc,(BYTE)0xff,(BYTE)0xff,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xa8 },
    { 0x5e50,(BYTE)0x13,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x2c,(BYTE)0x6a,(BYTE)0x4a,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x46,(BYTE)0x67,(BYTE)0x0a },
    { 0x5e60,(BYTE)0x4a,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x47,(BYTE)0x67,(BYTE)0x00,(BYTE)0x02,(BYTE)0x0c,(BYTE)0x20,(BYTE)0x7c,(BYTE)0x00,(BYTE)0x00,(BYTE)0x2a,(BYTE)0x16 },
    { 0x5e70,(BYTE)0x61,(BYTE)0x00,(BYTE)0x04,(BYTE)0xde,(BYTE)0x30,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xa8,(BYTE)0x52,(BYTE)0x40,(BYTE)0x33,(BYTE)0xc0,(BYTE)0x00,(BYTE)0x00 },
    { 0x5e80,(BYTE)0x05,(BYTE)0xa8,(BYTE)0x0c,(BYTE)0x40,(BYTE)0x00,(BYTE)0x07,(BYTE)0x62,(BYTE)0x00,(BYTE)0x01,(BYTE)0xec,(BYTE)0x4a,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x46 },
    { 0x5e90,(BYTE)0x67,(BYTE)0x20,(BYTE)0x02,(BYTE)0x00,(BYTE)0x00,(BYTE)0x07,(BYTE)0x01,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x47,(BYTE)0x66,(BYTE)0x00,(BYTE)0x00,(BYTE)0x1e },
    { 0x5ea0,(BYTE)0x52,(BYTE)0x40,(BYTE)0x0c,(BYTE)0x40,(BYTE)0x00,(BYTE)0x08,(BYTE)0x64,(BYTE)0x00,(BYTE)0x01,(BYTE)0xcc,(BYTE)0x52,(BYTE)0x79,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xa8 },
    { 0x5eb0,(BYTE)0x60,(BYTE)0xe4,(BYTE)0x4a,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00,(BYTE)0x2c,(BYTE)0x6a,(BYTE)0x66,(BYTE)0x00,(BYTE)0x01,(BYTE)0xba,(BYTE)0x76,(BYTE)0x01,(BYTE)0x3f,(BYTE)0x03 },
    { 0x5ec0,(BYTE)0x30,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xa8,(BYTE)0x02,(BYTE)0x40,(BYTE)0x00,(BYTE)0x07,(BYTE)0xd0,(BYTE)0x3c,(BYTE)0x00,(BYTE)0x30,(BYTE)0x13,(BYTE)0xc0 },
    { 0x5ed0,(BYTE)0x00,(BYTE)0x00,(BYTE)0x2a,(BYTE)0x2d,(BYTE)0x20,(BYTE)0x7c,(BYTE)0x00,(BYTE)0x00,(BYTE)0x2a,(BYTE)0x26,(BYTE)0x61,(BYTE)0x00,(BYTE)0x04,(BYTE)0x74,(BYTE)0x36,(BYTE)0x1f },
    { 0x5ee0,(BYTE)0x20,(BYTE)0x79,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xf4,(BYTE)0x42,(BYTE)0x90,(BYTE)0x2f,(BYTE)0x08,(BYTE)0x08,(BYTE)0x39,(BYTE)0x00,(BYTE)0x04,(BYTE)0x00,(BYTE)0x00 },
    { 0x5ef0,(BYTE)0x05,(BYTE)0xa9,(BYTE)0x67,(BYTE)0x00,(BYTE)0x00,(BYTE)0x4e,(BYTE)0x3f,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xa8,(BYTE)0x61,(BYTE)0x00,(BYTE)0xf6,(BYTE)0x72 },
    { 0x5f00,(BYTE)0x5c,(BYTE)0x4f,(BYTE)0x4a,(BYTE)0x40,(BYTE)0x66,(BYTE)0x00,(BYTE)0x01,(BYTE)0x38,(BYTE)0x20,(BYTE)0x79,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xf4,(BYTE)0x11,(BYTE)0x7c },
    { 0x5f10,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x5e,(BYTE)0xd1,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x36,(BYTE)0x61,(BYTE)0x00,(BYTE)0x04,(BYTE)0x34,(BYTE)0x20,(BYTE)0x7c },
    { 0x5f20,(BYTE)0x00,(BYTE)0x00,(BYTE)0x2a,(BYTE)0x40,(BYTE)0x61,(BYTE)0x00,(BYTE)0x04,(BYTE)0x2a,(BYTE)0x2f,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xf4,(BYTE)0x61,(BYTE)0x00 },
    { 0x5f30,(BYTE)0xf7,(BYTE)0x8e,(BYTE)0x58,(BYTE)0x8f,(BYTE)0x43,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x01,(BYTE)0x90,(BYTE)0x32,(BYTE)0xc1,(BYTE)0x32,(BYTE)0x82,(BYTE)0x60,(BYTE)0x00 },
    { 0x5f40,(BYTE)0x00,(BYTE)0x92,(BYTE)0x50,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x18,(BYTE)0x2f,(BYTE)0x3f,(BYTE)0x3c,(BYTE)0x00,(BYTE)0x30,(BYTE)0x3f,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00 },
    { 0x5f50,(BYTE)0x05,(BYTE)0xa8,(BYTE)0x61,(BYTE)0x00,(BYTE)0xf8,(BYTE)0xa0,(BYTE)0x50,(BYTE)0x8f,(BYTE)0x51,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x18,(BYTE)0x2f,(BYTE)0x4a,(BYTE)0x40 },
    { 0x5f60,(BYTE)0x6b,(BYTE)0x00,(BYTE)0x00,(BYTE)0xdc,(BYTE)0x57,(BYTE)0xcb,(BYTE)0xff,(BYTE)0x7a,(BYTE)0x67,(BYTE)0x16,(BYTE)0x20,(BYTE)0x7c,(BYTE)0x00,(BYTE)0x00,(BYTE)0x2a,(BYTE)0x44 },
    { 0x5f70,(BYTE)0x61,(BYTE)0x00,(BYTE)0x03,(BYTE)0xde,(BYTE)0xb0,(BYTE)0x7c,(BYTE)0x00,(BYTE)0x02,(BYTE)0x66,(BYTE)0x00,(BYTE)0x00,(BYTE)0x58,(BYTE)0x60,(BYTE)0x00,(BYTE)0x00,(BYTE)0xd2 },
    { 0x5f80,(BYTE)0x20,(BYTE)0x79,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xf4,(BYTE)0x11,(BYTE)0x7c,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x20,(BYTE)0x50,(BYTE)0x88,(BYTE)0x61,(BYTE)0x00 },
    { 0x5f90,(BYTE)0x03,(BYTE)0xc0,(BYTE)0x20,(BYTE)0x7c,(BYTE)0x00,(BYTE)0x00,(BYTE)0x2a,(BYTE)0x40,(BYTE)0x61,(BYTE)0x00,(BYTE)0x03,(BYTE)0xb6,(BYTE)0x20,(BYTE)0x79,(BYTE)0x00,(BYTE)0x00 },
    { 0x5fa0,(BYTE)0x05,(BYTE)0xf4,(BYTE)0x4a,(BYTE)0x10,(BYTE)0x66,(BYTE)0x00,(BYTE)0x00,(BYTE)0xaa,(BYTE)0x08,(BYTE)0x39,(BYTE)0x00,(BYTE)0x03,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xa9 },
    { 0x5fb0,(BYTE)0x66,(BYTE)0x10,(BYTE)0x32,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xa8,(BYTE)0x02,(BYTE)0x41,(BYTE)0x00,(BYTE)0x07,(BYTE)0x03,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00 },
    { 0x5fc0,(BYTE)0x05,(BYTE)0xba,(BYTE)0x08,(BYTE)0x28,(BYTE)0x00,(BYTE)0x07,(BYTE)0x00,(BYTE)0x01,(BYTE)0x67,(BYTE)0x08,(BYTE)0x08,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x06,(BYTE)0x00,(BYTE)0x00 },
    { 0x5fd0,(BYTE)0x05,(BYTE)0xa9,(BYTE)0x3f,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xa8,(BYTE)0x54,(BYTE)0x57,(BYTE)0x42,(BYTE)0x67,(BYTE)0x3f,(BYTE)0x3c,(BYTE)0x00,(BYTE)0x01 },
    { 0x5fe0,(BYTE)0x2f,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xf4,(BYTE)0x3f,(BYTE)0x3c,(BYTE)0x00,(BYTE)0x0a,(BYTE)0x61,(BYTE)0x00,(BYTE)0xe6,(BYTE)0xfa,(BYTE)0xde,(BYTE)0xfc },
    { 0x5ff0,(BYTE)0x00,(BYTE)0x0c,(BYTE)0x33,(BYTE)0xc0,(BYTE)0x00,(BYTE)0x00,(BYTE)0x2c,(BYTE)0x68,(BYTE)0x67,(BYTE)0x0a,(BYTE)0x08,(BYTE)0x39,(BYTE)0x00,(BYTE)0x06,(BYTE)0x00,(BYTE)0x00 },
    { 0x6000,(BYTE)0x05,(BYTE)0xa9,(BYTE)0x67,(BYTE)0x44,(BYTE)0x52,(BYTE)0x79,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0xd2,(BYTE)0x08,(BYTE)0x39,(BYTE)0x00,(BYTE)0x04,(BYTE)0x00,(BYTE)0x00 },
    { 0x6010,(BYTE)0x05,(BYTE)0xa9,(BYTE)0x66,(BYTE)0x18,(BYTE)0x08,(BYTE)0x39,(BYTE)0x00,(BYTE)0x03,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xa9,(BYTE)0x67,(BYTE)0x08,(BYTE)0x52,(BYTE)0x79 },
    { 0x6020,(BYTE)0x00,(BYTE)0x00,(BYTE)0x01,(BYTE)0x72,(BYTE)0x60,(BYTE)0x06,(BYTE)0x52,(BYTE)0x79,(BYTE)0x00,(BYTE)0x00,(BYTE)0x01,(BYTE)0x70,(BYTE)0x61,(BYTE)0x00,(BYTE)0x00,(BYTE)0x8e },
    { 0x6030,(BYTE)0x4a,(BYTE)0x40,(BYTE)0x67,(BYTE)0x1c,(BYTE)0x0c,(BYTE)0x40,(BYTE)0xff,(BYTE)0xf2,(BYTE)0x67,(BYTE)0x00,(BYTE)0xfe,(BYTE)0x82,(BYTE)0x60,(BYTE)0x12,(BYTE)0x20,(BYTE)0x7c },
    { 0x6040,(BYTE)0x00,(BYTE)0x00,(BYTE)0x2a,(BYTE)0x32,(BYTE)0x61,(BYTE)0x00,(BYTE)0x03,(BYTE)0x0a,(BYTE)0x13,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x01,(BYTE)0x00,(BYTE)0x00,(BYTE)0x2c,(BYTE)0x6a },
    { 0x6050,(BYTE)0x02,(BYTE)0x79,(BYTE)0x00,(BYTE)0x1f,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xa8,(BYTE)0x08,(BYTE)0x39,(BYTE)0x00,(BYTE)0x04,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xa9 },
    { 0x6060,(BYTE)0x66,(BYTE)0x00,(BYTE)0xfd,(BYTE)0x94,(BYTE)0x08,(BYTE)0x39,(BYTE)0x00,(BYTE)0x03,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xa9,(BYTE)0x66,(BYTE)0x00,(BYTE)0xfd,(BYTE)0x88 },
    { 0x6070,(BYTE)0x60,(BYTE)0x00,(BYTE)0xfe,(BYTE)0x02,(BYTE)0x9b,(BYTE)0xcd,(BYTE)0x23,(BYTE)0xed,(BYTE)0x04,(BYTE)0x72,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xc4,(BYTE)0x23,(BYTE)0xed },
    { 0x6080,(BYTE)0x04,(BYTE)0x76,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xc8,(BYTE)0x23,(BYTE)0xed,(BYTE)0x04,(BYTE)0x7e,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xcc,(BYTE)0x2b,(BYTE)0x7c },
    { 0x6090,(BYTE)0x00,(BYTE)0x00,(BYTE)0x06,(BYTE)0x1a,(BYTE)0x04,(BYTE)0x72,(BYTE)0x2b,(BYTE)0x7c,(BYTE)0x00,(BYTE)0x00,(BYTE)0x06,(BYTE)0x2c,(BYTE)0x04,(BYTE)0x76,(BYTE)0x2b,(BYTE)0x7c },
    { 0x60a0,(BYTE)0x00,(BYTE)0x00,(BYTE)0x06,(BYTE)0x40,(BYTE)0x04,(BYTE)0x7e,(BYTE)0x2b,(BYTE)0x7c,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0xd2,(BYTE)0x05,(BYTE)0x16,(BYTE)0x23,(BYTE)0xfc },
    { 0x60b0,(BYTE)0x00,(BYTE)0x00,(BYTE)0x01,(BYTE)0x24,(BYTE)0x00,(BYTE)0x00,(BYTE)0x01,(BYTE)0x28,(BYTE)0x60,(BYTE)0x00,(BYTE)0xec,(BYTE)0xc4,(BYTE)0x33,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x00 },
    { 0x60c0,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xaa,(BYTE)0x08,(BYTE)0x39,(BYTE)0x00,(BYTE)0x06,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xa9,(BYTE)0x67,(BYTE)0x08,(BYTE)0x4a,(BYTE)0x79 },
    { 0x60d0,(BYTE)0x00,(BYTE)0x00,(BYTE)0x2c,(BYTE)0x68,(BYTE)0x66,(BYTE)0x54,(BYTE)0x72,(BYTE)0x03,(BYTE)0x20,(BYTE)0x79,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xf4,(BYTE)0x0c,(BYTE)0x68 },
    { 0x60e0,(BYTE)0x55,(BYTE)0xaa,(BYTE)0x01,(BYTE)0xfe,(BYTE)0x66,(BYTE)0x06,(BYTE)0x61,(BYTE)0x00,(BYTE)0x00,(BYTE)0x78,(BYTE)0x60,(BYTE)0x04,(BYTE)0x61,(BYTE)0x00,(BYTE)0x00,(BYTE)0xe0 },
    { 0x60f0,(BYTE)0x4a,(BYTE)0x40,(BYTE)0x66,(BYTE)0x6a,(BYTE)0x08,(BYTE)0x39,(BYTE)0x00,(BYTE)0x06,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xa9,(BYTE)0x67,(BYTE)0x60,(BYTE)0x32,(BYTE)0x39 },
    { 0x6100,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xa8,(BYTE)0x08,(BYTE)0x01,(BYTE)0x00,(BYTE)0x03,(BYTE)0x66,(BYTE)0x08,(BYTE)0x41,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x16 },
    { 0x6110,(BYTE)0x60,(BYTE)0x06,(BYTE)0x41,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x24,(BYTE)0x02,(BYTE)0x41,(BYTE)0x00,(BYTE)0x07,(BYTE)0x12,(BYTE)0x30,(BYTE)0x10,(BYTE)0x00 },
    { 0x6120,(BYTE)0x92,(BYTE)0x79,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xaa,(BYTE)0x63,(BYTE)0x36,(BYTE)0x60,(BYTE)0x22,(BYTE)0x32,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xa8 },
    { 0x6130,(BYTE)0x08,(BYTE)0x01,(BYTE)0x00,(BYTE)0x03,(BYTE)0x66,(BYTE)0x08,(BYTE)0x41,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x16,(BYTE)0x60,(BYTE)0x06,(BYTE)0x41,(BYTE)0xf9 },
    { 0x6140,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x24,(BYTE)0x02,(BYTE)0x41,(BYTE)0x00,(BYTE)0x07,(BYTE)0x12,(BYTE)0x30,(BYTE)0x10,(BYTE)0x00,(BYTE)0x53,(BYTE)0x41,(BYTE)0x3f,(BYTE)0x01 },
    { 0x6150,(BYTE)0x61,(BYTE)0x00,(BYTE)0x01,(BYTE)0x4e,(BYTE)0x32,(BYTE)0x1f,(BYTE)0x4a,(BYTE)0x40,(BYTE)0x6b,(BYTE)0x04,(BYTE)0x51,(BYTE)0xc9,(BYTE)0xff,(BYTE)0xf2,(BYTE)0x4e,(BYTE)0x75 },
    { 0x6160,(BYTE)0xd0,(BYTE)0xfc,(BYTE)0x01,(BYTE)0xbe,(BYTE)0x48,(BYTE)0xe7,(BYTE)0x40,(BYTE)0x80,(BYTE)0x51,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x06,(BYTE)0x0c,(BYTE)0x61,(BYTE)0x00 },
    { 0x6170,(BYTE)0xe0,(BYTE)0xd6,(BYTE)0x4a,(BYTE)0x00,(BYTE)0x67,(BYTE)0x48,(BYTE)0x0c,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0x66,(BYTE)0x26,(BYTE)0x50,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00 },
    { 0x6180,(BYTE)0x06,(BYTE)0x0c,(BYTE)0x23,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x06,(BYTE)0x12,(BYTE)0x23,(BYTE)0xc1,(BYTE)0x00,(BYTE)0x00 },
    { 0x6190,(BYTE)0x06,(BYTE)0x0e,(BYTE)0x61,(BYTE)0x00,(BYTE)0xe0,(BYTE)0xe0,(BYTE)0x4a,(BYTE)0x00,(BYTE)0x67,(BYTE)0x24,(BYTE)0x6b,(BYTE)0x16,(BYTE)0x0c,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05 },
    { 0x61a0,(BYTE)0x67,(BYTE)0xf0,(BYTE)0x48,(BYTE)0xe7,(BYTE)0x60,(BYTE)0x80,(BYTE)0x61,(BYTE)0x00,(BYTE)0x00,(BYTE)0x98,(BYTE)0x4c,(BYTE)0xdf,(BYTE)0x01,(BYTE)0x06,(BYTE)0x4a,(BYTE)0x40 },
    { 0x61b0,(BYTE)0x67,(BYTE)0x04,(BYTE)0x50,(BYTE)0x8f,(BYTE)0x60,(BYTE)0x16,(BYTE)0x4a,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00,(BYTE)0x06,(BYTE)0x0c,(BYTE)0x66,(BYTE)0xd4,(BYTE)0x4c,(BYTE)0xdf },
    { 0x61c0,(BYTE)0x01,(BYTE)0x02,(BYTE)0xd0,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x10,(BYTE)0x51,(BYTE)0xc9,(BYTE)0xff,(BYTE)0x9c,(BYTE)0x70,(BYTE)0x00,(BYTE)0x4e,(BYTE)0x75,(BYTE)0xd0,(BYTE)0xfc },
    { 0x61d0,(BYTE)0x01,(BYTE)0xc2,(BYTE)0x4a,(BYTE)0x98,(BYTE)0x67,(BYTE)0x66,(BYTE)0x48,(BYTE)0xe7,(BYTE)0x40,(BYTE)0x80,(BYTE)0x51,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x06,(BYTE)0x0c },
    { 0x61e0,(BYTE)0x61,(BYTE)0x00,(BYTE)0xe1,(BYTE)0x8a,(BYTE)0x4a,(BYTE)0x00,(BYTE)0x67,(BYTE)0x48,(BYTE)0x0c,(BYTE)0x00,(BYTE)0x00,(BYTE)0x58,(BYTE)0x66,(BYTE)0x26,(BYTE)0x50,(BYTE)0xf9 },
    { 0x61f0,(BYTE)0x00,(BYTE)0x00,(BYTE)0x06,(BYTE)0x0c,(BYTE)0x23,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x06,(BYTE)0x12,(BYTE)0x23,(BYTE)0xc1 },
    { 0x6200,(BYTE)0x00,(BYTE)0x00,(BYTE)0x06,(BYTE)0x0e,(BYTE)0x61,(BYTE)0x00,(BYTE)0xe1,(BYTE)0xc6,(BYTE)0x4a,(BYTE)0x00,(BYTE)0x67,(BYTE)0x24,(BYTE)0x6b,(BYTE)0x16,(BYTE)0x0c,(BYTE)0x00 },
    { 0x6210,(BYTE)0x00,(BYTE)0x58,(BYTE)0x67,(BYTE)0xf0,(BYTE)0x48,(BYTE)0xe7,(BYTE)0x60,(BYTE)0x80,(BYTE)0x61,(BYTE)0x00,(BYTE)0x00,(BYTE)0x26,(BYTE)0x4c,(BYTE)0xdf,(BYTE)0x01,(BYTE)0x06 },
    { 0x6220,(BYTE)0x4a,(BYTE)0x40,(BYTE)0x67,(BYTE)0x04,(BYTE)0x50,(BYTE)0x8f,(BYTE)0x60,(BYTE)0x16,(BYTE)0x4a,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00,(BYTE)0x06,(BYTE)0x0c,(BYTE)0x66,(BYTE)0xd4 },
    { 0x6230,(BYTE)0x4c,(BYTE)0xdf,(BYTE)0x01,(BYTE)0x02,(BYTE)0xd0,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x0c,(BYTE)0x51,(BYTE)0xc9,(BYTE)0xff,(BYTE)0x9c,(BYTE)0x70,(BYTE)0x00,(BYTE)0x4e,(BYTE)0x75 },
    { 0x6240,(BYTE)0x0c,(BYTE)0x79,(BYTE)0x00,(BYTE)0x0e,(BYTE)0x00,(BYTE)0x00,(BYTE)0x2a,(BYTE)0x62,(BYTE)0x6c,(BYTE)0x00,(BYTE)0x00,(BYTE)0xcc,(BYTE)0x2f,(BYTE)0x01,(BYTE)0x3f,(BYTE)0x39 },
    { 0x6250,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xa8,(BYTE)0x54,(BYTE)0x57,(BYTE)0x3f,(BYTE)0x3c,(BYTE)0xff,(BYTE)0xff,(BYTE)0x3f,(BYTE)0x3c,(BYTE)0x00,(BYTE)0x01,(BYTE)0x2f,(BYTE)0x3c },
    { 0x6260,(BYTE)0x00,(BYTE)0x00,(BYTE)0x2a,(BYTE)0x68,(BYTE)0x3f,(BYTE)0x3c,(BYTE)0x00,(BYTE)0x0a,(BYTE)0x61,(BYTE)0x00,(BYTE)0xe4,(BYTE)0x7c,(BYTE)0xde,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x10 },
    { 0x6270,(BYTE)0x4a,(BYTE)0x40,(BYTE)0x66,(BYTE)0x00,(BYTE)0x00,(BYTE)0xa4,(BYTE)0x20,(BYTE)0x7c,(BYTE)0x00,(BYTE)0x00,(BYTE)0x2a,(BYTE)0x68,(BYTE)0x70,(BYTE)0x0b,(BYTE)0x61,(BYTE)0x00 },
    { 0x6280,(BYTE)0xe4,(BYTE)0x14,(BYTE)0x4a,(BYTE)0x40,(BYTE)0x67,(BYTE)0x26,(BYTE)0xb0,(BYTE)0x79,(BYTE)0x00,(BYTE)0x00,(BYTE)0x01,(BYTE)0x2e,(BYTE)0x63,(BYTE)0x06,(BYTE)0x33,(BYTE)0xc0 },
    { 0x6290,(BYTE)0x00,(BYTE)0x00,(BYTE)0x01,(BYTE)0x2e,(BYTE)0x72,(BYTE)0x09,(BYTE)0xe2,(BYTE)0x68,(BYTE)0x33,(BYTE)0xc0,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xfc,(BYTE)0x60,(BYTE)0x14 },
    { 0x62a0,(BYTE)0x0c,(BYTE)0x79,(BYTE)0x00,(BYTE)0x0e,(BYTE)0x00,(BYTE)0x00,(BYTE)0x2a,(BYTE)0x62,(BYTE)0x6c,(BYTE)0x00,(BYTE)0x00,(BYTE)0x6c,(BYTE)0x33,(BYTE)0xfc,(BYTE)0x00,(BYTE)0x01 },
    { 0x62b0,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xfc,(BYTE)0x22,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00,(BYTE)0x2a,(BYTE)0x64,(BYTE)0x20,(BYTE)0x38,(BYTE)0x04,(BYTE)0xc2,(BYTE)0x80,(BYTE)0x81 },
    { 0x62c0,(BYTE)0x21,(BYTE)0xc0,(BYTE)0x04,(BYTE)0xc2,(BYTE)0xe3,(BYTE)0x81,(BYTE)0x23,(BYTE)0xc1,(BYTE)0x00,(BYTE)0x00,(BYTE)0x2a,(BYTE)0x64,(BYTE)0x30,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00 },
    { 0x62d0,(BYTE)0x2a,(BYTE)0x62,(BYTE)0x41,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0xd6,(BYTE)0x11,(BYTE)0xb9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xa9,(BYTE)0x00,(BYTE)0x00 },
    { 0x62e0,(BYTE)0x41,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x01,(BYTE)0xd0,(BYTE)0x11,(BYTE)0xb9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x05,(BYTE)0xfd,(BYTE)0x00,(BYTE)0x00,(BYTE)0x41,(BYTE)0xf9 },
    { 0x62f0,(BYTE)0x00,(BYTE)0x00,(BYTE)0x01,(BYTE)0x82,(BYTE)0x11,(BYTE)0xbc,(BYTE)0x00,(BYTE)0x01,(BYTE)0x00,(BYTE)0x00,(BYTE)0x41,(BYTE)0xf9,(BYTE)0x00,(BYTE)0x00,(BYTE)0x01,(BYTE)0x74 },
    { 0x6300,(BYTE)0x11,(BYTE)0xbc,(BYTE)0x00,(BYTE)0x02,(BYTE)0x00,(BYTE)0x00,(BYTE)0x52,(BYTE)0x79,(BYTE)0x00,(BYTE)0x00,(BYTE)0x2a,(BYTE)0x62,(BYTE)0x52,(BYTE)0x79,(BYTE)0x00,(BYTE)0x00 },
    { 0x6310,(BYTE)0x05,(BYTE)0xaa,(BYTE)0x70,(BYTE)0x00,(BYTE)0x60,(BYTE)0x02,(BYTE)0x70,(BYTE)0xff,(BYTE)0x4e,(BYTE)0x75,(BYTE)0x24,(BYTE)0x38,(BYTE)0x82,(BYTE)0x42,(BYTE)0x31,(BYTE)0xfc },
    { 0x6320,(BYTE)0x05,(BYTE)0x55,(BYTE)0x82,(BYTE)0x42,(BYTE)0x31,(BYTE)0xfc,(BYTE)0x02,(BYTE)0x22,(BYTE)0x82,(BYTE)0x44,(BYTE)0x70,(BYTE)0x01,(BYTE)0x32,(BYTE)0x38,(BYTE)0x82,(BYTE)0x42 },
    { 0x6330,(BYTE)0xc2,(BYTE)0x7c,(BYTE)0x07,(BYTE)0x77,(BYTE)0xb2,(BYTE)0x7c,(BYTE)0x05,(BYTE)0x55,(BYTE)0x66,(BYTE)0x14,(BYTE)0x32,(BYTE)0x38,(BYTE)0x82,(BYTE)0x44,(BYTE)0xc2,(BYTE)0x7c },
    { 0x6340,(BYTE)0x07,(BYTE)0x77,(BYTE)0xb2,(BYTE)0x7c,(BYTE)0x02,(BYTE)0x22,(BYTE)0x66,(BYTE)0x06,(BYTE)0x70,(BYTE)0x00,(BYTE)0x21,(BYTE)0xc2,(BYTE)0x82,(BYTE)0x42,(BYTE)0x4e,(BYTE)0x75 },
    { 0x6350,(BYTE)0x4a,(BYTE)0x39,(BYTE)0x00,(BYTE)0x00,(BYTE)0x06,(BYTE)0x18,(BYTE)0x66,(BYTE)0x0a,(BYTE)0x2f,(BYTE)0x08,(BYTE)0x3f,(BYTE)0x3c,(BYTE)0x00,(BYTE)0x09,(BYTE)0x4e,(BYTE)0x41 },
    { 0x6360,(BYTE)0x5c,(BYTE)0x8f,(BYTE)0x4e,(BYTE)0x75,(BYTE)0x55,(BYTE)0x73,(BYTE)0x69,(BYTE)0x6e,(BYTE)0x67,(BYTE)0x20,(BYTE)0x42,(BYTE)0x4c,(BYTE)0x69,(BYTE)0x54,(BYTE)0x54,(BYTE)0x45 },
    { 0x6370,(BYTE)0x52,(BYTE)0x20,(BYTE)0x66,(BYTE)0x6f,(BYTE)0x72,(BYTE)0x20,(BYTE)0x49,(BYTE)0x44,(BYTE)0x45,(BYTE)0x20,(BYTE)0x49,(BYTE)0x2f,(BYTE)0x4f,(BYTE)0x0d,(BYTE)0x0a,(BYTE)0x00 },
    { 0x6380,(BYTE)0x44,(BYTE)0x6f,(BYTE)0x69,(BYTE)0x6e,(BYTE)0x67,(BYTE)0x20,(BYTE)0x70,(BYTE)0x72,(BYTE)0x6f,(BYTE)0x67,(BYTE)0x72,(BYTE)0x61,(BYTE)0x6d,(BYTE)0x6d,(BYTE)0x65,(BYTE)0x64 },
    { 0x6390,(BYTE)0x20,(BYTE)0x49,(BYTE)0x2f,(BYTE)0x4f,(BYTE)0x20,(BYTE)0x66,(BYTE)0x6f,(BYTE)0x72,(BYTE)0x20,(BYTE)0x49,(BYTE)0x44,(BYTE)0x45,(BYTE)0x0d,(BYTE)0x0a,(BYTE)0x00,(BYTE)0x00 },
    { 0x63a0,(BYTE)0x4c,(BYTE)0x6f,(BYTE)0x61,(BYTE)0x64,(BYTE)0x69,(BYTE)0x6e,(BYTE)0x67,(BYTE)0x2e,(BYTE)0x2e,(BYTE)0x2e,(BYTE)0x0d,(BYTE)0x0a,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d },
    { 0x63b0,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d },
    { 0x63c0,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x0d,(BYTE)0x0a,(BYTE)0x41,(BYTE)0x74,(BYTE)0x61,(BYTE)0x72,(BYTE)0x69,(BYTE)0x20,(BYTE)0x48,(BYTE)0x61,(BYTE)0x72,(BYTE)0x64,(BYTE)0x20 },
    { 0x63d0,(BYTE)0x44,(BYTE)0x69,(BYTE)0x73,(BYTE)0x6b,(BYTE)0x20,(BYTE)0x44,(BYTE)0x72,(BYTE)0x69,(BYTE)0x76,(BYTE)0x65,(BYTE)0x72,(BYTE)0x20,(BYTE)0x0d,(BYTE)0x0a,(BYTE)0x41,(BYTE)0x48 },
    { 0x63e0,(BYTE)0x44,(BYTE)0x49,(BYTE)0x20,(BYTE)0x76,(BYTE)0x35,(BYTE)0x2e,(BYTE)0x30,(BYTE)0x30,(BYTE)0x20,(BYTE)0x44,(BYTE)0x65,(BYTE)0x63,(BYTE)0x20,(BYTE)0x30,(BYTE)0x33,(BYTE)0x20 },
    { 0x63f0,(BYTE)0x31,(BYTE)0x39,(BYTE)0x39,(BYTE)0x31,(BYTE)0x20,(BYTE)0x0d,(BYTE)0x0a,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d },
    { 0x6400,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x2d,(BYTE)0x0d,(BYTE)0x0a },
    { 0x6410,(BYTE)0x00,(BYTE)0x00,(BYTE)0x49,(BYTE)0x44,(BYTE)0x45,(BYTE)0x20,(BYTE)0x64,(BYTE)0x65,(BYTE)0x76,(BYTE)0x69,(BYTE)0x63,(BYTE)0x65,(BYTE)0x73,(BYTE)0x3a,(BYTE)0x0d,(BYTE)0x0a },
    { 0x6420,(BYTE)0x00,(BYTE)0x00,(BYTE)0x53,(BYTE)0x43,(BYTE)0x53,(BYTE)0x49,(BYTE)0x20,(BYTE)0x64,(BYTE)0x65,(BYTE)0x76,(BYTE)0x69,(BYTE)0x63,(BYTE)0x65,(BYTE)0x73,(BYTE)0x3a,(BYTE)0x0d },
    { 0x6430,(BYTE)0x0a,(BYTE)0x00,(BYTE)0x41,(BYTE)0x43,(BYTE)0x53,(BYTE)0x49,(BYTE)0x20,(BYTE)0x64,(BYTE)0x65,(BYTE)0x76,(BYTE)0x69,(BYTE)0x63,(BYTE)0x65,(BYTE)0x73,(BYTE)0x3a,(BYTE)0x0d },
    { 0x6440,(BYTE)0x0a,(BYTE)0x00,(BYTE)0x20,(BYTE)0x20,(BYTE)0x55,(BYTE)0x6e,(BYTE)0x69,(BYTE)0x74,(BYTE)0x20,(BYTE)0x30,(BYTE)0x20,(BYTE)0x2d,(BYTE)0x20,(BYTE)0x00,(BYTE)0x4e,(BYTE)0x6f },
    { 0x6450,(BYTE)0x74,(BYTE)0x20,(BYTE)0x72,(BYTE)0x65,(BYTE)0x73,(BYTE)0x70,(BYTE)0x6f,(BYTE)0x6e,(BYTE)0x64,(BYTE)0x69,(BYTE)0x6e,(BYTE)0x67,(BYTE)0x0d,(BYTE)0x0a,(BYTE)0x00,(BYTE)0x00 },
    { 0x6460,(BYTE)0x49,(BYTE)0x64,(BYTE)0x65,(BYTE)0x6e,(BYTE)0x74,(BYTE)0x69,(BYTE)0x66,(BYTE)0x69,(BYTE)0x63,(BYTE)0x61,(BYTE)0x74,(BYTE)0x69,(BYTE)0x6f,(BYTE)0x6e,(BYTE)0x20,(BYTE)0x75 },
    { 0x6470,(BYTE)0x6e,(BYTE)0x61,(BYTE)0x76,(BYTE)0x61,(BYTE)0x69,(BYTE)0x6c,(BYTE)0x61,(BYTE)0x62,(BYTE)0x6c,(BYTE)0x65,(BYTE)0x0d,(BYTE)0x0a,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00 },
    { 0x6480,(BYTE)0x00,(BYTE)0x52,(BYTE)0x16,(BYTE)0x0c,(BYTE)0x01,(BYTE)0x01,(BYTE)0x01,(BYTE)0x01,(BYTE)0x01,(BYTE)0xb8,(BYTE)0x14,(BYTE)0x12,(BYTE)0x14,(BYTE)0x0e,(BYTE)0x0e,(BYTE)0x06 },
    { 0x6490,(BYTE)0x06,(BYTE)0x12,(BYTE)0x18,(BYTE)0x0c,(BYTE)0x06,(BYTE)0x2a,(BYTE)0x0a,(BYTE)0x08,(BYTE)0x0e,(BYTE)0x30,(BYTE)0x08,(BYTE)0x0a,(BYTE)0x0c,(BYTE)0x16,(BYTE)0x12,(BYTE)0x0e },
    { 0x64a0,(BYTE)0x2a,(BYTE)0x14,(BYTE)0x0a,(BYTE)0x06,(BYTE)0x16,(BYTE)0x0c,(BYTE)0x0e,(BYTE)0x08,(BYTE)0x4e,(BYTE)0x6c,(BYTE)0x0c,(BYTE)0x06,(BYTE)0x14,(BYTE)0x14,(BYTE)0x0a,(BYTE)0x06 },
    { 0x64b0,(BYTE)0x16,(BYTE)0x0c,(BYTE)0x80,(BYTE)0x90,(BYTE)0x0c,(BYTE)0x06,(BYTE)0x10,(BYTE)0x3c,(BYTE)0x0e,(BYTE)0x16,(BYTE)0x44,(BYTE)0x12,(BYTE)0x2a,(BYTE)0x06,(BYTE)0x06,(BYTE)0x10 },
    { 0x64c0,(BYTE)0x0a,(BYTE)0x1a,(BYTE)0x10,(BYTE)0x0e,(BYTE)0x08,(BYTE)0x08,(BYTE)0x0c,(BYTE)0x5c,(BYTE)0x0a,(BYTE)0x0c,(BYTE)0x08,(BYTE)0x08,(BYTE)0x06,(BYTE)0x06,(BYTE)0x0c,(BYTE)0x0e },
    { 0x64d0,(BYTE)0x22,(BYTE)0x10,(BYTE)0x0c,(BYTE)0x06,(BYTE)0x06,(BYTE)0x06,(BYTE)0x06,(BYTE)0x1c,(BYTE)0x0a,(BYTE)0x08,(BYTE)0x06,(BYTE)0x08,(BYTE)0x06,(BYTE)0x0e,(BYTE)0x10,(BYTE)0x16 },
    { 0x64e0,(BYTE)0x06,(BYTE)0x38,(BYTE)0x2a,(BYTE)0x0c,(BYTE)0x08,(BYTE)0x06,(BYTE)0x0a,(BYTE)0x06,(BYTE)0x06,(BYTE)0x06,(BYTE)0x0c,(BYTE)0x08,(BYTE)0x0c,(BYTE)0x5a,(BYTE)0x2a,(BYTE)0x12 },
    { 0x64f0,(BYTE)0x08,(BYTE)0x0e,(BYTE)0x06,(BYTE)0x06,(BYTE)0x48,(BYTE)0x12,(BYTE)0x08,(BYTE)0x0e,(BYTE)0x12,(BYTE)0x0a,(BYTE)0x0c,(BYTE)0x0a,(BYTE)0x06,(BYTE)0x10,(BYTE)0x0c,(BYTE)0x12 },
    { 0x6500,(BYTE)0x08,(BYTE)0x08,(BYTE)0x08,(BYTE)0x22,(BYTE)0x0e,(BYTE)0x18,(BYTE)0x52,(BYTE)0x0c,(BYTE)0x22,(BYTE)0x50,(BYTE)0x0a,(BYTE)0x28,(BYTE)0x78,(BYTE)0x04,(BYTE)0x10,(BYTE)0x0a },
    { 0x6510,(BYTE)0x10,(BYTE)0x3a,(BYTE)0x70,(BYTE)0x30,(BYTE)0x0c,(BYTE)0x4c,(BYTE)0x10,(BYTE)0x0c,(BYTE)0x08,(BYTE)0x08,(BYTE)0x08,(BYTE)0x0e,(BYTE)0x0a,(BYTE)0x1a,(BYTE)0x06,(BYTE)0x06 },
    { 0x6520,(BYTE)0x06,(BYTE)0x2c,(BYTE)0x08,(BYTE)0x04,(BYTE)0x06,(BYTE)0x10,(BYTE)0x04,(BYTE)0x06,(BYTE)0x04,(BYTE)0x08,(BYTE)0x10,(BYTE)0x06,(BYTE)0x08,(BYTE)0x0a,(BYTE)0x08,(BYTE)0x06 },
    { 0x6530,(BYTE)0x20,(BYTE)0x08,(BYTE)0x06,(BYTE)0x08,(BYTE)0x0e,(BYTE)0x0e,(BYTE)0x1a,(BYTE)0x0e,(BYTE)0x12,(BYTE)0x1e,(BYTE)0x1e,(BYTE)0x8c,(BYTE)0x0a,(BYTE)0x06,(BYTE)0x1e,(BYTE)0x0a },
    { 0x6540,(BYTE)0x10,(BYTE)0x0a,(BYTE)0x8e,(BYTE)0x0c,(BYTE)0x06,(BYTE)0x08,(BYTE)0x06,(BYTE)0x08,(BYTE)0x01,(BYTE)0x01,(BYTE)0x12,(BYTE)0x36,(BYTE)0x2a,(BYTE)0x06,(BYTE)0x16,(BYTE)0x88 },
    { 0x6550,(BYTE)0x7e,(BYTE)0xca,(BYTE)0x4c,(BYTE)0x96,(BYTE)0x58,(BYTE)0x3c,(BYTE)0xb2,(BYTE)0x78,(BYTE)0x1a,(BYTE)0x2c,(BYTE)0x36,(BYTE)0x3a,(BYTE)0x26,(BYTE)0x36,(BYTE)0x36,(BYTE)0x10 },
    { 0x6560,(BYTE)0x22,(BYTE)0x32,(BYTE)0x08,(BYTE)0x01,(BYTE)0x8a,(BYTE)0x9c,(BYTE)0x98,(BYTE)0x52,(BYTE)0x10,(BYTE)0x10,(BYTE)0x0e,(BYTE)0x06,(BYTE)0x0a,(BYTE)0x12,(BYTE)0x06,(BYTE)0x12 },
    { 0x6570,(BYTE)0x0a,(BYTE)0x0e,(BYTE)0x0c,(BYTE)0x0a,(BYTE)0x10,(BYTE)0x06,(BYTE)0x2c,(BYTE)0x1c,(BYTE)0x0a,(BYTE)0x0e,(BYTE)0x0c,(BYTE)0x0a,(BYTE)0x0e,(BYTE)0x12,(BYTE)0x0e,(BYTE)0x0a },
    { 0x6580,(BYTE)0x06,(BYTE)0x06,(BYTE)0x06,(BYTE)0x04,(BYTE)0x08,(BYTE)0x0a,(BYTE)0x06,(BYTE)0x08,(BYTE)0x0a,(BYTE)0x08,(BYTE)0x0c,(BYTE)0x0a,(BYTE)0x0c,(BYTE)0x08,(BYTE)0x0a,(BYTE)0x08 },
    { 0x6590,(BYTE)0x0a,(BYTE)0x0a,(BYTE)0x08,(BYTE)0x0e,(BYTE)0x0c,(BYTE)0x14,(BYTE)0x08,(BYTE)0x0a,(BYTE)0x0c,(BYTE)0x08,(BYTE)0x06,(BYTE)0x08,(BYTE)0x0a,(BYTE)0x0a,(BYTE)0x08,(BYTE)0x0e },
    { 0x65a0,(BYTE)0x0c,(BYTE)0x14,(BYTE)0x08,(BYTE)0x0e,(BYTE)0x0e,(BYTE)0x06,(BYTE)0x0c,(BYTE)0x0c,(BYTE)0x0a,(BYTE)0x12,(BYTE)0x16,(BYTE)0x0a,(BYTE)0x0c,(BYTE)0x0e,(BYTE)0x0a,(BYTE)0x0c },
    { 0x65b0,(BYTE)0x12,(BYTE)0x16,(BYTE)0x12,(BYTE)0x0a,(BYTE)0x0e,(BYTE)0x08,(BYTE)0x0a,(BYTE)0x10,(BYTE)0x06,(BYTE)0x0e,(BYTE)0x12,(BYTE)0x0a,(BYTE)0x08,(BYTE)0x08,(BYTE)0x0a,(BYTE)0x08 },
    { 0x65c0,(BYTE)0x08,(BYTE)0x18,(BYTE)0x0c,(BYTE)0x08,(BYTE)0x08,(BYTE)0x0c,(BYTE)0x12,(BYTE)0x08,(BYTE)0x08,(BYTE)0x06,(BYTE)0x08,(BYTE)0x08,(BYTE)0x08,(BYTE)0x08,(BYTE)0x04,(BYTE)0x0c },
    { 0x65d0,(BYTE)0x08,(BYTE)0x08,(BYTE)0x0a,(BYTE)0x1e,(BYTE)0x08,(BYTE)0x0c,(BYTE)0x08,(BYTE)0x0e,(BYTE)0x0a,(BYTE)0x0c,(BYTE)0x08,(BYTE)0x2a,(BYTE)0x14,(BYTE)0x0a,(BYTE)0x06,(BYTE)0x2a },
    { 0x65e0,(BYTE)0x24,(BYTE)0x14,(BYTE)0x0a,(BYTE)0x06,(BYTE)0x2a,(BYTE)0x1a,(BYTE)0x0c,(BYTE)0x10,(BYTE)0x18,(BYTE)0x10,(BYTE)0x08,(BYTE)0x0a,(BYTE)0x0a,(BYTE)0x0c,(BYTE)0x06,(BYTE)0x12 },
    { 0x65f0,(BYTE)0x06,(BYTE)0x06,(BYTE)0x06,(BYTE)0x08,(BYTE)0x06,(BYTE)0x08,(BYTE)0x0c,(BYTE)0x0c,(BYTE)0x06,(BYTE)0x44,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00 },
    { 0x6600,(BYTE)0x23,(BYTE)0x61,(BYTE)0x30,(BYTE)0x30,(BYTE)0x30,(BYTE)0x30,(BYTE)0x30,(BYTE)0x30,(BYTE)0x0d,(BYTE)0x0a,(BYTE)0x23,(BYTE)0x62,(BYTE)0x30,(BYTE)0x30,(BYTE)0x30,(BYTE)0x30 },
    { 0x6610,(BYTE)0x30,(BYTE)0x30,(BYTE)0x0d,(BYTE)0x0a,(BYTE)0x23,(BYTE)0x63,(BYTE)0x37,(BYTE)0x37,(BYTE)0x37,(BYTE)0x30,(BYTE)0x30,(BYTE)0x30,(BYTE)0x37,(BYTE)0x30,(BYTE)0x30,(BYTE)0x30 },
    { 0x6620,(BYTE)0x36,(BYTE)0x30,(BYTE)0x30,(BYTE)0x30,(BYTE)0x37,(BYTE)0x30,(BYTE)0x30,(BYTE)0x35,(BYTE)0x35,(BYTE)0x32,(BYTE)0x30,(BYTE)0x30,(BYTE)0x35,(BYTE)0x30,(BYTE)0x35,(BYTE)0x35 },
    { 0x6630,(BYTE)0x35,(BYTE)0x32,(BYTE)0x32,(BYTE)0x32,(BYTE)0x30,(BYTE)0x37,(BYTE)0x37,(BYTE)0x30,(BYTE)0x35,(BYTE)0x35,(BYTE)0x37,(BYTE)0x30,(BYTE)0x37,(BYTE)0x35,(BYTE)0x30,(BYTE)0x35 },
    { 0x6640,(BYTE)0x35,(BYTE)0x35,(BYTE)0x30,(BYTE)0x37,(BYTE)0x37,(BYTE)0x30,(BYTE)0x33,(BYTE)0x31,(BYTE)0x31,(BYTE)0x31,(BYTE)0x31,(BYTE)0x30,(BYTE)0x33,(BYTE)0x0d,(BYTE)0x0a,(BYTE)0x23 },
    { 0x6650,(BYTE)0x64,(BYTE)0x20,(BYTE)0x20,(BYTE)0x20,(BYTE)0x20,(BYTE)0x20,(BYTE)0x20,(BYTE)0x20,(BYTE)0x20,(BYTE)0x20,(BYTE)0x20,(BYTE)0x20,(BYTE)0x20,(BYTE)0x20,(BYTE)0x20,(BYTE)0x20 },
    { 0x6660,(BYTE)0x20,(BYTE)0x20,(BYTE)0x20,(BYTE)0x20,(BYTE)0x20,(BYTE)0x20,(BYTE)0x20,(BYTE)0x20,(BYTE)0x20,(BYTE)0x20,(BYTE)0x20,(BYTE)0x20,(BYTE)0x20,(BYTE)0x20,(BYTE)0x20,(BYTE)0x20 },
    { 0x6670,(BYTE)0x20,(BYTE)0x20,(BYTE)0x20,(BYTE)0x20,(BYTE)0x20,(BYTE)0x20,(BYTE)0x20,(BYTE)0x20,(BYTE)0x20,(BYTE)0x20,(BYTE)0x20,(BYTE)0x20,(BYTE)0x20,(BYTE)0x20,(BYTE)0x0d,(BYTE)0x0a },
    { 0x6680,(BYTE)0x23,(BYTE)0x4b,(BYTE)0x20,(BYTE)0x34,(BYTE)0x46,(BYTE)0x20,(BYTE)0x35,(BYTE)0x33,(BYTE)0x20,(BYTE)0x34,(BYTE)0x43,(BYTE)0x20,(BYTE)0x30,(BYTE)0x30,(BYTE)0x20,(BYTE)0x34 },
    { 0x6690,(BYTE)0x36,(BYTE)0x20,(BYTE)0x34,(BYTE)0x32,(BYTE)0x20,(BYTE)0x34,(BYTE)0x33,(BYTE)0x20,(BYTE)0x35,(BYTE)0x37,(BYTE)0x20,(BYTE)0x34,(BYTE)0x35,(BYTE)0x20,(BYTE)0x35,(BYTE)0x38 },
    { 0x66a0,(BYTE)0x20,(BYTE)0x30,(BYTE)0x30,(BYTE)0x20,(BYTE)0x30,(BYTE)0x30,(BYTE)0x20,(BYTE)0x30,(BYTE)0x30,(BYTE)0x20,(BYTE)0x30,(BYTE)0x30,(BYTE)0x20,(BYTE)0x30,(BYTE)0x30,(BYTE)0x20 },
    { 0x66b0,(BYTE)0x30,(BYTE)0x30,(BYTE)0x20,(BYTE)0x30,(BYTE)0x30,(BYTE)0x20,(BYTE)0x30,(BYTE)0x30,(BYTE)0x20,(BYTE)0x30,(BYTE)0x30,(BYTE)0x20,(BYTE)0x30,(BYTE)0x30,(BYTE)0x20,(BYTE)0x30 },
    { 0x66c0,(BYTE)0x30,(BYTE)0x20,(BYTE)0x30,(BYTE)0x30,(BYTE)0x20,(BYTE)0x30,(BYTE)0x30,(BYTE)0x20,(BYTE)0x35,(BYTE)0x32,(BYTE)0x20,(BYTE)0x30,(BYTE)0x30,(BYTE)0x20,(BYTE)0x30,(BYTE)0x30 },
    { 0x66d0,(BYTE)0x20,(BYTE)0x34,(BYTE)0x44,(BYTE)0x20,(BYTE)0x35,(BYTE)0x36,(BYTE)0x20,(BYTE)0x35,(BYTE)0x30,(BYTE)0x20,(BYTE)0x30,(BYTE)0x30,(BYTE)0x20,(BYTE)0x40,(BYTE)0x0d,(BYTE)0x0a },
    { 0x66e0,(BYTE)0x23,(BYTE)0x45,(BYTE)0x20,(BYTE)0x31,(BYTE)0x38,(BYTE)0x20,(BYTE)0x30,(BYTE)0x33,(BYTE)0x20,(BYTE)0x30,(BYTE)0x30,(BYTE)0x20,(BYTE)0x30,(BYTE)0x36,(BYTE)0x20,(BYTE)0x0d },
    { 0x66f0,(BYTE)0x0a,(BYTE)0x23,(BYTE)0x51,(BYTE)0x20,(BYTE)0x34,(BYTE)0x31,(BYTE)0x20,(BYTE)0x34,(BYTE)0x30,(BYTE)0x20,(BYTE)0x34,(BYTE)0x33,(BYTE)0x20,(BYTE)0x34,(BYTE)0x30,(BYTE)0x20 },
    { 0x6700,(BYTE)0x34,(BYTE)0x33,(BYTE)0x20,(BYTE)0x34,(BYTE)0x30,(BYTE)0x20,(BYTE)0x0d,(BYTE)0x0a,(BYTE)0x23,(BYTE)0x57,(BYTE)0x20,(BYTE)0x30,(BYTE)0x30,(BYTE)0x20,(BYTE)0x30,(BYTE)0x30 },
    { 0x6710,(BYTE)0x20,(BYTE)0x30,(BYTE)0x30,(BYTE)0x20,(BYTE)0x30,(BYTE)0x37,(BYTE)0x20,(BYTE)0x34,(BYTE)0x43,(BYTE)0x20,(BYTE)0x30,(BYTE)0x43,(BYTE)0x20,(BYTE)0x30,(BYTE)0x30,(BYTE)0x20 },
    { 0x6720,(BYTE)0x43,(BYTE)0x3a,(BYTE)0x5c,(BYTE)0x2a,(BYTE)0x2e,(BYTE)0x2a,(BYTE)0x40,(BYTE)0x0d,(BYTE)0x0a,(BYTE)0x23,(BYTE)0x57,(BYTE)0x20,(BYTE)0x30,(BYTE)0x30,(BYTE)0x20,(BYTE)0x30 },
    { 0x6730,(BYTE)0x30,(BYTE)0x20,(BYTE)0x30,(BYTE)0x32,(BYTE)0x20,(BYTE)0x30,(BYTE)0x42,(BYTE)0x20,(BYTE)0x34,(BYTE)0x43,(BYTE)0x20,(BYTE)0x30,(BYTE)0x39,(BYTE)0x20,(BYTE)0x30,(BYTE)0x30 },
    { 0x6740,(BYTE)0x20,(BYTE)0x40,(BYTE)0x0d,(BYTE)0x0a,(BYTE)0x23,(BYTE)0x57,(BYTE)0x20,(BYTE)0x30,(BYTE)0x30,(BYTE)0x20,(BYTE)0x30,(BYTE)0x30,(BYTE)0x20,(BYTE)0x30,(BYTE)0x41,(BYTE)0x20 },
    { 0x6750,(BYTE)0x30,(BYTE)0x46,(BYTE)0x20,(BYTE)0x33,(BYTE)0x34,(BYTE)0x20,(BYTE)0x30,(BYTE)0x39,(BYTE)0x20,(BYTE)0x30,(BYTE)0x30,(BYTE)0x20,(BYTE)0x40,(BYTE)0x0d,(BYTE)0x0a,(BYTE)0x23 },
    { 0x6760,(BYTE)0x57,(BYTE)0x20,(BYTE)0x30,(BYTE)0x30,(BYTE)0x20,(BYTE)0x30,(BYTE)0x30,(BYTE)0x20,(BYTE)0x30,(BYTE)0x45,(BYTE)0x20,(BYTE)0x30,(BYTE)0x31,(BYTE)0x20,(BYTE)0x33,(BYTE)0x34 },
    { 0x6770,(BYTE)0x20,(BYTE)0x30,(BYTE)0x39,(BYTE)0x20,(BYTE)0x30,(BYTE)0x30,(BYTE)0x20,(BYTE)0x40,(BYTE)0x0d,(BYTE)0x0a,(BYTE)0x23,(BYTE)0x57,(BYTE)0x20,(BYTE)0x30,(BYTE)0x30,(BYTE)0x20 },
    { 0x6780,(BYTE)0x30,(BYTE)0x30,(BYTE)0x20,(BYTE)0x30,(BYTE)0x34,(BYTE)0x20,(BYTE)0x30,(BYTE)0x37,(BYTE)0x20,(BYTE)0x34,(BYTE)0x43,(BYTE)0x20,(BYTE)0x30,(BYTE)0x43,(BYTE)0x20,(BYTE)0x30 },
    { 0x6790,(BYTE)0x30,(BYTE)0x20,(BYTE)0x40,(BYTE)0x0d,(BYTE)0x0a,(BYTE)0x23,(BYTE)0x57,(BYTE)0x20,(BYTE)0x30,(BYTE)0x30,(BYTE)0x20,(BYTE)0x30,(BYTE)0x30,(BYTE)0x20,(BYTE)0x30,(BYTE)0x43 },
    { 0x67a0,(BYTE)0x20,(BYTE)0x30,(BYTE)0x42,(BYTE)0x20,(BYTE)0x34,(BYTE)0x43,(BYTE)0x20,(BYTE)0x30,(BYTE)0x39,(BYTE)0x20,(BYTE)0x30,(BYTE)0x30,(BYTE)0x20,(BYTE)0x40,(BYTE)0x0d,(BYTE)0x0a },
    { 0x67b0,(BYTE)0x23,(BYTE)0x57,(BYTE)0x20,(BYTE)0x30,(BYTE)0x30,(BYTE)0x20,(BYTE)0x30,(BYTE)0x30,(BYTE)0x20,(BYTE)0x30,(BYTE)0x38,(BYTE)0x20,(BYTE)0x30,(BYTE)0x46,(BYTE)0x20,(BYTE)0x33 },
    { 0x67c0,(BYTE)0x34,(BYTE)0x20,(BYTE)0x30,(BYTE)0x39,(BYTE)0x20,(BYTE)0x30,(BYTE)0x30,(BYTE)0x20,(BYTE)0x40,(BYTE)0x0d,(BYTE)0x0a,(BYTE)0x23,(BYTE)0x57,(BYTE)0x20,(BYTE)0x30,(BYTE)0x30 },
    { 0x67d0,(BYTE)0x20,(BYTE)0x30,(BYTE)0x30,(BYTE)0x20,(BYTE)0x30,(BYTE)0x36,(BYTE)0x20,(BYTE)0x30,(BYTE)0x31,(BYTE)0x20,(BYTE)0x33,(BYTE)0x34,(BYTE)0x20,(BYTE)0x30,(BYTE)0x39,(BYTE)0x20 },
    { 0x67e0,(BYTE)0x30,(BYTE)0x30,(BYTE)0x20,(BYTE)0x40,(BYTE)0x0d,(BYTE)0x0a,(BYTE)0x23,(BYTE)0x4e,(BYTE)0x20,(BYTE)0x46,(BYTE)0x46,(BYTE)0x20,(BYTE)0x30,(BYTE)0x34,(BYTE)0x20,(BYTE)0x30 },
    { 0x67f0,(BYTE)0x30,(BYTE)0x30,(BYTE)0x20,(BYTE)0x40,(BYTE)0x20,(BYTE)0x2a,(BYTE)0x2e,(BYTE)0x2a,(BYTE)0x40,(BYTE)0x20,(BYTE)0x40,(BYTE)0x20,(BYTE)0x0d,(BYTE)0x0a,(BYTE)0x23,(BYTE)0x44 },
    { 0x6800,(BYTE)0x20,(BYTE)0x46,(BYTE)0x46,(BYTE)0x20,(BYTE)0x30,(BYTE)0x31,(BYTE)0x20,(BYTE)0x30,(BYTE)0x30,(BYTE)0x30,(BYTE)0x20,(BYTE)0x40,(BYTE)0x20,(BYTE)0x2a,(BYTE)0x2e,(BYTE)0x2a },
    { 0x6810,(BYTE)0x40,(BYTE)0x20,(BYTE)0x40,(BYTE)0x20,(BYTE)0x0d,(BYTE)0x0a,(BYTE)0x23,(BYTE)0x47,(BYTE)0x20,(BYTE)0x30,(BYTE)0x33,(BYTE)0x20,(BYTE)0x46,(BYTE)0x46,(BYTE)0x20,(BYTE)0x30 },
    { 0x6820,(BYTE)0x30,(BYTE)0x30,(BYTE)0x20,(BYTE)0x2a,(BYTE)0x2e,(BYTE)0x41,(BYTE)0x50,(BYTE)0x50,(BYTE)0x40,(BYTE)0x20,(BYTE)0x40,(BYTE)0x20,(BYTE)0x40,(BYTE)0x20,(BYTE)0x0d,(BYTE)0x0a },
    { 0x6830,(BYTE)0x23,(BYTE)0x47,(BYTE)0x20,(BYTE)0x30,(BYTE)0x33,(BYTE)0x20,(BYTE)0x46,(BYTE)0x46,(BYTE)0x20,(BYTE)0x30,(BYTE)0x30,(BYTE)0x30,(BYTE)0x20,(BYTE)0x2a,(BYTE)0x2e,(BYTE)0x50 },
    { 0x6840,(BYTE)0x52,(BYTE)0x47,(BYTE)0x40,(BYTE)0x20,(BYTE)0x40,(BYTE)0x20,(BYTE)0x40,(BYTE)0x20,(BYTE)0x0d,(BYTE)0x0a,(BYTE)0x23,(BYTE)0x59,(BYTE)0x20,(BYTE)0x30,(BYTE)0x33,(BYTE)0x20 },
    { 0x6850,(BYTE)0x46,(BYTE)0x46,(BYTE)0x20,(BYTE)0x30,(BYTE)0x30,(BYTE)0x30,(BYTE)0x20,(BYTE)0x2a,(BYTE)0x2e,(BYTE)0x47,(BYTE)0x54,(BYTE)0x50,(BYTE)0x40,(BYTE)0x20,(BYTE)0x40,(BYTE)0x20 },
    { 0x6860,(BYTE)0x40,(BYTE)0x20,(BYTE)0x0d,(BYTE)0x0a,(BYTE)0x23,(BYTE)0x50,(BYTE)0x20,(BYTE)0x30,(BYTE)0x33,(BYTE)0x20,(BYTE)0x46,(BYTE)0x46,(BYTE)0x20,(BYTE)0x30,(BYTE)0x30,(BYTE)0x30 },
    { 0x6870,(BYTE)0x20,(BYTE)0x2a,(BYTE)0x2e,(BYTE)0x54,(BYTE)0x54,(BYTE)0x50,(BYTE)0x40,(BYTE)0x20,(BYTE)0x40,(BYTE)0x20,(BYTE)0x40,(BYTE)0x20,(BYTE)0x0d,(BYTE)0x0a,(BYTE)0x23,(BYTE)0x46 },
    { 0x6880,(BYTE)0x20,(BYTE)0x30,(BYTE)0x33,(BYTE)0x20,(BYTE)0x30,(BYTE)0x34,(BYTE)0x20,(BYTE)0x30,(BYTE)0x30,(BYTE)0x30,(BYTE)0x20,(BYTE)0x2a,(BYTE)0x2e,(BYTE)0x54,(BYTE)0x4f,(BYTE)0x53 },
    { 0x6890,(BYTE)0x40,(BYTE)0x20,(BYTE)0x40,(BYTE)0x20,(BYTE)0x40,(BYTE)0x20,(BYTE)0x0d,(BYTE)0x0a,(BYTE)0x23,(BYTE)0x4d,(BYTE)0x20,(BYTE)0x30,(BYTE)0x30,(BYTE)0x20,(BYTE)0x30,(BYTE)0x31 },
    { 0x68a0,(BYTE)0x20,(BYTE)0x30,(BYTE)0x30,(BYTE)0x20,(BYTE)0x46,(BYTE)0x46,(BYTE)0x20,(BYTE)0x43,(BYTE)0x20,(BYTE)0x42,(BYTE)0x4F,(BYTE)0x4F,(BYTE)0x54,(BYTE)0x20,(BYTE)0x44,(BYTE)0x49 },
    { 0x68b0,(BYTE)0x53,(BYTE)0x4b,(BYTE)0x40,(BYTE)0x20,(BYTE)0x40,(BYTE)0x20,(BYTE)0x0d,(BYTE)0x0a,(BYTE)0x23,(BYTE)0x4d,(BYTE)0x20,(BYTE)0x30,(BYTE)0x30,(BYTE)0x20,(BYTE)0x30,(BYTE)0x30 },
    { 0x68c0,(BYTE)0x20,(BYTE)0x30,(BYTE)0x30,(BYTE)0x20,(BYTE)0x46,(BYTE)0x46,(BYTE)0x20,(BYTE)0x41,(BYTE)0x20,(BYTE)0x46,(BYTE)0x4c,(BYTE)0x4f,(BYTE)0x50,(BYTE)0x50,(BYTE)0x59,(BYTE)0x20 },
    { 0x68d0,(BYTE)0x44,(BYTE)0x49,(BYTE)0x53,(BYTE)0x4b,(BYTE)0x40,(BYTE)0x20,(BYTE)0x40,(BYTE)0x20,(BYTE)0x0d,(BYTE)0x0a,(BYTE)0x23,(BYTE)0x4d,(BYTE)0x20,(BYTE)0x30,(BYTE)0x31,(BYTE)0x20 },
    { 0x68e0,(BYTE)0x30,(BYTE)0x30,(BYTE)0x20,(BYTE)0x30,(BYTE)0x30,(BYTE)0x20,(BYTE)0x46,(BYTE)0x46,(BYTE)0x20,(BYTE)0x42,(BYTE)0x20,(BYTE)0x46,(BYTE)0x4c,(BYTE)0x4f,(BYTE)0x50,(BYTE)0x50 },
    { 0x68f0,(BYTE)0x59,(BYTE)0x20,(BYTE)0x44,(BYTE)0x49,(BYTE)0x53,(BYTE)0x4b,(BYTE)0x40,(BYTE)0x20,(BYTE)0x40,(BYTE)0x20,(BYTE)0x0d,(BYTE)0x0a,(BYTE)0x23,(BYTE)0x54,(BYTE)0x20,(BYTE)0x30 },
    { 0x6900,(BYTE)0x30,(BYTE)0x20,(BYTE)0x30,(BYTE)0x37,(BYTE)0x20,(BYTE)0x30,(BYTE)0x32,(BYTE)0x20,(BYTE)0x46,(BYTE)0x46,(BYTE)0x20,(BYTE)0x20,(BYTE)0x20,(BYTE)0x54,(BYTE)0x52,(BYTE)0x41 },
    { 0x6910,(BYTE)0x53,(BYTE)0x48,(BYTE)0x40,(BYTE)0x20,(BYTE)0x40,(BYTE)0x20,(BYTE)0x0d,(BYTE)0x0a,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00,(BYTE)0x00 },
};


BOOL FRWVirtualDisk(BOOL fWrite, ULONG buf, ULONG sec, ULONG count)
{
    // Reading or writing a sector from virtual drive.
    // Sector numbers are $1x0000 too big. To convert to file
    // offset in virtual disk, subtract $1x0000 and multiply by 512.
    //
    // Returns non-zero error code

    ULONG cb;
    unsigned iVHD;
    int i;
    PDI pdi;

#ifndef NDEBUG
    Assert(buf < vi.cbRAM[0]);
    Assert(count >= 1);
#endif

    // upper 3 bits of byte1 are unit #. Map it to the SCSI device
    // of the same number, skipping the 2 virtual disks for floppies.
    // Thus unit 1 is 1+2 = 3;

    iVHD = ((sec >> 21) & 7) + 2;
    sec  = sec & 0x1FFFFF;

#if TRACEDISK
    DebugStr("FRWVirtualDisk: fWrite = %d, buf = %08X, id = %d, sec = %08X, count = %d\n",
          fWrite, buf, iVHD, sec, count);
#endif

    if (!vmCur.fUseVHD)
        PokeW(0x446, (WORD)vmCur.iBootDisk);

    if ((buf >= vi.cbRAM[0]) || (count == 0))
        return 1;

    if (iVHD >= 9)
        return 1;

    pdi = vi.rgpdi[iVHD];

    if (pdi == NULL)    // nothing mapped to this drive
        {
        if (!fWrite && (iVHD == vmCur.iBootDisk) && (vmCur.iBootDisk >= 2))
            {
            unsigned int i;

            // Atari ST boot device must return something

            for (i = (sec << 9); i < ((sec + count) << 9); i++)
                {
                unsigned char b = 0;
                unsigned id;

                for (id = 0; id < (sizeof(d16HDX) / sizeof(data16)); id++)
                    {
                    if (d16HDX[id].off == (i & 0xFFF0))
                        {
                        b = d16HDX[id].rgb[i & 15];
                        break;
                        }
                    }

                vmPokeB(buf++, b);
                }
            return 0;
            }

        return 1;
        }

    if (!fWrite && sec < 2 && pdi->fst != FS_RAW && (pdi->offsec == 0))
        {
        ReadBootSect(vsthw.rgbDiskBuffer, sec);
        HostToGuestDataCopy(vsthw.DMAbasis, vsthw.rgbDiskBuffer,
             ACCESS_PHYSICAL_WRITE, 512);
        return 0;
        }

    if (fWrite && sec < 2 && pdi->fst != FS_RAW && (pdi->offsec == 0))
        {
        if (GuestToHostDataCopy(vsthw.DMAbasis, vsthw.rgbDiskBuffer,
             ACCESS_PHYSICAL_READ, 512))
            {
            WriteBootSect(vsthw.rgbDiskBuffer, sec, vsthw.DMAbasis);
            }
        return 0;
        }

    if (!vmCur.fUseVHD && (iVHD > vmCur.iBootDisk))
        return 1;

    cb = count*512;

    // REVIEW: There is some stupid SCSI hack whereby large
    // block transfers (on the order of 200+ sectors) return
    // an error. So do the stupid thing of breaking up large
    // disk I/O operations.
    //
    // We're arbitrarily going to use 120 sector (60K) blocks

    i = fTrue;

    while (count && i)
        {
        void *pv = NULL;

        pdi->sec   = sec & 0xFFFFF;
        pdi->count = min(120,count);

        pv = malloc(pdi->count * 512);

        if (pv == NULL)
            return 1;

        memset(pv, 0, cb);
        pdi->lpBuf = pv;

        if (fWrite)
            {
            i = GuestToHostDataCopy(buf, pv, ACCESS_PHYSICAL_READ, cb) &&
                FRawDiskRWPdi(pdi, fWrite);
            }
        else
            {
            i = FRawDiskRWPdi(pdi, fWrite);

            // read may fail if past EOF, so always return OK.
            i = fTrue;

            i = HostToGuestDataCopy(buf, pv, ACCESS_PHYSICAL_WRITE, cb);
            }

        count -= min(120,count);

        sec += 120;
        cb -= (120 * 512);
        buf += (120 * 512);
        }

    return !i;
}


ULONG __cdecl DumpHW(void)
{
    unsigned i;

    printf("\nAtari hardware dump:\n\n");

    for (i = 0; i < 4; i++)
        {
        printf("RAM range %d start %08X size %08X\n",
             i, vi.eaRAM[i], vi.cbRAM[i]);
        }
    printf("\n");

    for (i = 0; i < 4; i++)
        {
        printf("ROM range %d start %08X size %08X\n",
             i, vi.eaROM[i], vi.cbROM[i]);
        }
    printf("\n");

    printf("Screen video base: %08X\n", vsthw.dbase);
    printf("\n");

    return TRUE;
}


#ifndef NDEBUG

void DumpDMA(int byte0, int byte1, int byte2, int byte3, int byte4, int byte5)
{
    if (!fDebug)
        return;

    if (byte0 & 0xFFFFFF00)
        {
        // contents of hard disk partition table

        unsigned char *pch = (unsigned char *)byte0;
        int i;

        for (i = 0; i < 512; i++)
            {
            if ((i & 15) == 0)
                {
                DebugStr("\n%02X:  ", i);
                }
            else if ((i & 15) == 8)
                {
                DebugStr(" ");
                }

            DebugStr("%02X ", *pch++);
            }

        DebugStr("\n");
        return;
        }

    DebugStr("DMA: %02X %02X %02X %02X %02X %02X  ",
        byte0, byte1, byte2, byte3, byte4, byte5);

    DebugStr("Ctl: %d  Drv: %d  ", byte0>>5, byte1>>5);

    switch(byte0 & 0x1F)
        {
    default:
        DebugStr("Unknown command!!!");
        break;

    case 0:
        DebugStr("Test Unit Ready");
        break;

    case 1:
        DebugStr("Restore");
        break;

    case 3:
        DebugStr("Request Sense");
        break;

    case 8:
        DebugStr("Read sector: %d  count: %d",
              ((byte1 & 0x1F) << 16) + (byte2 << 8) + byte3, byte4);
        break;

    case 10:
        DebugStr("Write sector: %d  count: %d",
              ((byte1 & 0x1F) << 16) + (byte2 << 8) + byte3, byte4);
        break;

    case 11:
        DebugStr("Seek");
        break;

    case 18:
        DebugStr("Inquiry");
        break;

    case 27:
        DebugStr("Park");
        break;
        }

    DebugStr("\n");

}

void DumpAssert(int x)
{
    DebugStr("ASSERT %d FAIL\n", x);
}

void DumpACIA(ULONG b)
{
    if (!fDebug)
        return;

    DebugStr("putting %d in keyboard buffer\n", b & 0xFF);
}

#endif  // !NDEBUG


//
// OutPrinter is called from the TRAP #13 hook
//

void __cdecl OutPrinter(char ch)
{
    ByteToPrinter(ch);
    return;
}


ULONG __cdecl Exec(BOOL fStep, BOOL fCont)
{
    ULONG tickStart, ticks;  // used to time the exeuction time of this function
    ULONG stat;

    extern void InvalidateDispatch();

#if 0
    vi.countR = 0;
    vi.countW = 0;
#endif
    vi.cFreq  = 0;

    /* valid combinations of fStep and fCont
     *
     * fStep=0 fCont=0
     * fStep=0 fCont=1 = execute
     * fStep=1 fCont=0 = single step
     * fStep=1 fCont=1 = trace
     *
     */

    if (!fStep)
        {
        static ULONG lTick;         // current reading of millisecond timer
        static ULONG lTick1000;     // previous reading of millisecond timer

        /* non-tracing execution loop */

        QueryTickCtr();
        tickStart = Getms();

        if (vi.fMouseMoved && vi.fHaveFocus && vi.fGEMMouse)
            {
            POINT pt;

//                    DebugStr("m");
            vi.fMouseMoved = FALSE;
            GetCursorPos(&pt);
            ScreenToClient(vi.hWnd, &pt);
            FMouseMsgST(vi.hWnd, WM_MOUSEMOVE, -1, pt.x, pt.y);
            }

        for(;;)
            {
            // execute some 68000 code

            stat = vpci->pfnGo(5000);

            if (stat != 0)
                {
//                DebugStr("exiting due to exit code %d\n", stat);
                break;
                }

// this is in case we want to start tracing after a fixed # of instructions.
//            if (vpregs->count > 250000)
//                goto totrace;

            lTick1000 = lTick;
            QueryTickCtr();
            lTick = Getms();

#if 0
            if (vmCur.fSound)
                {
                // check sound

                UpdateSound();
                SoundDoneCallback();
                }
#endif

            // check for 200 Hz timer tick

            if ((lTick - vsthw.lTick200) >= 5)
                {
                ULONG dTick = (lTick - vsthw.lTick200)/5;

                // during disk I/O, we may miss many ticks,
                // so this bumps up the 200 Hz timer to compensate.
                // Only do it up to 800 ticks (4 seconds).
                // This makes disk benchmarks in Quick Index accurate.

                if ((dTick > 1) && (dTick < 800))
                    {
                    PokeL(0x4BA, PeekL(0x4BA) + dTick - 1);
                    }

                if (dTick > 2)
                    {
#if !defined(NDEBUG)
                    printf("200 Hz timeout! period = %d ms\n",
                        lTick - vsthw.lTick200);
#endif
                    }

                vsthw.lTick200 += dTick*5;

#ifndef NDEBUG
                if (vrgf.fTimerC)
                    {
                    DebugStr("C");
                    }
#endif

                // Every 200 Hz timer tick, force a 200Hz timer interrupt
                vrgf.fTimerC = 1;

#if 1
                if (vmCur.fSound)
                    {
                    // check sound

                    UpdateSound();
                    SoundDoneCallback();
                    }
#endif

                // decrement printer timer

                if (vi.cPrintTimeout && vmCur.fShare)
                    {
                    vi.cPrintTimeout--;
                    if (vi.cPrintTimeout == 0)
                        {
                        FlushToPrinter();
                        UnInitPrinter();
                        }
                    }
                }

            // check for 60 Hz / 70 Hz VBI

            if ((lTick - vsthw.lTickVBI) >= (ULONG)(vsthw.fMono ? 15 : 17))
                {
                if ((lTick - vsthw.lTickVBI) < 30)
                    {
                    vsthw.lTickVBI += (vsthw.fMono ? 15 : 17);
                    }
                else
                    {
#if !defined(NDEBUG)
                    printf("VBI timeout! period = %d ms\n",
                        lTick - vsthw.lTickVBI);
#endif
                    vsthw.lTickVBI = lTick - 8;
                    }

#ifndef NDEBUG
                if (vrgf.fVBIPending)
                    {
                    DebugStr("V");
                    }
#endif

                vrgf.fVBIPending = 1;

                // check for mouse movement
                if (vi.fMouseMoved && vi.fHaveFocus && vi.fGEMMouse)
                    {
                    POINT pt;

//                    DebugStr("m");
                    vi.fMouseMoved = FALSE;
                    GetCursorPos(&pt);
                    ScreenToClient(vi.hWnd, &pt);
                    FMouseMsgST(vi.hWnd, WM_MOUSEMOVE, -1, pt.x, pt.y);
                    }

                CchSerialPending();

                if (GetQueueStatus(QS_ALLINPUT))
                    break;
                }

            // about once a second, force full screen refresh and
            // force printer to flush

            if ((lTick - vsthw.lTickSec) >= 1000)
                {
                FlushToPrinter();

#if !defined(NDEBUG)
                DisplayStatus();
#endif
                if ((lTick - vsthw.lTickSec) < 2000)
                    vsthw.lTickSec += 1000;
                else
                    vsthw.lTickSec = lTick;

                // To avoid the mysterious deadlock, break out every second

                break;
                }
            }

        QueryTickCtr();
        ticks = Getms() - tickStart;
        }

    else if (fCont)
        {
        int i = 0;

        /* trace execution */

        do
            {
            InvalidateDispatch();
            stat = vpci->pfnStep(1);
            FDumpRegsVM();
            } while (!stat && (i++ < 20));
        }
    else
        {
        /* step execution */

        InvalidateDispatch();
        stat = vpci->pfnStep(1);
        FDumpRegsVM();
        }

    return stat;
}


void ReadBootSect(BYTE *buf, ULONG sec)
{
    HFILE h;
    OFSTRUCT ofs;
    unsigned short int w = 0;
    unsigned short int *pw = (unsigned short int *)buf;
    int i;

    if (!vmCur.fUseVHD)
        {
        memset(buf, 0, 512);

        if (vmCur.iBootDisk)
            {
            PokeW(0x446, (WORD)vmCur.iBootDisk);  // _bootdev
            PokeL(0x4C2, 0x03FFFFFF); // _drvbits
            }
        else
            {
            PokeW(0x446, 0);
            PokeL(0x4C2, 0x03);
            }
#if 0
        PokeL(0xE40004,PeekL(0x84);
        PokeL(0x84, 0xE000000);
        PokeL(0xE00000, 0x7
#endif
        return;
        }

#ifndef NDEBUG
    if (fDebug) DebugStr("in ReadBootSect\n");
#endif

    if (!vmCur.iBootDisk)
        {
        PokeW(0x446, 0);
        PokeL(0x4C2, 0x03);
        return;
        }

    h = OpenFile("STHDBOOT.SEC", &ofs, OF_READ);

    if (h != HFILE_ERROR)
        {
        if (SetFilePointer(h, sec*512, NULL, FILE_BEGIN) != 0xFFFFFFFF)
            ReadFile(h, buf, 512, (int *)&ofs, NULL);     // read Atari HD boot sector
        CloseHandle(h);

        if (fDebug) DebugStr("Loaded STHDBOOT.SEC\n");
        }
    else
        {
        static unsigned char rgbSTHDBOOTSEC[512] =
            {
            0x60, 0x04, 0x60, 0x00, 0x00, 0xb2, 0x08, 0xb8, 0x00, 0x02, 0x04, 0xc5, 0x31, 0xfc, 0x00, 0x00,
            0x04, 0x46, 0x48, 0xe7, 0x14, 0x00, 0x3f, 0x3c, 0xff, 0xff, 0x3f, 0x3c, 0x00, 0x0b, 0x4e, 0x4d,
            0x58, 0x4f, 0x08, 0x00, 0x00, 0x03, 0x66, 0x42, 0x70, 0x03, 0x72, 0xf8, 0x41, 0xfa, 0x01, 0x98,
            0x0c, 0x83, 0x44, 0x4d, 0x41, 0x72, 0x67, 0x06, 0x78, 0xff, 0x74, 0x80, 0x60, 0x04, 0x14, 0x05,
            0x67, 0x16, 0x08, 0x10, 0x00, 0x00, 0x67, 0x06, 0xc3, 0x10, 0xb4, 0x10, 0x67, 0x36, 0xd0, 0xfc,
            0x00, 0x0c, 0x51, 0xc8, 0xff, 0xee, 0x60, 0x12, 0x08, 0x10, 0x00, 0x00, 0x67, 0x04, 0xc3, 0x10,
            0x66, 0x22, 0xd0, 0xfc, 0x00, 0x0c, 0x51, 0xc8, 0xff, 0xf0, 0x50, 0x4f, 0x20, 0x78, 0x04, 0xf2,
            0x20, 0x28, 0x00, 0x18, 0x48, 0x40, 0xb0, 0xbc, 0x19, 0x87, 0x04, 0x22, 0x64, 0x04, 0x3e, 0x3c,
            0x00, 0xe0, 0x4e, 0x75, 0x2c, 0x28, 0x00, 0x04, 0x7a, 0x01, 0x49, 0xfa, 0x01, 0x74, 0x61, 0x26,
            0x4a, 0x40, 0x66, 0xd6, 0x20, 0x4c, 0x32, 0x3c, 0x00, 0xff, 0x70, 0x00, 0xd0, 0x58, 0x51, 0xc9,
            0xff, 0xfc, 0xb0, 0x7c, 0x12, 0x34, 0x66, 0xc2, 0x47, 0xfa, 0x00, 0x0c, 0x4b, 0xfa, 0xff, 0xbe,
            0x4c, 0xdf, 0x00, 0x28, 0x4e, 0xd4, 0x4a, 0x44, 0x6b, 0x16, 0x3f, 0x04, 0x2f, 0x0c, 0x3f, 0x05,
            0x2f, 0x06, 0x3f, 0x3c, 0x00, 0x2a, 0x4e, 0x4e, 0xde, 0xfc, 0x00, 0x0e, 0x60, 0x00, 0x00, 0x9e,
            0x50, 0xf8, 0x04, 0x3e, 0x20, 0x38, 0x04, 0xba, 0x54, 0x80, 0xb0, 0xb8, 0x04, 0xba, 0x64, 0xfa,
            0x3c, 0x7c, 0x86, 0x09, 0x48, 0xe7, 0x02, 0x08, 0x1d, 0x6f, 0x00, 0x07, 0x00, 0x04, 0x1d, 0x6f,
            0x00, 0x06, 0x00, 0x02, 0x1c, 0xaf, 0x00, 0x05, 0x4d, 0xfa, 0x00, 0x74, 0x1d, 0x6f, 0x00, 0x01,
            0x00, 0x01, 0x1d, 0x6f, 0x00, 0x02, 0x00, 0x05, 0x1d, 0x6f, 0x00, 0x03, 0x00, 0x09, 0x1d, 0x45,
            0x00, 0x0d, 0x50, 0x4f, 0x3c, 0x7c, 0x86, 0x06, 0x3a, 0x7c, 0x86, 0x04, 0x3c, 0xbc, 0x01, 0x98,
            0x3c, 0xbc, 0x00, 0x98, 0x3a, 0x85, 0x3c, 0xbc, 0x00, 0x88, 0x70, 0x00, 0x10, 0x07, 0x80, 0x3c,
            0x00, 0x08, 0x48, 0x40, 0x30, 0x3c, 0x00, 0x8a, 0x61, 0x44, 0x41, 0xfa, 0x00, 0x32, 0x74, 0x03,
            0x20, 0x18, 0x61, 0x3a, 0x51, 0xca, 0xff, 0xfa, 0x2a, 0xbc, 0x00, 0x00, 0x00, 0x0a, 0x22, 0x3c,
            0x00, 0x00, 0x01, 0x90, 0x61, 0x2c, 0x3c, 0xbc, 0x00, 0x8a, 0x30, 0x15, 0xc0, 0x7c, 0x00, 0xff,
            0x67, 0x02, 0x70, 0xff, 0x3c, 0xbc, 0x00, 0x80, 0x51, 0xf8, 0x04, 0x3e, 0x4e, 0x75, 0x00, 0x00,
            0x00, 0x8a, 0x00, 0x00, 0x00, 0x8a, 0x00, 0x00, 0x00, 0x8a, 0x00, 0x01, 0x00, 0x8a, 0x2a, 0x80,
            0x72, 0x0a, 0xd2, 0xb8, 0x04, 0xba, 0x08, 0x38, 0x00, 0x05, 0xfa, 0x01, 0x67, 0xde, 0xb2, 0xb8,
            0x04, 0xba, 0x64, 0xf2, 0x58, 0x4f, 0x60, 0xca, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x03, 0x67, 0x08, 0x00, 0x01, 0x2c, 0x01, 0x2c, 0x6c, 0x02,
#if 0
//          0x01, 0x23, 0x00, 0x03, 0x20, 0xa0, 0x81, 0x47, 0x45, 0x4d, 0x00, 0x10, 0x00, 0x00, 0x00, 0x01,
//          0x00, 0x00, 0x01, 0x47, 0x45, 0x4d, 0x00, 0x14, 0x00, 0x00, 0x00, 0x01, 0x00, 0x00, 0x01, 0x47,
//          0x45, 0x4d, 0x00, 0x18, 0x00, 0x00, 0x00, 0x01, 0x00, 0x00, 0x01, 0x2D, 0x2D, 0x2D, 0x00, 0x1c,
//          0x00, 0x00, 0x00, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xa8, 0xdc,
#else
            // instead of 4 32-meg partitions, pretend there is a single partition
            0x01, 0x23, 0x00, 0x03, 0x20, 0xa0, 0x81, 0x47, 0x45, 0x4d, 0x00, 0x10, 0x00, 0x00, 0x00, 0x01,
            0x00, 0x00, 0x00, 0x2D, 0x2D, 0x2D, 0x00, 0x14, 0x00, 0x00, 0x00, 0x01, 0x00, 0x00, 0x00, 0x2D,
            0x2D, 0x2D, 0x00, 0x18, 0x00, 0x00, 0x00, 0x01, 0x00, 0x00, 0x00, 0x2D, 0x2D, 0x2D, 0x00, 0x1c,
            0x00, 0x00, 0x00, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xa8, 0xdc,
#endif
            };

        memcpy(buf, rgbSTHDBOOTSEC, 512);

        DebugStr("Couldn't load STHDBOOT.SEC\n");
        }

    PokeW(0x446, (WORD)vmCur.iBootDisk);

#if 0
    if (sec)
        return;
#endif

    for (i = 0; i < 255; i++)
        {
        w += swab(pw[i]);
        }

#ifndef NDEBUG
    if (fDebug)
        {
        DebugStr("chksum of first 255 words = %04X\n", w);
        DebugStr("even out before = %04X, chksum = %04X\n", swab(pw[255]),
            (unsigned short int)(w + swab(pw[255])));
        }
#endif

    if (sec)
        return;

    pw[255] = swab(0x1234-w);

#ifndef NDEBUG
    if (fDebug)
        {
        DebugStr("even out after = %04X, chksum = %04X\n", swab(pw[255]),
            (unsigned short int)(w + swab(pw[255])));
        }
#endif
}


void WriteBootSect(buf, sec, lAddr)
BYTE *buf;
ULONG sec;
ULONG lAddr;
{
    HFILE h;
    OFSTRUCT ofs;
    unsigned short int w = 0;
    unsigned short int *pw = (unsigned short int *)buf;
    int i;

#ifndef NDEBUG
    if (fDebug)
        {
        DebugStr("in WriteBootSect\n");

        for (i = 0; i < 5; i++)
            {
            DebugStr("DMAbasis = %08X\n", lAddr);
            DebugStr("sector offset %d = %02X\n", i, PeekB((lAddr & 0x00FFFFFF) +i));
            }
        }
#endif

    h = OpenFile("STHDBOOT.SEC", &ofs, OF_CREATE);

    if (h == HFILE_ERROR)
        {
#ifndef NDEBUG
        DebugStr("WriteBootSect: could not open STHDBOOT.SEC\n");
#endif
        goto chksum;
        }
    if (SetFilePointer(h, sec*512, NULL, FILE_BEGIN) != 0xFFFFFFFF)
        WriteFile(h, buf, 512, (int *)&ofs, NULL); // write Atari's HD boot sector
    CloseHandle(h);

chksum:
#if 0
    if (sec)
        return;
#endif

    for (i = 0; i < 255; i++)
        {
        w += swab(pw[i]);
        }

#ifndef NDEBUG
    if (!fDebug)
        return;

    DebugStr("chksum of first 255 words = %04X\n", w);
    DebugStr("even out = %04X, chksum = %04X\n", swab(pw[255]),
        (unsigned short int)(w + swab(pw[255])));
#endif
}


/////////////////////////////////////////////////////////////////////////////
//
// ATARI ST HARDWARE EMULATION
//
//   ReadHWByte(ea,pb)  - read byte at ea to *pb
//   ReadHWWord(ea,pw)  - read word at ea to *pw
//   WriteHWByte(ea,pb) - write byte to ea from *pb
//   WriteHWWord(ea,pw) - write word to ea from *pw
//
// All routines return HRESULT
//
/////////////////////////////////////////////////////////////////////////////

STHW vsthw;

//
// ReadHWByte(ea)
//
// Handler for emulating Atari ST hardware byte reads.
//
// Returns:
//  $000000xx - success, xx is return value
//  $8000xxxx - xxxx is trap value
//

static ULONG __cdecl st_ReadHWByte(ULONG ea, BYTE *pb)
{
    ea = ea & vsthw.lAddrMask;

    if (ea < vi.cbRAM[0])
        {
        // it's just RAM

        return *(BYTE *)(vi.pbRAM[0] - (ea - vi.eaRAM[0]));
        }

    if (ea < 0x00400000)
        {
        // non-RAM below 4 meg returns garbage

        return 0xFF;
        }

    if ((ea >= vi.eaROM[0]) && (ea < vi.eaROM[0] + vi.cbROM[0]))
        {
        // it's just TOS ROM

        return *(BYTE *)(vi.pbROM[0] - (ea - vi.eaROM[0]));
        }

    if ((ea >= vi.eaROM[1]) && (ea < vi.eaROM[1] + vi.cbROM[1]))
        {
        // it's just cartridge ROM

        return *(BYTE *)(vi.pbROM[1] - (ea - vi.eaROM[1]));
        }

#if TRACEHW
    printf("ReadHWByte: %08X\n", ea);
#endif

    if ((vpregs->SR & 0x2000) == 0)
        {
        // oh oh, not in supervisor mode!

        return 0x80000008; // priviledge violation
        }

    switch (ea & 0x00FFFF00)
        {
    default:
//        _asm { int 3 } ;
        break;

    case 0x00FF8000: // memory configuration

        // $FF8000 reads $FF0x, all others bus error

        if (ea == 0x00FF8001)
            {
            // some magic memory config, 5 on 1040ST, return 10?
            return 10;
            }
        break;

    case 0x00FF8200: // video chip
        ea &= 255;

        if (ea >= 0x40 && ea < 0x60)
            {
            // do a byte swap

            return vsthw.bColor[(ea^1)-0x40];
            }

        switch (ea & 255)
            {
        case 0x01:
            return vsthw.dbaseh;

        case 0x03:
            return vsthw.dbasel;

        case 0x05:
            vsthw.vcount = ((vsthw.vcount - vsthw.dbase + 1) % 32000) + vsthw.dbase;
            return vsthw.vcounthi;

        case 0x07:
            vsthw.vcount = ((vsthw.vcount - vsthw.dbase + 1) % 32000) + vsthw.dbase;
            return vsthw.vcountmid;

        case 0x09:
            vsthw.vcount = ((vsthw.vcount - vsthw.dbase + 1) % 32000) + vsthw.dbase;
            return vsthw.vcountlo & 0xFC;

        case 0x0A:
            return 0xFC;    // syncmode

        case 0x0B:
            return 0xFF;

        case 0x0D:
            if (vrgf.fSTE)
                return vsthw.dbase0;
            return 0xFF;

        case 0x0F:
            if (vrgf.fSTE)
                return vsthw.scanskip;
            return 0xFF;

        case 0x60:
            return vsthw.shiftmd;

        case 0x62:
            if (vrgf.fSTE > 1)
                return 0;

        case 0x65:
            if (vrgf.fSTE)
                return vsthw.horscroll;
            break;
            }
        break;

    case 0x00FF8600: // disk drive

        switch (ea & 255)
            {
        case 0x09:
            return vsthw.dmahi;

        case 0x0B:
            return vsthw.dmamid;

        case 0x0D:
            return vsthw.dmalo;
            }

        break;

    case 0x00FF8700: // AHDI 5.0 accesses $FF8781, TT DMA
        if (vrgf.fSTE > 1)
            {
//            _asm { int 3 } ;
            return 0;
            }
        break;

    case 0x00FF8800: // sound chip

        if ((ea & 3) == 0)
            {
            return FReadYamaha();
            }

        return 0xFF; // sound chip doesn't bus error

    case 0x00FF8900: // stereo sound chip on STE
        if (vrgf.fSTE && ea < 0xFF8940)
            {
            if (ea >= 0xFF8921)
                return 0;

            ea = (ea ^ 1) & 0x3F;   // byte swap

            return vsthw.rgbSTESound[ea];
            }
        break;

    case 0x00FF8A00: // blitter
        if (vmCur.fBlitter && ea < 0xFF8A3E)
            {
            ea = (ea ^ 1) & 0x3F;   // byte swap

            return vsthw.blitter.rgbBlitter[ea];
            }
        break;

    case 0x00FF8E00: // Mega STE / TT SCC ports
        if (vrgf.fSTE > 1)
            {
//            _asm { int 3 } ;
            return 0;
            }
        break;

    case 0x00FF9200: // paddle ports (NYI)
        break;

    case 0x00FF9800: // Falcon color registers
    case 0x00FF9900: // Falcon color registers
    case 0x00FF9A00: // Falcon color registers
    case 0x00FF9B00: // Falcon color registers
        break;

    case 0x00FFFA00: // MFP

        // MFP is odd byte access from $FFFA01 to $FFFA3F.
        // All other bytes bus error.
        //
        // On TT, $FFFA81 to $FFFABF is the second MFP (NYI)

        ea &= 255;

        if ((ea < 0x40) && (ea & 1))
            {
            if (ea >= 0x30) // unused bytes
                return 0xFF;

            // MFP registers are on odd addresses

            switch (ea)
                {
            case 0x01:
                {
                // reading $FFFA01
                //
                // bit 7, 0 = mono, 1 = color
                // bit 6, RS-232 ring
                // bit 5, FDC/HDC done interrupt
                // bit 4, keyboard/MIDI interrupt
                // bit 3, reserved
                // bit 2, RS232 CTS
                // bit 1, RS232 carrier detect
                // bit 0, 0 = printer ready, 1 = printer busy

                BYTE  b = vsthw.fMono ? 0x00 : 0x80; // BIT 7
                int w;

                b |= !FPrinterReady();          // BIT 0

                w = GetModemStatus() & 0x4646;
                b |= (w >> 8);                  // BITS 6 2 1

                // TODO: fill in other bits later

                if (vi.keyhead != vi.keytail)
                    {
                    vsthw.gpip &= 0xEF;
                    b          &= 0xEF;
                    }
                else
                    {
                    vsthw.gpip |= 0x10;
                    b          |= 0x10;
                    }
                return b | 0x10;
                }

            case 0x0B:
                // interrupt pending register, set timer A bit (??)
                return vsthw.ipra | 0x20;

            case 0x21:
                // timer B data register, used by TOS to detect VBI,
                // so it has to return this funky pattern

                vsthw.tbcnt--;
                vsthw.tbcnt &= 1023;

                if (vsthw.tbcnt < 620)
                    return 0;

                return ((vsthw.tbcnt - 620) >> 1) & 255;

            case 0x23:
                // timer C data register
                // need to count down every 2 reads

                vsthw.tccnt--;
                return (vsthw.tccnt >> 2) & 255;

            case 0x2D:
                // USART transmitter status register

                vsthw.tsr &= 0x09;  // mask everything but enable and break bits
                vsthw.tsr |= 0x80;  // set bit to indicate no char being sent
                return vsthw.tsr;

            case 0x2F:
                // USART data register (read from modem port)

                if (vi.cchserial)
                    {
                    // if >1 characters, don't clear Buffer Full bit

                    if (vi.cchserial == 1)
                        vsthw.rsr &= 1; // REVIEW: is this right??

                    vi.cchserial -= CchSerialRead(&vsthw.udr, 1);
                    }

                return vsthw.udr;
                }

            return vsthw.rgbMFP[ea >> 1];
            }
        else if ((ea == 0x8D) && (vrgf.fSTE > 1))
            {
            return 0x20;
            }
        else if ((ea >= 0x80) && (ea < 0xB0) && (ea & 1) && (vrgf.fSTE > 1))
            {
            return 0;
            }
        else if ((ea >= 0x00FFFA40) && (ea < 0x00FFFA80))
            {
            DebugStr("reading byte at FPU at address %08X\n", ea);
            }
        break;

    case 0x00FFFC00: // ACIA

        //
        //  ACIA - keyboard and MIDI
        //
        //  FFFC00 - keyboard status register
        //  FFFC02 - keyboard data read/write
        //  FFFC04 - MIDI status register
        //  FFFC06 - MIDI data read/write
        //

        switch (ea & 255)
            {
        case 0x00:
            return vsthw.KeyCtrl;

        case 0x04:
            return vsthw.MIDICtrl;

        case 0x02:
            if (1 || vi.keyhead == vi.keytail)
                {
                vsthw.KeyCtrl &= 0x7F;  // clear high bit
                vsthw.gpip |= 0x10;     // set bit 4
                }
            return vsthw.KeyData;

        case 0x06:
            vsthw.MIDICtrl &= 0x7F;  // clear high bit
            return vsthw.MIDIData;

        default:
            return 0xFF; // ACIA doesn't bus error
            }

    case 0x00FFFD00: // ????, doesn't bus error
        return 0xFF;
        }

    return 0x80000002; // default is bus error
}


//
// ReadHWWord(ea)
//
// Handler for emulating Atari ST hardware word reads.
// In most cases will just call ReadHWByte.
//
// Returns:
//  $000000xx - success, xx is return value
//  $8000xxxx - xxxx is trap value
//

ULONG WordFromOB(ULONG l)
{
    return 0xFF00 | l;
}

ULONG WordFromEB(ULONG l)
{
    if (l & 0xFFFF0000)
        return l;

    return 0xFF | (l << 8);
}

static ULONG __cdecl st_ReadHWWord(ULONG ea, WORD *pw)
{
    ea = ea & vsthw.lAddrMask;

    if ((ea & 1) && (vmCur.bfCPU < cpu68020))
        {
        return 0x80000003; // odd address error
        }

    if (ea < (vi.cbRAM[0] - 1))
        {
        // it's just RAM

        return *(unsigned short *)(vi.pbRAM[0] - 1 - (ea - vi.eaRAM[0]));
        }

    if (ea < 0x00400000)
        {
        // non-RAM below 4 meg returns garbage

        return 0xFF;
        }

    if (vi.cbROM[0] &&
            (ea >= vi.eaROM[0]) && (ea < vi.eaROM[0] + vi.cbROM[0] - 1))
        {
        // it's just TOS ROM

        return *(unsigned short *)(vi.pbROM[0] - 1 - (ea - vi.eaROM[0]));
        }

    if (vi.cbROM[1] &&
            (ea >= vi.eaROM[1]) && (ea < vi.eaROM[1] + vi.cbROM[1] - 1))
        {
        // it's just cartridge ROM

        return *(unsigned short *)(vi.pbROM[1] - 1 - (ea - vi.eaROM[1]));
        }

#if TRACEHW
    printf("ReadHWWord: %08X\n", ea);
#endif

    if ((vpregs->SR & 0x2000) == 0)
        {
        // oh oh, not in supervisor mode!

        return 0x80000008; // priviledge violation
        }

    switch (ea & 0x00FFFF00)
        {
    default:
//        _asm { int 3 } ;
        break;

    case 0x00FF8000: // memory configuration

        // $FF8000 reads $FF0x, all others bus error

        if (ea == 0x00FF8000)
            {
            // some magic memory config, 5 on 1040ST, return 10?
            return 0xFF00 | 10;
            }
        break;

    case 0x00FF8200: // video chip
        ea &= 255;

        if (ea >= 0x40 && ea < 0x60)
            {
            // do a byte swap

            return vsthw.wcolor[(ea>>1)-0x20];
            }

        switch (ea & 255)
            {
        case 0x00:
        case 0x02:
        case 0x04:
        case 0x06:
        case 0x08:
        case 0x0C:
        case 0x0E:
            return WordFromOB(st_ReadHWByte(ea + 1, NULL));

        case 0x0A:
            return 0xFCFF;  // syncmode

        case 0x60:
        case 0x62:
            return WordFromEB(vsthw.shiftmd);

        case 0x64:
            if (vrgf.fSTE)
                return WordFromEB(vsthw.horscroll);
            break;

        case 0x7E:
            // STacy display state, checked by Maccel 3.0

            return 0x6100;
            }
        break;

    case 0x00FF8600: // disk drive

        switch (ea & 255)
            {
        case 0x08:
        case 0x0A:
        case 0x0C:
            return WordFromOB(st_ReadHWByte(ea + 1, NULL));

        case 0x04:
            // FDC/HDC registers

            if (vsthw.idiskreg < 5)
                {
                if ((vsthw.idiskreg == 0) && vsthw.fMediaChange)
                    return vsthw.rgwFDC[0] ^ 0x40; // toggle write protect bit

                return vsthw.rgwFDC[vsthw.idiskreg];
                }
            return 0xFF;

        case 0x06:
            // FDC/HDC status

            return vsthw.dmastat;
            }

        break;

    case 0x00FF8700: // AHDI 5.0 accesses $FF8781, TT DMA
        if (vrgf.fSTE > 1)
            {
//            _asm { int 3 } ;
            return 0;
            }
        break;

    case 0x00FF8800: // sound chip

        if ((ea & 3) == 0)
            {
            return WordFromEB(FReadYamaha());
            }

        return 0xFF; // sound chip doesn't bus error

    case 0x00FF8900: // stereo sound chip on STE
        if (vrgf.fSTE && ea < 0xFF8940)
            {
            if (ea >= 0xFF8921)
                return 0;

            if (ea == 0xFF8924)
                return 0x0FFE;

            ea = ea & 0x3E;

            return vsthw.rgwSTESound[ea>>1];
            }
        break;

    case 0x00FF8A00: // blitter
        if (vmCur.fBlitter && ea < 0xFF8A3E)
            {
            ea = ea & 0x3E;

            return vsthw.blitter.rgwBlitter[ea>>1];
            }
        break;

    case 0x00FF8E00: // Mega STE / TT SCC ports
        if (vrgf.fSTE > 1)
            {
//            _asm { int 3 } ;
            return 0;
            }
        break;

    case 0x00FF9200: // paddle ports (NYI)
        if (vrgf.fSTE)
            {
            switch (ea & 255)
                {
            default:
                break;

            case 0x00:
            case 0x02:
            case 0x10:
            case 0x12:
            case 0x14:
            case 0x16:
            case 0x20:
            case 0x22:
                return 0xFF;
                }
            }
        break;

    case 0x00FF9800: // Falcon color registers
    case 0x00FF9900: // Falcon color registers
    case 0x00FF9A00: // Falcon color registers
    case 0x00FF9B00: // Falcon color registers
        break;

    case 0x00FFFA00: // MFP

        // MFP is odd byte access from $FFFA01 to $FFFA3F.
        // All other bytes bus error.
        //
        // On TT, $FFFA81 to $FFFABF is the second MFP (NYI)

        if (ea < 0x00FFFA40)
            {
            return WordFromOB(st_ReadHWByte(ea + 1, NULL));
            }
        else if ((ea >= 0x00FFFA80) && (ea < 0x00FFFAB0) && (ea & 1) && (vrgf.fSTE > 1))
            {
            return 0;
            }
        else if ((ea >= 0x00FFFA40) && (ea < 0x00FFFA80))
            {
            // 68881 in peripheral mode (Mega ST / Mega STE)
            // Bus error for now

            DebugStr("reading word from FPU at address %08X\n", ea);
            return 0x80000002;
            }
        break;

    case 0x00FFFC00: // ACIA
        return WordFromOB(st_ReadHWByte(ea + 1, NULL));

    case 0x00FFFD00: // ????, doesn't bus error
        return 0xFFFF;
        }

    return 0x80000002; // default is bus error
}


//
// WriteHWByte(ea,pb)
//
// Handler for emulating Atari ST hardware byte writes.
//
// Returns: 0 = success, non-zero = trap number
//

static ULONG __cdecl st_WriteHWByte(ULONG ea, BYTE b)
{
    ea = ea & vsthw.lAddrMask;

    if (ea < vi.cbRAM[0])
        {
        // it's just RAM

        *(BYTE *)(vi.pbRAM[0] - (ea - vi.eaRAM[0])) = b;
        return 0;
        }

    if (ea < 0x00400000)
        {
        // non-RAM below 4 meg does nothing

        return 0;
        }

#if TRACEHW
    printf("WriteHWByte: %08X\n", ea, b);
#endif

    if ((vpregs->SR & 0x2000) == 0)
        {
        // oh oh, not in supervisor mode!

        return 0x80000008; // priviledge violation
        }

    switch (ea & 0x00FFFF00)
        {
    default:
//        _asm { int 3 } ;
        break;

    case 0x00FF8000: // memory configuration
        if (ea == 0x00FF8001)
            {
            // not much to do!
            return 0;
            }
        break;

    case 0x00FF8200: // video chip

        ea &= 255;

        if (ea >= 0x40 && ea < 0x60)
            {
            // do a byte swap

            if (vsthw.bColor[(ea^1)-0x40] != b)
                {
                // color register changing

                vsthw.bColor[(ea^1)-0x40] = b;
                MarkAllPagesDirty();
                }
            return 0;
            }

        switch (ea & 255)
            {
        case 0x01:
            if (b != vsthw.dbaseh)
                {
                vsthw.dbaseh = b;
                MarkAllPagesDirty();
                }
            return 0;

        case 0x03:
            if (b != vsthw.dbasel)
                {
                vsthw.dbasel = b;
                MarkAllPagesDirty();
                }
            return 0;

        case 0x05:
        case 0x07:
        case 0x09:
        case 0x0A:
        case 0x0B:
            return 0;

        case 0x0D:
            if (vrgf.fSTE)
                {
                if (b != vsthw.dbase0)
                    {
                    vsthw.dbase0 = b;
                    MarkAllPagesDirty();
                    }
                }
            return 0;

        case 0x0F:
            if (vrgf.fSTE)
                {
                if (vsthw.scanskip != b)
                    {
                    vsthw.scanskip = b;
                    MarkAllPagesDirty();
                    }
                }
            return 0;

        case 0x60:
            if (vsthw.fMono)
                b = 2;
            else
                b &= 3;

            if (b != vsthw.shiftmd)
                {
                vsthw.shiftmd = b;

#if 0
                if (!vsthw.fMono && b < 2)
#endif
                    EnterMode(1);
                }
            return 0;

        case 0x62:
            if (vrgf.fSTE > 1)
                return 0;

        case 0x65:
            if (vrgf.fSTE)
                {
                if (vsthw.horscroll != b)
                    {
                    vsthw.horscroll = b;
                    MarkAllPagesDirty();
                    }
                return 0;
                }
            break;
            }
        break;

    case 0x00FF8600: // disk drive

        switch (ea & 255)
            {
        case 0x09:
            vsthw.dmahi = b;
            return 0;

        case 0x0B:
            vsthw.dmamid = b;
            return 0;

        case 0x0D:
            vsthw.dmalo = b;
            return 0;
            }

        break;

    case 0x00FF8700: // AHDI 5.0 accesses $FF8781, TT DMA
        if (vrgf.fSTE > 1)
            {
//            _asm { int 3 } ;
            return 0;
            }
        break;

    case 0x00FF8800: // sound chip

        FWriteYamaha(ea & 3, b);
        return 0;    // sound chip doesn't bus error

    case 0x00FF8900: // stereo sound chip on STE
        if (vrgf.fSTE && ea < 0xFF8940)
            {
            ea = (ea ^ 1) & 0x3F;   // byte swap

            vsthw.rgbSTESound[ea] = b;
            return 0;
            }
        break;

    case 0x00FF8A00: // blitter
        if (vmCur.fBlitter && ea < 0xFF8A3E)
            {
            ea = (ea ^ 1) & 0x3F;   // byte swap

            vsthw.blitter.rgbBlitter[ea] = b;

            if ((ea == 0x3D) && (b & 0x80))  // $FF8A3C
                {
                // start blitter operation

                DoBlitter((struct _blitregs *)&vsthw.blitter);
                vsthw.blitter.rgbBlitter[ea] = b & 0x7F;
                }

            return 0;
            }
        break;

    case 0x00FF8E00: // Mega STE / TT SCC ports
        if (vrgf.fSTE > 1)
            {
//            _asm { int 3 } ;
            return 0;
            }
        break;

    case 0x00FF9200: // paddle ports (NYI)
        break;

    case 0x00FF9800: // Falcon color registers
    case 0x00FF9900: // Falcon color registers
    case 0x00FF9A00: // Falcon color registers
    case 0x00FF9B00: // Falcon color registers
        break;

    case 0x00FFFA00: // MFP

        if ((ea < 0x00FFFA40) && (ea & 1))
            {
            ea &= 255;

            if (ea >= 0x30) // unused bytes
                return 0;

            // MFP registers are on odd addresses

            switch (ea)
                {
            case 0x01:
                return 0;   // read only register

            case 0x1D:
                // timer C & D control

                if (b != vsthw.tcdcr)
                    goto L_setbaud;
                break;

            case 0x23:
                // timer C data

                vsthw.tccnt = (b<<2) + 2;
                break;

            case 0x25:
                // timer D data

                if (b != vsthw.tddr)
                    goto L_setbaud;
                break;

            case 0x29:
                // USART control register

                if (b != vsthw.ucr)
                    goto L_setbaud;
                break;

            case 0x2D:
                // USART transmitter status register

                if (b != vsthw.ucr)
                    {
L_setbaud:
                    FSetBaudRate(vsthw.tsr, vsthw.ucr, vsthw.tddr, vsthw.tcdcr);
                    }
                break;

            case 0x2F:
                // USART data register (write to modem port)

                FWriteSerialPort(b);
                return 0;
                }

            vsthw.rgbMFP[ea >> 1] = b;
            return 0;
            }
        else if ((ea >= 0x00FFFA80) && (ea < 0x00FFFAB0) && (ea & 1) && (vrgf.fSTE > 1))
            {
            return 0;
            }
        else if ((ea >= 0x00FFFA40) && (ea < 0x00FFFA80))
            {
            // 68881 in peripheral mode (Mega ST / Mega STE)
            // Bus error for now

            DebugStr("writing byte to FPU at address %08X\n", ea);
            return 0x80000002;
            }
        break;

    case 0x00FFFC00: // ACIA

        //
        //  ACIA - keyboard and MIDI
        //
        //  FFFC00 - keyboard status register
        //  FFFC02 - keyboard data read/write
        //  FFFC04 - MIDI status register
        //  FFFC06 - MIDI data read/write
        //

        if (ea == 0x00FFFC02)
            FWriteIKBD(b);

        return 0; // ACIA doesn't bus error

    case 0x00FFFD00: // ????, doesn't bus error
        return 0;
        }

    return 0x80000002; // default is bus error
}

static ULONG __cdecl st_WriteHWWord(ULONG ea, unsigned short w)
{
    PDI pdi;

    ea = ea & vsthw.lAddrMask;
    w &= 65535;

    if ((ea & 1) && (vmCur.bfCPU < cpu68020))
        {
        return 0x80000003; // odd address error
        }

    if (ea < (vi.cbRAM[0] - 1))
        {
        // it's just RAM

        *(WORD *)(vi.pbRAM[0] - 1 - (ea - vi.eaRAM[0])) = w;
        return 0;
        }

    if (ea < 0x00400000)
        {
        // non-RAM below 4 meg does nothing

        return 0;
        }

#if TRACEHW
    printf("WriteHWWord: %08X\n", ea, w);
#endif

    if ((vpregs->SR & 0x2000) == 0)
        {
        // oh oh, not in supervisor mode!

        return 0x80000008; // priviledge violation
        }

    switch (ea & 0x00FFFF00)
        {
    default:
//        _asm { int 3 } ;
        break;

    case 0x00FF8000: // memory configuration
        if (ea == 0x00FF8000)
            {
            // not much to do!
            return 0;
            }
        break;

    case 0x00FF8200: // video chip

        ea &= 255;

        if (ea >= 0x40 && ea < 0x60)
            {
            if (vsthw.wcolor[(ea>>1)-0x20] != w)
                {
                // color register changing

                vsthw.wcolor[(ea>>1)-0x20] = w;
                MarkAllPagesDirty();
                }
            return 0;
            }

        switch (ea & 255)
            {
        case 0x00:
        case 0x02:
        case 0x04:
        case 0x06:
        case 0x08:
        case 0x0A:
        case 0x0C:
        case 0x0E:
        case 0x64:
            // OB registers
            return st_WriteHWByte(ea+1,(BYTE)w);
            break;

        case 0x60:
        case 0x62:
            return st_WriteHWByte(ea,(BYTE)(w>>8));

        case 0x7E:
            // STacy display state, checked by Maccel 3.0

            return 0;
            }
        break;

    case 0x00FF8600: // disk drive

        switch (ea & 255)
            {
        case 0x0E:
            return 0; //???

        case 0x04:
            // FDC/HDC registers

            vsthw.diskctrl = (BYTE)w;

            if (vsthw.fHDC)
                return 0;       // HD operation in progress

            // floppy disk operation

            if (vsthw.idiskreg > 4)
                {
                // bogus register select - do what??

                return 0;
                }
            else if (vsthw.idiskreg != 0)
                {
                vsthw.rgwFDC[vsthw.idiskreg] = w;
                return 0;
                }

            // writing to FDC command register
            // set cmdreg = $A0 assuming success

            vsthw.cmdreg = 0xA0;

            switch (w & 0xF0)
                {
            case 0x00:
                // RESTORE - seek to track 0

                vsthw.trackreg = 0;
                goto seek;

            case 0x10:
                // SEEK

                vsthw.trackreg = vsthw.datareg;
                if (vsthw.trackreg & 0x8000)
                    vsthw.trackreg = 0;
seek:
                vsthw.cmdreg = (vsthw.trackreg == 0) ? 0xA4 : 0xA0;
                break;

            case 0x20:
            case 0x40:
            case 0x60:
                break;

            case 0x30:
                // STEP - do what?
                break;

            case 0x50:
                // STEP IN

                vsthw.trackreg++;
                break;

            case 0x70:
                // STEP OUT

                vsthw.trackreg--;
                if (vsthw.trackreg & 0x8000)
                    vsthw.trackreg = 0;

                goto seek;

            case 0x80:
                // READ SECTOR

                vsthw.countreg = 1;
                // fall through

            case 0x90:
                // READ SECTORS

                {
                pdi = vi.rgpdi[vsthw.disksel];

                if (pdi == NULL)
                    goto fail_fdc;

                pdi->lpBuf = (BYTE *)&vsthw.rgbDiskBuffer;
                pdi->track = vsthw.trackreg & 0xFF;
                pdi->side  = vsthw.sidereg;
                pdi->sec   = vsthw.secreg - 1; // 1 based -> 0 based
                pdi->count = vsthw.countreg;

                if (FRawDiskRWPdi(pdi, fFalse /* fWrite=0 */))
                    {
                    HostToGuestDataCopy(vsthw.DMAbasis, vsthw.rgbDiskBuffer,
                         ACCESS_PHYSICAL_WRITE, vsthw.countreg * 512);

                    vsthw.cmdreg = 0x80;
                    vsthw.dmastat = 3;  // DMA OK
                    }
                else
                    {
fail_fdc:
                    vsthw.cmdreg = 0x90;
                    vsthw.dmastat = 1;
                    }
                }
                break;

            case 0xA0:
                // WRITE SECTOR

                vsthw.countreg = 1;
                // fall through

            case 0xB0:
                // WRITE SECTORS

                if (GuestToHostDataCopy(vsthw.DMAbasis, vsthw.rgbDiskBuffer,
                     ACCESS_PHYSICAL_READ, vsthw.countreg * 512))
                {
                PDI pdi = vi.rgpdi[vsthw.disksel];

                if (pdi == NULL)
                    goto fail_fdc;

                pdi->lpBuf = (BYTE *)&vsthw.rgbDiskBuffer;
                pdi->track = vsthw.trackreg & 0xFF;
                pdi->side  = vsthw.sidereg;
                pdi->sec   = vsthw.secreg - 1; // 1 based -> 0 based
                pdi->count = vsthw.countreg;

                if (FRawDiskRWPdi(pdi, fTrue /* fWrite=1 */))
                    {
                    vsthw.cmdreg = 0x80;
                    vsthw.dmastat = 3;  // DMA OK
                    }
                else
                    goto fail_fdc;
                }
                else
                    goto fail_fdc;
                break;
                }
            return 0;

        case 0x06:
            // DMA control register

            vsthw.dmacmd = w;

            switch (w & 255)
                {
            case 0x80:
                vsthw.idiskreg = 0; // select stat/cmd register
                vsthw.fHDC = 0;     // end of hard disk operation
                break;

            case 0x82:
                vsthw.idiskreg = 1; // select track register
                break;

            case 0x84:
                vsthw.idiskreg = 2; // select sector register
                break;

            case 0x86:
                vsthw.idiskreg = 3; // select data register
                break;

            case 0x90:
                vsthw.idiskreg = 4; // select sector count register
                break;

            case 0x88:
                vsthw.fHDC = 1;     // start hard disk operation
                break;

            case 0x8A:              // next byte of hard disk op
            case 0x0A:              // last byte of hard disk op
                vsthw.idiskreg = 0;

                if (vsthw.fHDC != 7)
                    {
                    // stuff in next byte

                    vsthw.rgbHDC[vsthw.fHDC++] = (BYTE)vsthw.diskctrl;
                    break;
                    }

                // All 7 bytes arrived, start hard disk operation

                vsthw.fHDC = 0;         // reset index counter

                // Check that the SCSI device is mapped to something

                pdi = vi.rgpdi[((vsthw.byte0 & 0xE0) >> 5) + 2];

                // special case SCSI device 0 (C: drive), we fake it

                if (pdi == NULL)
                  if (((vsthw.byte0 & 0xE0) >> 5) && vmCur.fUseVHD)
                    goto badHD;

                if (vsthw.byte1 & 0xE0) // check for drive 0
                    goto badHD;

                // perform op on controller 0 drive 0

                switch (vsthw.byte0 & 0x1F)
                    {
                default:
                    // unknown command, pretend it's good (or bad??)

                    goto badHD;

                case 0x00:
                    // test unit ready

goodHD:
                    vsthw.cmdreg = 0;
                    vsthw.idiskreg = 0;
                    vsthw.dmastat = 0; // DMA OK
                    break;

badHD:
                    vsthw.cmdreg = 0xFF;
                    vsthw.idiskreg = 0;
                    vsthw.dmastat = 1;
                    break;

                case 0x01:
                    // restore

                    goto goodHD;

                case 0x03:
                    // request sense - return 4 zeroes in buffer

                    {
                    ULONG i = vsthw.DMAbasis;

                    if (i >= vi.cbRAM[0])
                        goto badHD;

                    vmPokeL(i, 0);    // 4 bytes of 0
                    }
                    goto goodHD;

                case 0x04:
                case 0x15:
                    // used during HDX formatting, just pretend it's ok

                    goto goodHD;

                case 0x08:
                    // read sector

                    {
                    ULONG sector = (((vsthw.byte1 & 0x1F) | (vsthw.byte0 & 0xE0)) << 16) |
                        ((vsthw.byte2 & 0xFF) << 8) | (vsthw.byte3 & 0xFF);

                    if (FRWVirtualDisk(fFalse,      // fWrite=0
                        vsthw.DMAbasis,             // buffer
                        sector,                     // sector
                        vsthw.byte4)                // count
                       )
                        goto badHD;
                    }
                    goto goodHD;

                case 0x0A:
                    // write sector

                    {
                    ULONG sector = (((vsthw.byte1 & 0x1F) | (vsthw.byte0 & 0xE0)) << 16) |
                        ((vsthw.byte2 & 0xFF) << 8) | (vsthw.byte3 & 0xFF);

                    if (FRWVirtualDisk(fTrue,       // fWrite=1
                        vsthw.DMAbasis,             // buffer
                        sector,                     // sector
                        vsthw.byte4)                // count
                       )
                        goto badHD;
                    }
                    goto goodHD;

                case 0x0B:
                    // seek

                    goto goodHD;

                case 0x12:
                    // inquiry
                    //
                    // sends 5 byte header, plus 31 or 49 bytes of data
                    // first 4 bytes are 0, 5th byte is $1F (31 bytes of stuff to follow)
                    // first 3 bytes are 0, 8 bytes of ID, 16 bytes of product ID

                    {
                    char *pch = "BRASOFT GEMULATOR DEVICE";
                    ULONG i = vsthw.DMAbasis;

                    vmPokeL(i, 0);    // 4 bytes of 0
                    i += 4;

                    vmPokeB(i++, 31); // indicates 31 bytes of data to come
                    vmPokeB(i++, 0);
                    vmPokeB(i++, 0);
                    vmPokeB(i++, 0);

                    // pdi may be NULL for faked C: drive

                    if (pdi && pdi->szVolume[0])
                        pch = &pdi->szVolume[0];

                    while (*pch)
                        vmPokeB(i++, *pch++);
                    }

                    goto goodHD;

                case 0x1A:
                    // mode sense
                    //
                    // sends 12 bytes of data, described as follows (from email by Julian Reschke)
                    //
                    // SCSI-Tool tries to get the number and size of blocks out of the
                    // MODE SENSE data. I would recommend to return 12 bytes:
                    //
                    // Byte  Contents
                    //
                    //   0  data length == 8 (8 + header of 4 bytes)
                    //   1  medium type == 0
                    //   2  specific == 0 (bit 7 would be set for write protected media)
                    //   3  block descriptor length == 8 (one block descriptor)
                    //
                    //   4  density == 0 (default)
                    //   5  number of blocks, bits 16..23
                    //   6  number of blocks, bits 8..15
                    //   7  number of blocks, bits 0..7
                    //   8  reserved == 0
                    //   9  block length, bits 16..23 == 0
                    //  10  block length, bits 8..15 == 2
                    //  11  block length, bits 0..7 == 0

                    {
                    ULONG i = vsthw.DMAbasis;
                    unsigned int cSecMax = 0x00040000, cbSec = 512;

                    if (i >= vi.cbRAM[0])
                        goto badHD;

                    // pdi may be NULL for faked C: drive

                    if (pdi && pdi->cbSector)
                        cbSec = pdi->cbSector;

                    if (pdi && pdi->size && pdi->size >= 20000)
                        cSecMax = (pdi->size * 2) / (cbSec / 512);

                    vmPokeB(i++, 8);
                    vmPokeB(i++, 0);
                    vmPokeB(i++, 0);
                    vmPokeB(i++, 8);

                    vmPokeB(i++, cSecMax >> 24);
                    vmPokeB(i++, cSecMax >> 16);
                    vmPokeB(i++, cSecMax >> 8);
                    vmPokeB(i++, cSecMax);

                    vmPokeB(i++, 0);
                    vmPokeB(i++, 0);
                    vmPokeB(i++, cbSec >> 8);
                    vmPokeB(i++, cbSec);
                    }

                    goto goodHD;

                case 0x1B:
                    // park

                    goto goodHD;
                    }
                break;
                }
            return 0;

#if 0
        case 0x08:
            vsthw.dmahi = (BYTE)w;
            return 0;

        case 0x0A:
            vsthw.dmamid = (BYTE)w;
            return 0;

        case 0x0C:
            vsthw.dmalo = (BYTE)w;
            return 0;
#endif
            }

        break;

    case 0x00FF8700: // AHDI 5.0 accesses $FF8781, TT DMA
        if (vrgf.fSTE > 1)
            {
//            _asm { int 3 } ;
            return 0;
            }
        break;

    case 0x00FF8800: // sound chip

        FWriteYamaha(ea & 3, w>>8);
        return 0;    // sound chip doesn't bus error

    case 0x00FF8900: // stereo sound chip on STE
        if (vrgf.fSTE && ea < 0xFF8940)
            {
            ea = ea & 0x3E;   // byte swap

            vsthw.rgwSTESound[ea>>1] = w;
            return 0;
            }
        break;

    case 0x00FF8A00: // blitter
        if (vmCur.fBlitter && ea < 0xFF8A3E)
            {
            ea = ea & 0x3E;   // byte swap

            vsthw.blitter.rgwBlitter[ea>>1] = w;

            if ((ea == 0x3C) && (w & 0x8000))  // $FF8A3C
                {
                // start blitter operation

                DoBlitter((struct _blitregs *)&vsthw.blitter);
                vsthw.blitter.rgwBlitter[ea>>1] = w & 0x7FFF;
                }

            return 0;
            }
        break;

    case 0x00FF8E00: // Mega STE / TT SCC ports
        if (vrgf.fSTE > 1)
            {
//            _asm { int 3 } ;
            return 0;
            }
        break;

    case 0x00FF9200: // paddle ports (NYI)
        break;

    case 0x00FF9800: // Falcon color registers
    case 0x00FF9900: // Falcon color registers
    case 0x00FF9A00: // Falcon color registers
    case 0x00FF9B00: // Falcon color registers
        break;

    case 0x00FFFA00: // MFP

        if ((ea < 0x00FFFA40) && (ea & 1))
            {
            // MFP registers are on odd addresses

            return st_WriteHWByte(ea+1,(BYTE)w);
            }
        else if ((ea >= 0x00FFFA80) && (ea < 0x00FFFAB0) && (ea & 1) && (vrgf.fSTE > 1))
            {
            return 0;
            }
        else if ((ea >= 0x00FFFA40) && (ea < 0x00FFFA80))
            {
            // 68881 in peripheral mode (Mega ST / Mega STE)
            // Bus error for now

            DebugStr("writing word to FPU at address %08X\n", ea);
            return 0x80000002;
            }
        break;

    case 0x00FFFC00: // ACIA

        //
        //  ACIA - keyboard and MIDI
        //
        //  FFFC00 - keyboard status register
        //  FFFC02 - keyboard data read/write
        //  FFFC04 - MIDI status register
        //  FFFC06 - MIDI data read/write
        //

        return st_WriteHWByte(ea,(BYTE)(w>>8));

    case 0x00FFFD00: // ????, doesn't bus error
        return 0;
        }

    return 0x80000002; // default is bus error
}


static HRESULT __cdecl ReadHWByte(ULONG ea, BYTE *pb)
{
    ULONG b = st_ReadHWByte(ea, pb);

    if (b & 0x80000000)
        {
        *pb = 0;
        return ERROR_INVALID_ACCESS;
        }

    *pb = (BYTE)(b & 255);
    return NO_ERROR;
}


static HRESULT __cdecl ReadHWWord(ULONG ea, WORD *pw)
{
    ULONG w = st_ReadHWWord(ea, pw);

    if (w & 0x80000000)
        {
        *pw = 0;
        return ERROR_INVALID_ACCESS;
        }

    *pw = (WORD)(w & 0xFFFF);
    return NO_ERROR;
}

static HRESULT __cdecl ReadHWLong(ULONG ea, ULONG *pl)
{
    WORD  w = 0;
    ULONG l = 0;

    if (NO_ERROR == ReadHWWord(ea, &w))
        {
        l = w;

        if (NO_ERROR == ReadHWWord(ea + 2, &w))
            {
            *pl = (l << 16) | (w & 0xFFFF);

            return NO_ERROR;
            }
        }

    *pl = 0;

    return ERROR_INVALID_ACCESS;
}

static HRESULT __cdecl WriteHWByte(ULONG ea, BYTE *pb)
{
    ULONG l = st_WriteHWByte(ea, *pb);

    if (l & 0x80000000)
        {
        return ERROR_INVALID_ACCESS;
        }

    return NO_ERROR;
}


static HRESULT __cdecl WriteHWWord(ULONG ea, WORD *pw)
{
    ULONG l = st_WriteHWWord(ea, *pw);

    if (l & 0x80000000)
        {
        return ERROR_INVALID_ACCESS;
        }

    return NO_ERROR;
}

static HRESULT __cdecl WriteHWLong(ULONG ea, unsigned *pl)
{
    WORD w = (WORD)(*pl >> 16);

    if (NO_ERROR != WriteHWWord(ea, &w))
        {
        return ERROR_INVALID_ACCESS;
        }

    w = (WORD)(*pl);

    if (NO_ERROR != WriteHWWord(ea + 2, &w))
        {
        return ERROR_INVALID_ACCESS;
        }

    return NO_ERROR;
}


//
// MapAddress[ForWrite]
//
// Return a host pointer to a the readable/writeable guest physical address 
// or NULL if the address is invalid or memory mapped I/O.
//

BYTE * __cdecl MapAddress(ULONG ea)
{
    ea = ea & vsthw.lAddrMask;

    if (ea < vi.cbRAM[0])
        {
        // it's just RAM

        return vi.pbRAM[0] - ((ea - vi.eaRAM[0]) & vsthw.lAddrMask);
        }

    if ((ea >= vi.eaROM[0]) && (ea < vi.eaROM[0] + vi.cbROM[0]))
        {
        // it's just TOS ROM

        return vi.pbROM[0] - ((ea - vi.eaROM[0]) & vsthw.lAddrMask);
        }

    if ((ea >= vi.eaROM[1]) && (ea < vi.eaROM[1] + vi.cbROM[1]))
        {
        // it's just cartridge ROM

        return vi.pbROM[1] - ((ea - vi.eaROM[1]) & vsthw.lAddrMask);
        }

    // bogus addresses return NULL so they fault if dereferenced

    return NULL;
}


BYTE * __cdecl MapAddressForWrite(ULONG ea)
{
    ea = ea & vsthw.lAddrMask;

    if (ea < vi.cbRAM[0])
        {
        // it's just RAM

        return vi.pbRAM[0] - ((ea - vi.eaRAM[0]) & vsthw.lAddrMask);
        }

    // bogus addresses return NULL so they fault if dereferenced

    return NULL;
}


//
// Initialize Atari ST hardware registers. Most will be cleared.
//

ULONG __cdecl Reset()
{
    memset(&vsthw, 0, sizeof(vsthw));
    memset(&vmachw, 0, sizeof(vmachw));

    // set address mask and invalidate the whole page table

    vsthw.lAddrMask = 0x00FFFFFF; // 24-bit addressing
    InvalidatePages(0x00000000, 0xFFFFFFFF);

    vsthw.fMono = FMonoFromBf(vmCur.bfMon);

    if (vsthw.fMono)
        {
        vsthw.planes = 1;
        vsthw.xpix = 640;
        vsthw.ypix = 400;
        }
    else // color Atari ST defaults to lo rez
        {
        vsthw.planes = 4;
        vsthw.xpix = 320;
        vsthw.ypix = 200;
        }

    CreateNewBitmap();
    DisplayStatus();

        {
        vsthw.dbase = 0x78000;
        vsthw.gpip = 0x5F;

        vsthw.KeyCtrl = 2;
        vsthw.MIDICtrl = 2;

        vmPokeW(0xFF8240,0x0FFF);   // color0 is white
        }

    if (!FInitSerialPort(vmCur.iCOM))
        vmCur.iCOM = 0;
    if (!InitPrinter(vmCur.iLPT))
        vmCur.iLPT = 0;

    InitSound(500);
    InitMIDI();
    FResetIKBD();
    FResetGemDosHooks();
    vi.fVMCapture = TRUE;

    return TRUE;
}


//
// VPollInterrupts
//
// Called by asm based PollInterrupts. Checks for any MFP interrupts that
// should go off, and sets the appropriate bits in vrgf.
// Returns the vector number.
//
// REVIEW: needs to be optimized to handle multiple interrupts
//

ULONG BitFromW(unsigned int i)
{
    ULONG bit = 0;
    ULONG lRet = 0;

    Assert(i);

    for (bit = 0; i; bit++, i>>=1)
        {
        if (i & 1)
            lRet = bit;
        }

    return lRet;
}

ULONG __cdecl VPollInterrupts()
{
    ULONG vector;

    if (!vi.fExecuting)
        return 0;

    // check for serial interrupts

    if (vi.cchserial && (vsthw.iera & 0x10)) // serial interrupt enabled?
        {
        vsthw.rsr &= 0x81;  // clear out bogus bits

        if (vsthw.rsr & 1)  // Receive Enable Bit is set in RSR
            {
            vsthw.rsr |= 0x80;  // set Buffer Full bit
            vsthw.isra |= 0x10;
            vrgf.fReceiveBuffFull = 1;

            goto L_check;   // serial overrides any key/mouse
            }
        }

    // check for keyboard/mouse packet

    if (vi.keyhead != vi.keytail)
        {
        if ((vsthw.ierb & 0x40)  // make sure mouse interrupt enabled
            && (!vrgf.fKeyMIDI))
            {
            vrgf.fKeyMIDI = 1;

            vsthw.KeyCtrl |= 0x81;
            vsthw.KeyData = vi.rgbKeybuf[vi.keytail++];
            vi.keytail &= 1023;
            vsthw.isrb |= 0x40;
            }
        }

L_check:
    vrgf.A &= vsthw.iera & vsthw.imra;
    vrgf.B &= vsthw.ierb & vsthw.imrb;

    if (vrgf.A && ((vpregs->SR & 0x0700) < 0x600))
        {
        vector = BitFromW(vrgf.A);
        vrgf.A &= ~(1 << vector);
        return (6 << 24) | (vector + (vsthw.vr & 0xF0) + 0x08);
        }

    if (vrgf.B && ((vpregs->SR & 0x0700) < 0x600))
        {
        vector = BitFromW(vrgf.B);
        vrgf.B &= ~(1 << vector);
        return (6 << 24) | (vector + (vsthw.vr & 0xF0));
        }

    if (vrgf.fVBIPending && ((vpregs->SR & 0x0700) < 0x400))
        {
        vrgf.fVBIPending = 0;
        return (4 << 24) | 0x1C;    // VBI vector, IPL 4
        }

    return 0;
}

ULONG __cdecl WinMsg(HWND hWnd, UINT msg, WPARAM uParam, LPARAM lParam)
{
    return FWinMsgST(hWnd, msg, uParam, lParam);
}

#endif // ATARIST

